/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Base/Allocate.h,v $
 * $Date: 2003/05/13 18:41:54 $
 *
 Description:
\*____________________________________________________________________________*/
//$HeaderTop:$

#ifndef ALLOCATE_H_
#define ALLOCATE_H_

#include <stddef.h>
#include <iostream.h>

/*
** Define this flag to enable memory management
** tracing
*/
/* #define USE_ALLOCATOR_DEBUG */

/*
** Never used for productions builds
*/
#if !defined(NDEBUG) && defined(USE_ALLOCATOR_DEBUG)
#	undef USE_ALLOCATOR_DEBUG
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
#if defined(USE_ALLOCATOR_DEBUG)
struct AllocInfo
{
	unsigned long ulIndex;
	size_t tSize;
	bool bPool;
	bool bFreed;
};

typedef unsigned long TSignatureType;
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class Allocator
{
private:
	void *pMemPool;
	size_t tPoolSize;
	size_t tPoolUsed;
	void *InternalAllocate( size_t tS );
	void InternalDeAllocate( void *pV );
#if defined(USE_ALLOCATOR_DEBUG)
	size_t tCurrentAllocatedMemory;
	size_t tMaximumAllocatedMemory;
	ostream *pDbgStream;
	inline ostream *GetDbgStream()		{ return pDbgStream; };
	static TSignatureType culSignature;
	static size_t tASize;			// aligned size of an AllocInfo structure
	static size_t tSSize;			// aligned size of an signature element
	void *DbgAllocate( size_t tS );
	void DbgDeAllocate( void *pV );
	void AddToList( AllocInfo *pAlloc );
	void DeleteFromList( AllocInfo *pAlloc );
	void InternalWriteOutList( ostream *pOs );
	unsigned long ulIndex;

	/* --- for tracking allocations, frees, un-free'd mem. etc. --- */
	enum { NUM_ALLOC_INFO=16384 };
	unsigned long aAllocInfo[NUM_ALLOC_INFO];
	int iPtr;

protected:
	inline unsigned long GetCounterValue()	{ return ++ulIndex; };
#endif

public:
	Allocator();
	Allocator( ostream *pDbgStream );
	~Allocator();
	void StartTransaction( void *pMem, size_t tSize );
	size_t EndTransaction();
#if !defined(USE_ALLOCATOR_DEBUG)
	inline void *Allocate( size_t tS )	{ return InternalAllocate( tS ); };
	inline void DeAllocate( void *pV )	{ InternalDeAllocate( pV ); };
#	if !defined(NDEBUG)
		/*
		** Stub functions
		*/
		inline void EnableTracing( ostream *pOs=0 )				{ (void)pOs; };
		inline void DisableTracing()							{};
		inline void BeforeAllocate( int, const char * )			{};
		inline void BeforeDeAllocate( int, const char * )		{};
		inline void Mark( int, const char * )					{};
#	endif

#else
	void EnableTracing( ostream *pOs=0 );
	void DisableTracing();
	static size_t InternalAlign( size_t tSize );
	inline void *Allocate( size_t tS )	{ return DbgAllocate( tS ); };
	inline void DeAllocate( void *pV )	{ DbgDeAllocate( pV ); };
	void BeforeAllocate( int iLine, const char *pFileName );
	void BeforeDeAllocate( int iLine, const char *pFileName );
	void Mark( int iLine, const char *pFileName );
#endif
};

#endif	/* ALLOCATE_H_ */

