/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Base/DblList.hpp,v $
 * $Date: 2003/05/13 18:41:55 $
 *
 Description:
	Definition of inline functions for doubly linked list class.
\*____________________________________________________________________________*/

#ifndef DBLLIST_HPP_
#define DBLLIST_HPP_

#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline DblListIterator::DblListIterator( DblList &tList )
:	pSentinel( &tList.Sentinel() ),
	p( tList.Begin() )
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline DblListIterator &DblListIterator::operator=( DblList &tList )
{
	pSentinel = &tList.Sentinel();
	p = tList.Begin();
	return *this;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline ConstDblListIterator::ConstDblListIterator( const DblList &tList )
:	pSentinel( &tList.Sentinel() ),
	p( tList.Begin() )
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline ConstDblListIterator &ConstDblListIterator::operator=(
		const DblList &tList )
{
	pSentinel = &tList.Sentinel();
	p = tList.Begin();
	return *this;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline void DblList::PushBack( _Node::type l )
{
	Sentinel().pNext=PI_NEW( _Node(l, &Sentinel(), Begin() ) );
	iSize++;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline void DblList::Append( _Node::type l )
{
	Sentinel().pLast=PI_NEW( _Node(l, End(), &Sentinel() ) );
	iSize++;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline _Node::type DblList::Cap()
{
	_Node *p_Node = Begin();
	_Node::type lV = p_Node->lObj;
	p_Node->Delete();
	iSize--;
	return lV;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline _Node::type DblList::operator[]( const int i )
{
	DblListIterator iter(*this);
	return iter[i];
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline const _Node::type DblList::operator[]( const int i ) const
{
	ConstDblListIterator iter(*this);
	return iter[i];
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline _Node::type DblList::Detach( const int i )
{
	DblListIterator iter(*this);
	_Node *pNode = iter.ObjectAtIndex( i );
	_Node::type lV = pNode->lObj;
	pNode->Delete();
	iSize--;
	return lV;
}

#endif /* DBLLIST_HPP_ */

