/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Base/OSEntity.h,v $
 * $Date: 2003/05/13 18:41:55 $
 *
 Description:
	This is a collection of classes denoting OS entities whose
	implementation are specific to the operating system.
	Thus these collection of abstract base classes provide
	a unified interface to complex operating system objects.
\*____________________________________________________________________________*/
//$HeaderTop:$

#ifndef OSENTITY_H_
#define OSENTITY_H_

#include "PICompat.h"

/*____________________________________________________________________________*\
 *
 Declarations:
\*____________________________________________________________________________*/
#define PISYNC_EBUSY	(-10)	/* if this is changed then also change in
									../PiAPI
								*/
/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Semaphore, base class for all synchronization objects, including
	semaphores, mutexes, condition variables and recursive mutexes.
\*____________________________________________________________________________*/
class Semaphore
{
public:
	virtual ~Semaphore()				{};

	/* ---
	These methods return 0 on success. Error code on failure.
	--- */
	virtual int Lock()=0;
	virtual int TryLock()=0;
	virtual int UnLock()=0;

	virtual bool IsOK() const=0;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Thread object.
\*____________________________________________________________________________*/
typedef unsigned long (* ThreadFn)(unsigned long);
class Thread
{
private:
protected:
public:
	virtual bool Begin( ThreadFn fnEntry, unsigned long ulData,
		int iPriority, int iFlags )=0;
	virtual bool Terminate( unsigned long ulExitCode )=0;
	virtual ~Thread()	{};
	virtual bool Suspend()=0;
	virtual bool Resume()=0;
	virtual int GetPriority()=0;
	virtual bool SetPriority( int iPriority )=0;
	virtual bool SetData( int iKey, void *pData )=0;
	virtual bool GetData( int iKey, void **ppData )=0;
	virtual void *GetSystemHandle()=0;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Encapsulates a dynamically linked library.
\*____________________________________________________________________________*/
class DLL
{
public:
	virtual ~DLL() {};
	virtual void *GetAddress( const char *pSymbolName )=0;
	virtual const char *GetLastErrorDescription() const=0;
	virtual bool IsOK() const=0;
    virtual void *GetSystemHandle() const=0;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class MappedFile
{
public:
	virtual ~MappedFile()	{};
	virtual void *Lock( int *piLength )=0;
	virtual bool UnLock()=0;
	virtual bool IsOK()=0;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Encapsulates a microsecond timer.
\*____________________________________________________________________________*/
class Timer
{
public:
	virtual bool Reset()=0;
	virtual bool Elapsed( long *plSeconds, long *plMicroSeconds )=0;
	virtual bool IsOK()=0;
};

#if 0
/* Under development */
/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Encapsulates a shared memory segment, with synchronization mutex.
\*____________________________________________________________________________*/
class ShMem
{
public:
	virtual bool 
	virtual bool Reset()=0;
	virtual bool Elapsed( long *plSeconds, long *plMicroSeconds )=0;
	virtual bool IsOK()=0;
};
#endif

#endif /* OSENTITY_H_ */

