/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/DB/PIDBTree.h,v $
 * $Date: 2003/05/13 18:41:56 $
 *
 Description:
\*____________________________________________________________________________*/
/* $HeaderTop:$ */

#ifndef PIDBTREE_H_
#define PIDBTREE_H_

#include "_PIDB.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PIDBTree : public _PIDB
{
private:
	DblList *pTypes;
	int iLastType;

	/* --- hash tables --- */
	enum { HASH_TABLE_SIZE=16 };
	inline int GetNumLists( int iTheLastType )
		{
		return (iTheLastType+1)*HASH_TABLE_SIZE;
		};
	static inline int GetFirstListIndexForType( int iType )
		{
		return iType*HASH_TABLE_SIZE;
		}; 
	static inline DblList &GetList( DblList *pLists, int iType,
		const char *pKey )
		{
		return pLists[
			/* --- 
			It is possible to use the modulo of the key address here
			because the key is a symbol table entry and is gauranteed to
			always be the same address for the same string 
			--- */
	/* NOTE is this any faster?, using simple bitmask operators instead
		of modulo? */
			( ( 0xF0 & ((short)pKey)) >> 8) + /* assumes HASH_TABLE_SIZE==16 */
/*			( pKey ? (((unsigned long)pKey)%HASH_TABLE_SIZE) : 0 ) + */
	/* -- also, if the module is used, why not skip the (?:) operator? --- */
			/* ---
			This is what it would look like otherwise
			Platform::HashFunction( pKey?pKey:"", HASH_TABLE_SIZE ) +
			--- */
			GetFirstListIndexForType( iType ) ];
		};
	inline const TypeFrame &GetTypeFrame( int iType )
		{
		return (iType<PIDBTYPE_BUILTINSIZE) ?
			aTypeFrames[iType] :
			aAdditionalTypeFrames[iType-PIDBTYPE_BUILTINSIZE];
		};

	/* --- implement internal functions --- */
	static void DestructTree( void *pV );
	static void RenderTree( ostream &os, const void *pV, int iLevel );
	static void *AddString( void *pV );
	static void RenderString( ostream &os, const void *pV, int iLevel );
	static void DestructString( void *pV );
	static void RenderOpaque( ostream &os, const void *pV, int iLevel );
	static void RenderUser( ostream &os, const void *pV, int iLevel );
	static void DestructUser( void *pV );

	/* --- typeframes for the built-in types --- */
public:
	enum { NUM_TYPES=PIDBTYPE_BUILTINSIZE };

private:
	static TypeFrame aTypeFrames[ NUM_TYPES ];
	TypeFrame *aAdditionalTypeFrames;

	/* --- internal functions --- */
	void InternalDestruct( DblList *paLists,
		int iNumLists, void (* fnDestructor)(void *) );

public:
	PIDBTree( _PIDB *pParent=0, const char *pName=0,
			int iNumAdditionalTypes=0,
			TypeFrame *aTheAdditionalTypeFrames=0 );
	~PIDBTree();

	static const char *GetFastKey( const char *pKey, int iType );
	static int PIDBTypeNameToNumber( const char *pType );
	static int PIDBFlagsNameToNumber( const char *pFlag );
	virtual int Add( int iType, const char *pKey, void *pValue, int iFlags );
	virtual int Replace( int iType, const char *pKey, void *pValue,
		int iFlags );
	virtual _PIDBIterator *GetIterator( int iType, const char *pKey,
		int iFlags );
	virtual void *Lookup( int iType, const char *pKey, int iFlags );
	virtual void Render( ostream &os ) const { RenderTree( os, this, 0 ); };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class UserDBType
{
private:
	TypeFrame &tFrame;
	void *pValue;

public:
	UserDBType( TypeFrame &tTheFrame, void *pTheValue )
	:	tFrame( tTheFrame ), pValue( pTheValue )	{};
	~UserDBType()
		{ if ( tFrame.fnDelete )		{ (tFrame.fnDelete)(pValue); }; };	
	inline TypeFrame &GetTypeFrame()	{ return tFrame; };
	inline void *GetValue()				{ return pValue; };
};

#endif	/* PIDBTREE_H_ */

