/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/EnhPi3/IO.cpp,v $
 * $Date: 2003/06/01 13:47:42 $
 *
 Description:

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <assert.h>

#include "DblList.h"
#include "PIString.h"

#include "resrc1.h"
#include "Dialogs.h"
#include "Config.h"


/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
static HWND hWndList = 0;
static HWND hWndType = 0;
static HWND hWndHost = 0;
static HWND hWndPort = 0;
static HWND hWndSend = 0;
static HWND hWndRecv = 0;
static HWND hWndAdd = 0;
static HWND hWndReplace = 0;
static HWND hWndDelete = 0;
static HWND hWndKeys = 0;
static DblList lTypes;
static DblList lHosts;
static DblList lPorts;
static DblList lSends;
static DblList lRecvs;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_verifyFields( HWND hDlg, char *pszBuf, int iBufSize )
{
	assert( pszBuf && iBufSize );
	*pszBuf = '\0';

	/*
	** Scan Hostname for spaces
	*/
	SendMessage( hWndHost, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	int iLen = strlen(pszBuf);
	for( int iTest=0; iTest<iLen; iTest++)
		{
		if ( isspace( pszBuf[iTest] ) )
			{
			/*
			** Give message and abort change
			*/
			MessageBox( hDlg, "'Hostname' field may not contain spaces or tabs.",
				"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
			SetFocus( hWndHost );
			SendMessage( hWndHost, EM_SETSEL, 0, -1 );
			SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
			goto verify_fail;
			};
		};

	/*
	** Check Port
	*/
	SendMessage( hWndPort, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	iTest = atoi( pszBuf );
	if ( iTest<0 || iTest>65535 )
		{
		/*
		** Give message and abort change
		*/
		MessageBox( hDlg, "Port must be in the range [0..65536]",
			"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
		SetFocus( hWndPort );
		SendMessage( hWndPort, EM_SETSEL, 0, -1 );
		SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
		goto verify_fail;
		};

	/*
	** Send Timeout
	*/
	::SendMessage( hWndSend, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	iLen = strlen(pszBuf);
	for( iTest=0; iTest<iLen; iTest++)
		{
		if ( isalpha( pszBuf[iTest] ) )
			{
			/*
			** Give message and abort change
			*/
			MessageBox( hDlg, "Field contains alphanumeric characters.",
				"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
			SetFocus( hWndSend );
			SendMessage( hWndSend, EM_SETSEL, 0, -1 );
			SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
			goto verify_fail;
			};
		};

	iTest = atoi( pszBuf );
	if ( iTest<1 && iTest!=-1 )
		{
		/*
		** Give message and abort change
		*/
		MessageBox( hDlg, "Send timeout must be in the range [-1,1..n]",
			"Configuration Error' Field", MB_OK );
		SetFocus( hWndSend );
		SendMessage( hWndSend, EM_SETSEL, 0, -1 );
		SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
		goto verify_fail;
		};

	/*
	** Receive Timeout
	*/
	::SendMessage( hWndRecv, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	iLen = strlen(pszBuf);
	for( iTest=0; iTest<iLen; iTest++)
		{
		if ( isalpha( pszBuf[iTest] ) )
			{
			/*
			** Give message and abort change
			*/
			MessageBox( hDlg, "Field contains alphanumeric characters.",
				"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
			SetFocus( hWndRecv );
			SendMessage( hWndRecv, EM_SETSEL, 0, -1 );
			SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
			goto verify_fail;
			};
		};

	iTest = atoi( pszBuf );
	if ( iTest<1 && iTest!=-1 )
		{
		/*
		** Give message and abort change
		*/
		MessageBox( hDlg, "Receive timeout must be in the range [-1,1..n]",
			"Configuration Error' Field", MB_OK );
		SetFocus( hWndRecv );
		SendMessage( hWndSend, EM_SETSEL, 0, -1 );
		SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
		goto verify_fail;
		};

	/*
	** Good
	*/
	return 1;

verify_fail:

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Insert values from listbox into values array. Sort
	the lines in reverse order of virtual path component length.
\*____________________________________________________________________________*/
static void Internal_saveValues( HWND hDlg )
{
	enum { BUF_SIZE=1023 };
	char pszBuf[BUF_SIZE+1];

	for(;;)
		{
		int iCount = SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

		if ( !iCount )
			{ return; };

		int iBiggest = -1;
		int iBiggestIndex = -1;

		for( int j=0; j<iCount; j++ )	
			{
			/*
			** Get string
			*/
			if ( SendMessage( hWndList, LB_GETTEXTLEN, j, 0 ) >= BUF_SIZE )
				{
				MessageBox( hDlg, "Line too long", "Internal Error", MB_OK );
				return;
				};
			SendMessage( hWndList, LB_GETTEXT, j, (LPARAM)pszBuf );

			
			/*
			** Get number of components in path
			*/
			int iNumComponents = 0;
			int i=0;
			for( ; pszBuf[i] && pszBuf[i]!='\t'; i++ );
			for( ; pszBuf[i] && isspace(pszBuf[i]); i++ );
			for( ; pszBuf[i] && !isspace(pszBuf[i]); i++ )
				{
				if ( pszBuf[i]=='/' )
					{
					iNumComponents++;
					i++; 
					if ( !pszBuf[i] || isspace(pszBuf[i]) )
						{ break; };
					};
				};

			if ( iNumComponents>iBiggest )
				{
				iBiggestIndex = j;
				iBiggest = iNumComponents;
				};
			};

		/*
		** Add biggest line to list and remove it from
		** listbox
		*/
		SendMessage( hWndList, LB_GETTEXT, iBiggestIndex, (LPARAM)pszBuf );
		SendMessage( hWndList, LB_DELETESTRING, iBiggestIndex, 0 );
		AD_addValue( __pConfig, "IOObjects", "IOObject", pszBuf );
		};
}
	
/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_addLine( const char *pLine )
{
	assert( pLine );
	assert( lTypes.Size()==lHosts.Size() );
	assert( lHosts.Size()==lPorts.Size() );
	assert( lPorts.Size()==lSends.Size() );
	assert( lSends.Size()==lRecvs.Size() );

	int i = 0;
	/* --- scan to tab --- */
	for( ; pLine[i] && pLine[i]!='\t'; i++ );
	lTypes.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );

	/* --- scan over whitespace to start of Host --- */
	for( ; pLine[i] && isspace(pLine[i]); i++ );
	pLine = &( pLine[i] );	
	i=0;
	/* --- scan to tab --- */
	for( ; pLine[i] && pLine[i]!='\t'; i++ );
	lHosts.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	

	/* --- scan over whitespace to start of Port --- */
	for( ; pLine[i] && isspace(pLine[i]); i++ );
	pLine = &( pLine[i] );	
	i=0;
	/* --- scan to tab --- */
	for( ; pLine[i] && pLine[i]!='\t'; i++ );
	lPorts.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	

	/* --- scan over whitespace to start of Send --- */
	for( ; pLine[i] && isspace(pLine[i]); i++ );
	pLine = &( pLine[i] );	
	i=0;
	/* --- scan to tab --- */
	for( ; pLine[i] && pLine[i]!='\t'; i++ );
	lSends.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	

	/* --- scan over whitespace to start of Recv --- */
	for( ; pLine[i] && isspace(pLine[i]); i++ );
	pLine = &( pLine[i] );	
	i=0;
	/* --- scan to tab --- */
	for( ; pLine[i] && pLine[i]!='\t'; i++ );
	lRecvs.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	

	assert( lTypes.Size()==lHosts.Size() );
	assert( lHosts.Size()==lPorts.Size() );
	assert( lPorts.Size()==lSends.Size() );
	assert( lSends.Size()==lRecvs.Size() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static void Internal_updateButtons( WPARAM wParam, LPARAM lParam, 
	const char *pHostname, const char *pPort, int iTypeChanged )
{
	int iCBSel = SendMessage( hWndType, CB_GETCURSEL, 0, 0 );
	const char *pType = 0;
	if ( iCBSel>-1 )
		{
		pType = AD_lookupValueEx( __pConfig, "Internal", 
			"IOTypes", iCBSel );
		}
	else
		{
		return;
		}

	int iExistingIndex = -1;
	assert( pType && pHostname && pPort );
	int iIndex = 0;
	if ( *pType )
		{
		if ( !stricmp( pType, "SSL" ) && iTypeChanged
		&& ( !stricmp( pPort, "" ) || !stricmp( pPort, "80" )))
			{
			SendMessage( hWndPort, WM_SETTEXT, 1, (LPARAM)"443" );
			return;
			}

		if ( !stricmp( pType, "TCP" ) && iTypeChanged
		&& ( !stricmp( pPort, "" ) || !stricmp( pPort, "443" )))
			{
			SendMessage( hWndPort, WM_SETTEXT, 1, (LPARAM)"80" );
			return;
			}

		/*
		** If the 'type', 'hostname' or 'port' changed, check is it an existing one.
		*/
		DblListIterator i( lTypes );
		DblListIterator j( lHosts );
		DblListIterator k( lPorts );
		for( ; !i.BadIndex(); i++, j++, k++, iIndex++ )
			{
			PIString *pT = (PIString *)i.Current();
			PIString *pH = (PIString *)j.Current();
			PIString *pP = (PIString *)k.Current();
			assert( pT && pH && pP);
			if ( (*pT)==pType && (*pH)==pHostname && (*pP)==pPort )
				{
				iExistingIndex = iIndex;
				break;
				};
			};
		};

	/*
	** Remove listbox selection or scroll to correct line
	** as appropriate 
	*/
	SendMessage( hWndList, LB_SETCURSEL, iExistingIndex, 0 );

	/*
	** Set buttons
	*/
	if ( iCBSel<0 || SendMessage( hWndType, CB_GETLBTEXTLEN, iCBSel, 0 )==0 ||
			SendMessage( hWndHost, EM_LINELENGTH, 0, 0 )==0 ||
			SendMessage( hWndPort, EM_LINELENGTH, 0, 0 )==0 )
		{
		if ( IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, FALSE );
		if ( IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, FALSE );
		}
	else
		{
		if ( iExistingIndex>-1 )	/* --- disable add, enable replace --- */
			{
			if ( IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, FALSE );
			if ( !IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, TRUE );
			}
		else						/* --- disable replace, enable add --- */
			{
			if ( !IsWindowEnabled( hWndAdd ) ) EnableWindow( hWndAdd, TRUE );
			if ( IsWindowEnabled( hWndReplace ) ) EnableWindow( hWndReplace, FALSE );
			};
		};

	/*
	** If there is a selected item in the listbox then
	** enable delete, otherwise disable delete	
	*/
	if ( SendMessage( hWndList, LB_GETCURSEL, 0, 0 )==LB_ERR )
		{
		EnableWindow( hWndDelete, FALSE );
		EnableWindow( hWndKeys, !stricmp( pType, "SSL" ) &&
			IsWindowEnabled( hWndAdd )
			);
		}
	else
		{
		EnableWindow( hWndDelete, TRUE );
		EnableWindow( hWndKeys, !stricmp( pType, "SSL" ));
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
{
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	const char *pTmp;

	switch (uMsg)
		{
		case WM_INITDIALOG:
			AD_hourglassOn();
			{
			int aTabstop[5];
			hWndList = GetDlgItem( hDlg, IDC_LIST );
			hWndType = GetDlgItem( hDlg, IDC_TYPE );
			hWndHost = GetDlgItem( hDlg, IDC_HOSTNAME );
			hWndPort = GetDlgItem( hDlg, IDC_PORT );
			hWndSend = GetDlgItem( hDlg, IDC_SEND );
			hWndRecv = GetDlgItem( hDlg, IDC_RECV );
			hWndAdd = GetDlgItem( hDlg, IDC_ADD );
			hWndReplace = GetDlgItem( hDlg, IDC_REPLACE );
			hWndDelete= GetDlgItem( hDlg, IDC_DELETE );
			hWndKeys= GetDlgItem( hDlg, IDC_NEW );

			assert( hWndList && hWndType && hWndHost && hWndPort );
			assert( hWndSend && hWndRecv );
			assert( hWndAdd && hWndReplace && hWndDelete );

			EnableWindow( hWndAdd, FALSE );
			EnableWindow( hWndReplace, FALSE );
			EnableWindow( hWndDelete, FALSE );
			EnableWindow( hWndKeys, FALSE );			
			/*
			** Add options to 'Types' combo-box
			*/
			int iTmp = 0;
			for(;; iTmp++ )
				{
				pTmp = AD_lookupValueEx( __pConfig, "Internal", 
					"IOTypes", iTmp );
				if ( !pTmp )
					{ break; };

				::SendMessage( hWndType, CB_ADDSTRING, 0, 
					(LPARAM)pTmp ); 
				};

			/*
			** Limit edit field lengths, really just a sanity thing
			*/
			SendMessage( hWndHost, EM_SETLIMITTEXT, 256, 0 );
			SendMessage( hWndPort, EM_SETLIMITTEXT, 256, 0 );
			SendMessage( hWndSend, EM_SETLIMITTEXT, 256, 0 );
			SendMessage( hWndRecv, EM_SETLIMITTEXT, 256, 0 );

			/*
			** Setup tabstops
			*/
			aTabstop[0] = 0;
			aTabstop[1] = 38;
			aTabstop[2] = 107;
			aTabstop[3] = 148;
			aTabstop[4] = 187;
			SendMessage( hWndList, LB_SETTABSTOPS, 5, (LPARAM)&aTabstop );

			/*
			** Load all mappings
			*/
			iTmp = 0;
			for(;; iTmp++ )
				{
				pTmp = AD_lookupValueEx( __pConfig, "IOObjects", 
					"IOObject", iTmp );
				if ( !pTmp )
					{ break; };

				if ( SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)pTmp )>=0 )
					{
					Internal_addLine( pTmp );
					}
				else
					{ break; /* error */; };
				};
			}

			AD_hourglassOff();
			break;

		case WM_DESTROY:
			{
			/*
			** Delete values
			*/
			DblListIterator x( lTypes );
			DblListIterator y( lHosts );
			DblListIterator z( lPorts );
			DblListIterator s( lSends );
			DblListIterator r( lRecvs );
			for( ; !x.BadIndex(); x++, y++, z++, s++, r++ )
				{
				assert( !x.BadIndex() && !y.BadIndex() && !z.BadIndex() 
					&& !s.BadIndex() && !r.BadIndex() );
				PI_DELETE( (PIString *)x.Current() );
				PI_DELETE( (PIString *)y.Current() );
				PI_DELETE( (PIString *)z.Current() );
				PI_DELETE( (PIString *)s.Current() );
				PI_DELETE( (PIString *)r.Current() );
				};
			lTypes.Clear();
			lHosts.Clear();
			lPorts.Clear();
			lSends.Clear();
			lRecvs.Clear();
			};
			break;

		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
				{
				case IDC_ADD:
					{
					/*
					** Verify fields are valid
					*/
					if ( !Internal_verifyFields( hDlg, szBuf, BUF_SIZE ) )
						{ break; };

					int i, j;
					*szBuf = '\0';

					SendMessage( hWndType, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					for( i=0; i<BUF_SIZE && szBuf[i]!='\t' && szBuf[i]; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lTypes.Append( (DblList::type)PI_NEW( PIString( szBuf ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lHosts.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lPorts.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndSend, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lSends.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndRecv, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lRecvs.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );					

					SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );

					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					SendMessage( hWndHost, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndSend, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndRecv, WM_SETTEXT, 0, (LPARAM)"" );
					EnableWindow( hWndKeys, FALSE );					
					SetFocus( hWndList );
					AD_Changed( hDlg );
					};
					break;
	
				case IDC_REPLACE:
					{
					/*
					** Verify fields are valid
					*/
					if ( !Internal_verifyFields( hDlg, szBuf, BUF_SIZE ) )
						{ break; };

					/*
					** Get selected item index 
					*/
					int iCurSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
					assert( iCurSel>=0 );

					/*
					** Get text to add and add it to internal lists only
					*/
					int i, j;
					*szBuf = '\0';

					SendMessage( hWndType, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
					for( i=0; i<BUF_SIZE && szBuf[i]!='\t' && szBuf[i]; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lTypes.Append( (DblList::type)PI_NEW( PIString( szBuf ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lHosts.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lPorts.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndSend, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lSends.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					szBuf[i++] = '\t';

					SendMessage( hWndRecv, WM_GETTEXT, BUF_SIZE-i, (LPARAM)&(szBuf[i]) );
					for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t'; i++ );
					if ( szBuf[i] ) { szBuf[i] = '\0'; };
					lRecvs.Append( (DblList::type)PI_NEW( PIString( &(szBuf[j]) ) ) );
					
					/*
					** Delete the item from the list and internal lists
					*/
					SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
					PI_DELETE( (PIString *)lTypes.Detach( iCurSel) );
					PI_DELETE( (PIString *)lHosts.Detach( iCurSel) );
					PI_DELETE( (PIString *)lPorts.Detach( iCurSel) );
					PI_DELETE( (PIString *)lSends.Detach( iCurSel) );
					PI_DELETE( (PIString *)lRecvs.Detach( iCurSel) );
					
					/*
					** Add new item to listbox
					*/
					SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );

					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					SendMessage( hWndHost, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndSend, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndRecv, WM_SETTEXT, 0, (LPARAM)"" );
					SetFocus( hWndList );
					AD_Changed( hDlg );
					};
					break;

				case IDC_DELETE:
					{
					int iCurSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
					SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
					PI_DELETE( (PIString *)lTypes.Detach( iCurSel ) );
					PI_DELETE( (PIString *)lHosts.Detach( iCurSel ) );
					PI_DELETE( (PIString *)lPorts.Detach( iCurSel ) );
					PI_DELETE( (PIString *)lSends.Detach( iCurSel ) );
					PI_DELETE( (PIString *)lRecvs.Detach( iCurSel ) );
			
					/*
					** Clear edit fields, focus to list
					*/
					SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					SendMessage( hWndHost, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndSend, WM_SETTEXT, 0, (LPARAM)"" );
					SendMessage( hWndRecv, WM_SETTEXT, 0, (LPARAM)"" );
					SetFocus( hWndList );
					AD_Changed( hDlg );
					};
					break;

				case IDC_TYPE:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)!=CBN_SELCHANGE )
						{ break; };
					*szBuf = '\0';
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf ); 
					pTmp = szBuf + strlen(szBuf) + 1;
					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)pTmp ); 
					Internal_updateButtons( wParam, lParam, szBuf, pTmp, 1 );
					break;

				case IDC_HOSTNAME:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
						{ break; };
					*szBuf = '\0';
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf ); 
					pTmp = szBuf + strlen(szBuf) + 1;
					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)pTmp ); 
					Internal_updateButtons( wParam, lParam, szBuf, pTmp, 0 );
					break;

				case IDC_PORT:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
						{ break; };
					*szBuf = '\0';
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE,(LPARAM)szBuf );
					pTmp = szBuf + strlen(szBuf) + 1;
					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)pTmp ); 
					Internal_updateButtons( wParam, lParam, szBuf, pTmp, 0 );
					break;

				case IDC_SEND:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
						{ break; };
					break;

				case IDC_RECV:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
						{ break; };
					break;

				case IDC_LIST:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)==LBN_SELCHANGE )
						{
						int iSel = SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
						if ( iSel!=LB_ERR )
							{
							assert( iSel>=0 && iSel<lTypes.Size() );
							EnableWindow( hWndDelete, TRUE );
							int iIndex = SendMessage( hWndType, CB_FINDSTRINGEXACT, 
								(WPARAM)-1, 
								(LPARAM)(const char *)*((PIString *)lTypes[iSel]) );
							assert( iIndex>=0 );
							SendMessage( hWndType, CB_SETCURSEL, (WPARAM)
								(iIndex==CB_ERR ? -1 : iIndex), 0 );
							SendMessage( hWndHost, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lHosts[iSel]) );
							SendMessage( hWndPort, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lPorts[iSel]) );
							SendMessage( hWndSend, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lSends[iSel]) );
							SendMessage( hWndRecv, WM_SETTEXT, 0,
								(LPARAM)(const char *)*((PIString *)lRecvs[iSel]) );
							};
						};
					break;

				case IDC_NEW:
					{
					DlgData data;
					pTmp = szBuf;

					// copy ssl servername
					data.pSslServername = pTmp;
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE,(LPARAM)pTmp );
					pTmp = data.pSslServername + strlen(data.pSslServername);
					pTmp++;

					// copy ssl private key filename
					data.pSslKeyFilename = pTmp;
					strcpy( (char *)pTmp, "./" );
					pTmp = data.pSslKeyFilename + strlen(data.pSslKeyFilename);
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE,(LPARAM)pTmp );
					pTmp = data.pSslKeyFilename + strlen(data.pSslKeyFilename);
					strcpy( (char *)pTmp, "_" );
					pTmp++;
					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)pTmp );
					pTmp = data.pSslKeyFilename + strlen(data.pSslKeyFilename);
					strcpy( (char *)pTmp, "_key.pem" );
					pTmp = data.pSslKeyFilename + strlen(data.pSslKeyFilename) + 1;

					// copy ssl certificate filename
					data.pSslCertFilename = pTmp;
					strcpy( (char *)pTmp, "./" );
					pTmp = data.pSslCertFilename + strlen(data.pSslCertFilename);
					SendMessage( hWndHost, WM_GETTEXT, BUF_SIZE,(LPARAM)pTmp );
					pTmp = data.pSslCertFilename + strlen(data.pSslCertFilename);
					strcpy( (char *)pTmp++, "_" );
					SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)pTmp );
					pTmp = data.pSslCertFilename + strlen(data.pSslCertFilename);
					strcpy( (char *)pTmp, "_cert.pem" );

					DialogBoxParam(
						::GetModuleHandle( NULL ),
						MAKEINTRESOURCE( IDD_GENKEYS ),
						hDlg,
						(DLGPROC)fnGenSslKeys,
						(LPARAM)&data
						);
					break;
					}

				default:;		
				}
			break;

		case WM_NOTIFY:
			{
			switch( ((NMHDR *)lParam)->code )
				{
				case PSN_SETACTIVE:
					{
					int i = *AD_lookupValue( __pConfig, "General", "ReuseThreads" ) ? 1 : 0;
					EnableWindow( hWndList, i ); EnableWindow( hWndType, i );
					EnableWindow( hWndHost, i ); EnableWindow( hWndPort, i );
					EnableWindow( hWndSend, i ); EnableWindow( hWndRecv, i );
					break;
					}

				case PSN_QUERYCANCEL:
					return AD_cancel( hDlg );

				case PSN_KILLACTIVE:
					{
					int i = 0;
					int iCurSel = ::SendMessage( hWndList, LB_GETCURSEL, 0, 0 );
					int iCount = ::SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

					/* --- remove old ones first --- */
					AD_deleteValue( __pConfig, "IOObjects", "IOObject" );

					/*
					** Save all mappings
					*/
					Internal_saveValues( hDlg );

					/*
					** Delete values
					*/
					DblListIterator x( lTypes );
					DblListIterator y( lHosts );
					DblListIterator z( lPorts );
					DblListIterator s( lSends );
					DblListIterator r( lRecvs );
					for( ; !x.BadIndex(); x++, y++, z++, s++, r++ )
						{
						assert( !x.BadIndex() && !y.BadIndex() && !z.BadIndex()
							&& !s.BadIndex() && !r.BadIndex() );
						PI_DELETE( (PIString *)x.Current() );
						PI_DELETE( (PIString *)y.Current() );
						PI_DELETE( (PIString *)z.Current() );
						PI_DELETE( (PIString *)s.Current() );
						PI_DELETE( (PIString *)r.Current() );
						};
					lTypes.Clear();
					lHosts.Clear();
					lPorts.Clear();
					lSends.Clear();
					lRecvs.Clear();

					/*
					** Load all mappings
					*/
					int iTmp = 0;
					const char *pTmp;
					for(;; iTmp++ )
						{
						pTmp = AD_lookupValueEx( __pConfig, "IOObjects", 
							"IOObject", iTmp );
						if ( !pTmp )
							{ break; };

						if ( SendMessage( hWndList, LB_ADDSTRING, 0,
							(LPARAM)pTmp )>=0 )
							{
							Internal_addLine( pTmp );
							}
						else
							{ break; /* error */; };
						};
					::SendMessage( hWndList, LB_SETCURSEL, iCurSel, 0 );
					::SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
					break;
					}

				default:
					return (FALSE);
				};
			};
			break;

		default:
			return (FALSE);
		};

	return (TRUE);
}

/*____________________________________________________________________________*\
 *
 Description:
	Definitions and global values
\*____________________________________________________________________________*/
DLGPROC fnIO = (DLGPROC)DialogProc;
