/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Intrface/Admin.h,v $
 * $Date: 2003/05/13 18:42:05 $
 *
 Description:

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#ifndef ADMIN_H_
#define ADMIN_H_

#include "QConfig.h"
#include "ConnPool.h"
#include "PIObject.h"
#include "HTTPDefs.h"

#define KEY_HTTP_X_ADMINCOMMAND     "X-AdminCommand"
#define KEY_ADMIN_GETCONFIG         "GetConfig"
#define KEY_ADMIN_SAVECONFIG        "SaveConfig"
#define ADMIN_CONTENTTYPE           "x-application/x-pi3-configuration-info"

#define KEY_BASIC_AUTH		"Basic"
#define KEY_DGST_AUTH		"Digest"
#define KEY_NTLM_AUTH		"NTLM"
#define KEY_REALM			"realm"
#define KEY_NONCE			"nonce"
#define ID_NO_AUTH			0
#define ID_BASIC_AUTH		1
#define ID_DGST_AUTH		2
#define ID_NTLM_AUTH		3

static const char *pMethodMap[] = 
{
    0,
    MD_NAME_GET,
    MD_NAME_POST,
    MD_NAME_HEAD,
    MD_NAME_PUT,
    MD_NAME_DELETE,
    MD_NAME_OPTIONS,
    MD_NAME_TRACE,
    MD_NAME_CONNECT,
    0
};

typedef struct _Admin Admin;
typedef int cbAuthData(void *pUserData);

Admin *Admin_newLocal( const char *pFileName );
Admin *Admin_newRemote(
    PIObject *pParent,
    const char *pHost,
    int iPort,
    const char *pUsername,
    const char *pPassword,
	cbAuthData *pGetAuthData,
	void *pUserData
	);
void Admin_delete( Admin *pAdmin );
int Admin_loadConfig( Admin *pAdmin, QConfig *pConfig, PIString &sError );
int Admin_saveConfig( Admin *pAdmin, QConfig *pConfig, PIString &sError );
const char *Admin_GetRealm(Admin *pAdmin);
const char *Admin_GetUsername(Admin *pAdmin);
void Admin_SetUsername(Admin *pAdmin, const char *pUsername);
void Admin_SetPassword(Admin *pAdmin, const char *pPassword);


#endif /* ADMIN_H_ */

