/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Intrface/Perform.cpp,v $
 * $Date: 2003/05/13 18:42:06 $
 *
 Description:

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <windowsx.h>

#include "IFaceBse.h"
#include "PIStrStr.h"
#include "StrToken.h"

#include "QConfig.h"
#include "Perform.hrc"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define C_RESOURCENAME                    "PerformancePage"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
    PerformancePage

Description:
    Creates the eponymous property page and is used by the PropSheet object.

Example:
	<PRE>
	&lt;Object&gt;
		Name PerformancePage
		Class PerformanceClass
		PropertyPage PerformancePage
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam);

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PerformancePage : public IFaceBase
{
private:
    QConfig *pConfig;
    PROPSHEETPAGE tPsp;
    Interface *pInterface;

    /*
    ** window private values
    */
	HWND hWndThreads;
	HWND hWndRecv;
	HWND hWndSend;
	HWND hWndKeepOpen;
	HWND hWndAliveCount;
	HWND hWndAliveTimeout;
	HWND hWndReverse;
	HWND hWndDebug;
	HWND hWndReuse;
	HWND hWndProcs;
	HWND hWndErrorLog;
	HWND hWndExitAfter;

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );

		PIOStrStream os;

		os << pWhere << "PerformancePage: ";

        if ( 0 )
            {
            }
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	PerformancePage( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	IFaceBase( pObject ),
        pConfig( FindParentQConfig( PIObject_getDB( pObject ) ) ),
        pInterface( 0 ),
		hWndThreads( 0 ),
		hWndRecv( 0 ),
		hWndSend( 0 ),
		hWndKeepOpen( 0 ),
		hWndAliveCount( 0 ),
		hWndAliveTimeout( 0 ),
		hWndReverse( 0 ),
		hWndDebug( 0 ),
		hWndReuse( 0 ),
		hWndProcs( 0 ),
		hWndErrorLog( 0 ),
		hWndExitAfter( 0 )
		{
		ReadParameters( iArgc, ppArgv );

        if ( !IsOK() )
            { return; };

        /*
        ** Setup the PROPSHEETPAGE
        */
        memset( &tPsp, 0, sizeof( PROPSHEETPAGE ) );
        tPsp.dwSize = sizeof( PROPSHEETPAGE );
        tPsp.dwFlags = PSP_DEFAULT;
        tPsp.hInstance = (HINSTANCE)PIDLL_getSystemHandle( PIClass_getLibrary(
            PIObject_getClass( Object() ) ) );

        /* Status page */
        tPsp.pszTemplate = C_RESOURCENAME;
        tPsp.pfnDlgProc = (DLGPROC)fnDialogProc;
        tPsp.lParam = (LPARAM)this;

        /*
        ** Set the data DB of this object
        */
        PIDB *pDB = PIObject_getDB( Object() );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_WINDOWTYPE,
            (void *)INTRFACE_PROPPAGE, 0 );
        assert(!PIDB_lookup( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE, 0));
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE,
            (void *)&tPsp, 0 );
        
        /*
        ** Done
        */
		};

	~PerformancePage()
		{
		};

    virtual int Handle( Interface &tWindow )
        {
        switch( tWindow.iMessage )
            {
            case INTRFACE_MSG_INIT:
                pInterface = &tWindow;
                break;

            case INTRFACE_MSG_WIN32CREATE:
				return PIAPI_ABORT;

            case INTRFACE_MSG_LOAD:
                return PIAPI_ABORT;

            case INTRFACE_MSG_DEFAULTS:
				M_AddValue( pConfig, "General", "Pocesses", "8" );
				M_AddValue( pConfig, "General", "Threads", "20" );
				M_AddValue( pConfig, "General", "ExitAfter", "1000" );
				M_AddValue( pConfig, "General", "SendTimeout", "350" );
				M_AddValue( pConfig, "General", "RecvTimeout", "400" );
				M_AddValue( pConfig, "General", "DNSReverseLookup", "" );
				M_AddValue( pConfig, "General", "Debug", "" );
				M_AddValue( pConfig, "General", "ErrorLog", "On" );
				M_AddValue( pConfig, "General", "ReuseThreads", "" );
				M_AddValue( pConfig, "General", "KeepOpen", "On" );
				M_AddValue( pConfig, "General", "KeepOpenCount", "100" );
				M_AddValue( pConfig, "General", "KeepOpenTimeout", "5" );
                break;

            default:
                assert( 0 );
                return PIAPI_ABORT;
            };

        return PIAPI_COMPLETED;
        };

	BOOL DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
	    {
	    switch (uMsg)
	        {
            case WM_INITDIALOG:
				{
				hWndThreads = GetDlgItem( hDlg, IDC_THREADS );
				hWndRecv = GetDlgItem( hDlg, IDC_RECV );
				hWndSend = GetDlgItem( hDlg, IDC_SEND );
				hWndKeepOpen = GetDlgItem( hDlg, IDC_KEEPALIVE );
				hWndAliveCount = GetDlgItem( hDlg, IDC_ALIVECOUNT );
				hWndAliveTimeout = GetDlgItem( hDlg, IDC_ALIVETIMEOUT );
				hWndReverse = GetDlgItem( hDlg, IDC_REVERSE );
				hWndDebug = GetDlgItem( hDlg, IDC_DEBUG );
				hWndReuse = GetDlgItem( hDlg, IDC_REUSE );
				hWndProcs = GetDlgItem( hDlg, IDC_PROCS );
				hWndErrorLog = GetDlgItem( hDlg, IDC_ERRORLOG );
				hWndExitAfter = GetDlgItem( hDlg, IDC_EXITAFTER );

				assert( hWndThreads );
				assert( hWndRecv && hWndSend );
				assert( hWndKeepOpen && hWndReverse );
				assert( hWndAliveCount && hWndAliveTimeout );
				assert( hWndDebug && hWndProcs && hWndReuse );
				assert( hWndErrorLog && hWndExitAfter );

				/* --- just for sanity --- */
				::SendMessage( hWndThreads, EM_SETLIMITTEXT, 256, 0 );
				::SendMessage( hWndRecv, EM_SETLIMITTEXT, 256, 0 );
				::SendMessage( hWndSend, EM_SETLIMITTEXT, 256, 0 );
				::SendMessage( hWndProcs, EM_SETLIMITTEXT, 256, 0 );
				::SendMessage( hWndExitAfter, EM_SETLIMITTEXT, 256, 0 );

				::SendMessage( hWndThreads, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "Threads" ));
				::SendMessage( hWndRecv, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "RecvTimeout" ));
				::SendMessage( hWndSend, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "SendTimeout" ));
				::SendMessage( hWndProcs, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "Processes" ));
				::SendMessage( hWndExitAfter, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "ExitAfter" ));
		
				if ( !stricmp( M_LookupValue( pConfig, "General", "KeepOpen" ), "On" ))
					{ ::SendMessage( hWndKeepOpen, BM_SETCHECK, BST_CHECKED, 0 ); }
				else
					{
					::SendMessage( hWndKeepOpen, BM_SETCHECK, BST_UNCHECKED, 0 );
					::EnableWindow( hWndAliveCount, FALSE );
					::EnableWindow( hWndAliveTimeout, FALSE );
					};
				::SendMessage( hWndAliveCount, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "KeepOpenCount" ));
				::SendMessage( hWndAliveTimeout, WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "KeepOpenTimeout" ));

				if ( *M_LookupValue( pConfig, "General", "DNSReverseLookup" ))
					{ ::SendMessage( hWndReverse, BM_SETCHECK, BST_CHECKED, 0 ); }
				else
					{ ::SendMessage( hWndReverse, BM_SETCHECK, BST_UNCHECKED, 0 ); };

				if ( *M_LookupValue( pConfig, "General", "Debug" ))
					{ ::SendMessage( hWndDebug, BM_SETCHECK, BST_CHECKED, 0 ); }
				else
					{ ::SendMessage( hWndDebug, BM_SETCHECK, BST_UNCHECKED, 0 ); };

				if ( *M_LookupValue( pConfig, "General", "ErrorLog" ))
					{ ::SendMessage( hWndErrorLog, BM_SETCHECK, BST_CHECKED, 0 ); }
				else
					{ ::SendMessage( hWndErrorLog, BM_SETCHECK, BST_UNCHECKED, 0 ); };

				if ( *M_LookupValue( pConfig, "General", "ReuseThreads" ))
					{ ::SendMessage( hWndReuse, BM_SETCHECK, BST_CHECKED, 0 ); }
				else
					{ ::SendMessage( hWndReuse, BM_SETCHECK, BST_UNCHECKED, 0 ); };

				Unchanged( hDlg );
				}
				break;

		case WM_COMMAND:
			if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE 
				|| GET_WM_COMMAND_CMD(wParam, lParam)==BN_CLICKED )
				{
				switch( GET_WM_COMMAND_ID( wParam, lParam) )
					{
					case IDC_KEEPALIVE:
						if ( ::SendMessage( hWndKeepOpen, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							::EnableWindow( hWndAliveCount, TRUE );
							::EnableWindow( hWndAliveTimeout, TRUE );
							}
						else
							{
							::EnableWindow( hWndAliveCount, FALSE );
							::EnableWindow( hWndAliveTimeout, FALSE );
							};
						break;

					default:;
					}
				Changed( hDlg );
				};
			break;

			case WM_NOTIFY:
				switch( ((NMHDR *)lParam)->code )
					{
					case PSN_QUERYCANCEL:
						return SHEET->Cancel(pInterface, hDlg);
	
					case PSN_KILLACTIVE:
						if ( IsChanged() )
						{
						enum { BUF_SIZE=1023 };
						char szBuf[BUF_SIZE+1];

						/*
						** Threads
						*/
						*szBuf = 0;
						SendMessage( hWndThreads, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						int iTest = atoi( szBuf );
						if ( iTest<1 || iTest>512 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Threads must be in the range [1..512]",
								"Configuration Error", MB_OK );
							SetFocus( hWndThreads );
							SendMessage( hWndThreads, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "Threads", szBuf ); 

						/*
						** Receive Timeout
						*/
						*szBuf = 0;
						::SendMessage( hWndRecv, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						int iLen = strlen(szBuf);
						for( iTest=0; iTest<iLen; iTest++)
							{
							if ( isalpha( szBuf[iTest] ) )
								{
								/*
								** Give message and abort change
								*/
								MessageBox( hDlg, "Field contains alphanumeric characters.",
									"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
								SetFocus( hWndRecv );
								SendMessage( hWndRecv, EM_SETSEL, 0, -1 );
								SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
								return (TRUE);
								};
							};

						iTest = atoi( szBuf );
						if ( iTest<1 && iTest!=-1 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Receive timeout must be in the range [-1,1..n]",
								"Configuration Error", MB_OK );
							SetFocus( hWndRecv );
							SendMessage( hWndRecv, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "RecvTimeout", szBuf ); 

						/*
						** Send Timeout
						*/
						*szBuf = 0;
						::SendMessage( hWndSend, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						iLen = strlen(szBuf);
						for( iTest=0; iTest<iLen; iTest++)
							{
							if ( isalpha( szBuf[iTest] ) )
								{
								/*
								** Give message and abort change
								*/
								MessageBox( hDlg, "Field contains alphanumeric characters.",
									"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
								SetFocus( hWndSend );
								SendMessage( hWndSend, EM_SETSEL, 0, -1 );
								SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
								return (TRUE);
								};
							};
						
						iTest = atoi( szBuf );
						if ( iTest<1 && iTest!=-1 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Send timeout must be in the range [-1,1..n]",
								"Configuration Error", MB_OK );
							SetFocus( hWndSend );
							SendMessage( hWndSend, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "SendTimeout", szBuf ); 

						/*
						** Keep-Alive
						*/
						if ( ::SendMessage( hWndKeepOpen, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{ M_ReplaceValue( pConfig, "General", "KeepOpen", "On" ); }
						else
							{ M_ReplaceValue( pConfig, "General", "KeepOpen", "Off" ); };

						/*
						** Keep-Alive Count
						*/
						*szBuf = 0;
						::SendMessage( hWndAliveCount, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						iTest = atoi( szBuf );
						if ( iTest<1 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Keep-alive count must be in the range [1..n]",
								"Configuration Error", MB_OK );
							SetFocus( hWndAliveCount );
							SendMessage( hWndAliveCount, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "KeepOpenCount", szBuf );

						/*
						** Keep-Alive Timeout
						*/
						*szBuf = 0;
						::SendMessage( hWndAliveTimeout, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						iLen = strlen(szBuf);
						for( iTest=0; iTest<iLen; iTest++)
							{
							if ( isalpha( szBuf[iTest] ) )
								{
								/*
								** Give message and abort change
								*/
								MessageBox( hDlg, "Field contains alphanumeric characters.",
									"Configuration Error", MB_OK | MB_ICONEXCLAMATION );
								SetFocus( hWndAliveTimeout );
								SendMessage( hWndAliveTimeout, EM_SETSEL, 0, -1 );
								SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
								return (TRUE);
								};
							};
						
						iTest = atoi( szBuf );
						if ( iTest<1 && iTest!=-1 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Keep-alive timeout must be in the range [-1,1..n]",
								"Configuration Error", MB_OK );
							SetFocus( hWndAliveTimeout );
							SendMessage( hWndAliveTimeout, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "KeepOpenTimeout", szBuf );

						/*
						** DNS Reverse Lookup
						*/
						if ( ::SendMessage( hWndReverse, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{ M_ReplaceValue( pConfig, "General", "DNSReverseLookup", "On" ); }
						else
							{ M_ReplaceValue( pConfig, "General", "DNSReverseLookup", "" ); };

						/*
						** Debug Log
						*/
						if ( ::SendMessage( hWndDebug, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							M_ReplaceValue( pConfig, "General", "Debug", "On" );
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Debug log is extremely verbose. Don't even \
think about using it in a production server.", "Debug Log Warning", MB_OK | MB_ICONWARNING );
							}
						else
							{ M_ReplaceValue( pConfig, "General", "Debug", "" ); };

						/*
						** Error-Log
						*/
						if ( ::SendMessage( hWndErrorLog, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{ M_ReplaceValue( pConfig, "General", "ErrorLog", "On" ); }
						else
							{ M_ReplaceValue( pConfig, "General", "ErrorLog", "" ); };
	
						/*
						** Processes
						*/
						*szBuf = 0;
						SendMessage( hWndProcs, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						iTest = atoi( szBuf );
						if ( iTest<1 || iTest>512 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Processes must be in the range [1..512]",
								"Configuration Error", MB_OK );
							SetFocus( hWndProcs );
							SendMessage( hWndProcs, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "Processes", szBuf ); 
	
						/*
						** ExitAfter
						*/
						*szBuf = 0;
						SendMessage( hWndExitAfter, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						iTest = atoi( szBuf );
						if ( iTest<1 && iTest!=-1 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Exit-after must be in the range [-1,1..n]",
								"Configuration Error", MB_OK );
							SetFocus( hWndExitAfter );
							SendMessage( hWndExitAfter, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "ExitAfter", szBuf ); 

						/*
						** Reuse Threads
						*/
						if ( ::SendMessage( hWndReuse, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{ M_ReplaceValue( pConfig, "General", "ReuseThreads", "On" ); }
						else
							{ M_ReplaceValue( pConfig, "General", "ReuseThreads", "" ); };
						}
						SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
						return (TRUE);
	
					case PSN_APPLY:
						return SHEET->SaveConfig(pInterface, hDlg);

					default:
						return (FALSE);
					};
				break;

            default:
                return (FALSE);
            };

        return FALSE;
        };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
    if ( uMsg==WM_INITDIALOG )
        {
        ::SetWindowLong( hDlg, DWL_USER, ((LPPROPSHEETPAGE)lParam)->lParam );
        };

    PerformancePage *pPage = (PerformancePage *)::GetWindowLong( hDlg, DWL_USER );
    return pPage ? pPage->DialogProc( hDlg, uMsg, wParam, lParam ) : FALSE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PerformancePage_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return IFaceBase_constructor( pObj, PI_NEW( PerformancePage( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name PerformancePageClass
		Type LogicExtension
		Library IntrFace
		OnClassLoad IFaceBase_onClassLoad
		Constructor PerformancePage_constructor
		CopyConstructor IFaceBase_copyConstructor
		Destructor IFaceBase_destructor
		Execute IFaceBase_execute
	</Class>

	<Object>
		Name PerformancePage
		Class PerformancePageClass
	</Object>

/*___+++CNF_END+++___*/
#endif

