/*____________________________________________________________________________*\
 *

 Copyright (c) 1999 Holger Zimmermann, John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/PHP4/PHP4Handler.cpp,v $
 * $Date: 2003/05/13 18:42:08 $
 *
 Description:
	Execute an PHP4 script.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#if WIN32
#include <windows.h>
#else
#include <ctype.h>
#endif

#include "pi3web_sapi.h"

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "DeQuote.h"
#include "Base64.h"


/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_EXTRAHDRPREFIX		"ExtraHeadersPrefix"
#define KEY_CONF_BEHAVIOR			"PHPMode"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	PHP4

Description:
	Wrapper for the PHP4 embedded script interpreter.

</PRE>
	The PHP4 Pi3 component is implemented using PHP4 Core Web Server API
	(SAPI), thus -
<PRE>
   &quot;This product includes software written by the PHP Development Team.&quot;
</PRE>

Options:

<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>ExtraHeadersPrefix
<TD>-
<TD>"HTTP_"
<TD>Prefix for HTTP extra request headers
<TD>ExtraHeadersPrefix "HTTP_"

<TR>
<TD>PHPMode
<TD>-
<TD>Either "Standard", "Highlight",  "Indent" or "SyntaxCheck"
<TD>Switch PHP to intended or highlight mode
<TD>PHPMode "Highlight"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>ExtraHeadersPrefix</H5>
The value is added before any HTTP extra header. The request header "User-Agent ..."
i.e. will generate a server variable "HTTP_USER-AGENT".

<H5>PHPMode</H5>
This value is used as a flag to control the general behaviour of PHP. The flag
is optional.
<ul>
<li><i>Standard:</i> If the <i>PHPMode</i> option is omitted, the <i>Standard</i>
mode is assumed and the script will be parsed as usual.
<li><i>Highlight:</i> In this mode the syntax highlighted source code will be send
to the client. The highlight colors are adjusted in <i>php.ini</i>.
<li><i>Indent:</i> The script source code is printed plain but with correct intended
lines.
<li><i>SyntaxCheck:</i> The script is only parsed for syntax errors and the result is
written to the client.
</ul>

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED, PIAPI_CONTINUE, PIAPI_ERROR, or INT_REDIRECT according
	to the status returned by the interpreter.

Note:
	This handler requires php4ts.dll 4.04 or greater from the PHP4 Win32 binary
      distribution.

<H5>PHP4 extensions</H5>
	The following PHP4 extensions are included in the PHP4-module:
	calendar
	crypt
	dbase
	filepro
	gd
	imap
	ldap
	msql
	mssql
	mysql
	odbc
	standard
	zlib
<H5>Server variables</H5>
	The following server variables are supported. "ALL_HTTP" will hold all
	request headers except of the values stored in other server variables.
	<P>AUTH_TYPE<BR>
	CONTENT_LENGTH<BR>
	CONTENT_TYPE<BR>
	GATEWAY_INTERFACE<BR>
	PATH_INFO<BR>
	PATH_TRANSLATED<BR>
	QUERY_STRING<BR>
	REQUEST_METHOD<BR>
	REMOTE_ADDR<BR>
	REMOTE_HOST<BR>
	REMOTE_USER<BR>
	SCRIPT_NAME<BR>
	SERVER_NAME<BR>
	SERVER_PORT<BR>
	SERVER_PROTOCOL<BR>
	SERVER_SOFTWARE<BR>
	ALL_HTTP


Example:
	<PRE>
	&lt;Object&gt;
		Name PHP4
		Class PHP4Class
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle Condition="&cmp($c,text/x-php3)" PHP4
		...
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Prototypes for callback functions
\*____________________________________________________________________________*/

BOOL fnGetServerVariable( HCONN hConn,
					LPSTR	lpszVariableName,
					LPVOID	lpvBuffer,
					LPDWORD	lpdwSize );

BOOL fnWriteClient( HCONN ConnID,
					LPVOID	Buffer,
					LPDWORD	lpdwBytes,
					DWORD	dwReserved );

BOOL fnReadClient( HCONN ConnID,
					LPVOID	lpvBuffer,
					LPDWORD	lpdwSize );

BOOL fnSendHeaderFunction( HCONN hConn,
					LPDWORD	lpdwSize,
					LPDWORD	lpdwDataType );

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	structure for PHP4 parameter mapping
\*____________________________________________________________________________*/
struct {
	const char *pName;
	int iFlag;
} aFlagMap[] =
	{
	{ "Standard",		PHP_MODE_STANDARD },
	{ "Highlight",		PHP_MODE_HIGHLIGHT },
	{ "Indent",			PHP_MODE_INDENT },
	{ "SyntaxCheck",	PHP_MODE_LINT },

	/* --- leave this last always --- */
	{ 0, 0 }
	};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Class for PHP4 context
\*____________________________________________________________________________*/
class PHP4;
class PHP4Context : public CONTROL_BLOCK
{
public:
	PHP4 *pPHP4;
	PIHTTP *pPIHTTP;
	int iRc;
private:
	inline char *DBL( PIDB *pDB, int iType, const char *pKey )
		{
		return (char *)PIDB_lookup( pDB, iType, pKey, 0 );
		};
	inline int SetAuth( PIHTTP *pPIHTTP)
		{
		/* --- test if remote_user already exists --- */
		const char *pAuthenticate  = DBL( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
			KEY_INT_REMOTEUSER );
		/* --- if Pi3Web has already done basic auth stuff --- */
		if ( pAuthenticate ) return PIAPI_COMPLETED;
		/* --- Get browser authentication string --- */
		pAuthenticate = DBL( pPIHTTP->pRequestDB, PIDBTYPE_RFC822,
			KEY_HTTP_AUTHORIZATION );
		if ( pAuthenticate )
			{
			/* --- Check authentication string --- */
			if ( !strncmp( pAuthenticate, "Basic ", 6 ) )
				{
				PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
					KEY_INT_AUTHTYPE, (void *)"Basic", 0 );

				/* --- base64 decoding stuff --- */
				const char *pBase64 = &( pAuthenticate[6] );
				char auth[256];
				Base64Decode( pBase64, auth, sizeof( auth ));
				/* --- split of authentication token --- */
				char *tok = strtok( auth, ":" );
				if ( tok != NULL )
					{
					PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
						KEY_INT_REMOTEUSER, tok, 0 );
					tok = strtok( NULL, ":" );
					if ( tok != NULL )
						{
						PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
							KEY_INT_AUTHPASS, tok, 0 );
						};
					};
				};
			};
			return PIAPI_COMPLETED;
		};
public:
	PHP4Context( PHP4 *pThePHP4, PIHTTP *pThePIHTTP, int iBehavior )
	:	pPHP4( pThePHP4 ), pPIHTTP( pThePIHTTP ), iRc( PIAPI_COMPLETED )
		{
		/* --- locals --- */
		PIDB *pQ = pPIHTTP->pRequestDB;
		PIDB *pR = pPIHTTP->pResponseDB;

		/* --- initialize member data --- */
		cbSize = sizeof( CONTROL_BLOCK );
		ConnID = (HCONN)this;
		dwHttpStatusCode = 0;
		*lpszLogData = '\0';

		/* --- input variables --- */
		lpszMethod = DBL( pQ, PIDBTYPE_STRING, KEY_HTTP_METHOD );
		lpszQueryString = DBL( pQ, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING );
		lpszPathInfo = DBL( pR, PIDBTYPE_STRING, KEY_INT_PATHINFO );
		lpszPathTranslated = DBL( pR, PIDBTYPE_STRING, KEY_INT_PATHTRANSLATED );
		lpszUri = DBL( pQ, PIDBTYPE_STRING, KEY_HTTP_URI );
		lpszReq = DBL( pQ, PIDBTYPE_STRING, KEY_HTTP_CLF );
		lpszFileName = DBL( pR, PIDBTYPE_STRING, KEY_INT_PATH );
		lpszContentType = DBL( pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE );
		dwBehavior = iBehavior;

		/* --- get user/passwd for basic auth with php --- */
		SetAuth( pPIHTTP );
		lpszUser = DBL( pR, PIDBTYPE_STRING, KEY_INT_REMOTEUSER );
		lpszPassword = DBL( pR, PIDBTYPE_STRING, KEY_INT_AUTHPASS );

		/* --- check for stdin --- */
		cbAvailable = 0;
		lpbData = NULL;
		cbTotalBytes = 0;
		const char *pContentLength = DBL( pQ, PIDBTYPE_RFC822,
			KEY_HTTP_CONTENTLENGTH );
		cbTotalBytes = pContentLength && pContentLength[0] ? atol( pContentLength ) : 0;

		/* --- callback functions --- */
		GetServerVariable = fnGetServerVariable;
		WriteClient = fnWriteClient;
		ReadClient = fnReadClient;
		SendHeaderFunction = fnSendHeaderFunction;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PHP4 : public HandlerBasePHP4
{
private:
	/* ---
	Configuration data
	--- */
	PIDB *pTypes;					/* DB for variable types */
   	PIString sExtraHeadersPrefix;	/* Extra Header for environment values, */
									/* by default "HTTP_" */
	int iBehavior;					/* PHP4 mode */
protected:
	/* ---
	Lookup functions (for GetServerVariable)
	--- */
	typedef BOOL (* PMFN)( PHP4 *pThis, PIHTTP *, LPVOID, LPDWORD );

	static BOOL fnAllHttp( PHP4 *pThis, PIHTTP *pPIHTTP,  char *lpvBuffer, LPDWORD lpdwSize )													
		{																		
		PIDB *pQ = pPIHTTP->pRequestDB;											
		int i = 0;																
		/* --- loop over all RFC822 headers in request block --- */				
		PIDBIterator *pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );	
		if ( !pIter )															
			{
				*lpdwSize = 0;
				PIPlatform_setLastError(PIAPI_INTERNALERROR);
				return -1;
			};														
		for(; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter ) )
			{																
			const char *pKey;													
			const char *pValue = (const char *)PIDBIterator_current( pIter,	&pKey );
			assert( pKey && pValue );											

			if ( !pKey || !pValue )												
				{ /* sanity */ continue; };										

			PMFN fnEnv = (PMFN)PIDB_lookup( pThis->pTypes, PIDBTYPE_OPAQUE, pKey, 0 );
			if ( fnEnv ) { continue; };

			int iVarLen = strlen( pKey );										
			int iVarPos = pThis->sExtraHeadersPrefix.Len();								
			int iValSize = strlen( pValue );									

			if ( (iVarPos+iValSize+i+3) < *lpdwSize )							
				{																
				memcpy( &(lpvBuffer[i]), pThis->sExtraHeadersPrefix, iVarPos );
				i += iVarPos;
				for( int j=0; j<iVarLen; j++, i++ )							
					{																
					if ( isalnum( pKey[j] ) )										
						{ lpvBuffer[i]=toupper(pKey[j]); }						
					else															
						{ lpvBuffer[i]='_'; };									
					};
				lpvBuffer[i++] = ':';
				lpvBuffer[i++] = ' ';
				memcpy( &(lpvBuffer[i]), pValue, iValSize ); i+=iValSize;		
				lpvBuffer[i++] = '\n';											
				}																
			else																
				{																
				PIDBIterator_delete( pIter );									
				*lpdwSize = 0;
				PIPlatform_setLastError(PIAPI_EINVAL);
				return PIAPI_FALSE;													
				};																
			};
		PIDBIterator_delete( pIter );
		*lpdwSize = i;
		return PIAPI_TRUE;															
		};

# define FN_DB_STR( name, db, type, key ) \
	static BOOL (name)( PHP4 *, PIHTTP *pPIHTTP, LPVOID lpvBuffer,		\
			LPDWORD lpdwSize )											\
		{																\
		const char *pValue = (const char *)PIDB_lookup( pPIHTTP->db,	\
			type, key, 0 );												\
		if ( !pValue )													\
			{															\
			pValue = ""; 												\
			};															\
		int iToCopy = strlen( pValue ) + 1;								\
		if ( DWORD(iToCopy)>*lpdwSize || !lpvBuffer )					\
			{															\
			*lpdwSize = iToCopy;										\
			PIPlatform_setLastError( PIAPI_EINVAL );					\
			return PIAPI_FALSE;											\
			};															\
		*lpdwSize = iToCopy;											\
		strncpy( (char *)lpvBuffer, pValue, iToCopy );					\
		return PIAPI_TRUE;												\
		};			

# define FN_STRING( name, value ) \
	static BOOL (name)( PHP4 *, PIHTTP *pPIHTTP, LPVOID lpvBuffer,		\
			LPDWORD lpdwSize )											\
		{																\
		const char *pValue = (value);									\
		if ( !pValue )													\
			{															\
			pValue = ""; 												\
			};															\
		int iToCopy = strlen( pValue ) + 1;								\
		if ( (DWORD)iToCopy>*lpdwSize || !lpvBuffer )					\
			{															\
			*lpdwSize = iToCopy;										\
			PIPlatform_setLastError( PIAPI_EINVAL );					\
			return PIAPI_FALSE;											\
			};															\
		*lpdwSize = iToCopy;											\
		strncpy( (char *)lpvBuffer, pValue, iToCopy );					\
		return PIAPI_TRUE;												\
		};			

	FN_DB_STR( fnAuthType, pResponseDB, PIDBTYPE_STRING, KEY_INT_AUTHTYPE );
	FN_DB_STR( fnContentLength, pRequestDB, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH );
	FN_DB_STR( fnContentType, pRequestDB, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE );
	FN_STRING( fnGatewayInterface, "CGI/1.1" );
	FN_DB_STR( fnPathInfo, pResponseDB, PIDBTYPE_STRING, KEY_INT_PATHINFO );
	FN_DB_STR( fnPathTranslated, pResponseDB, PIDBTYPE_STRING, KEY_INT_PATHTRANSLATED );
	FN_DB_STR( fnQueryString, pRequestDB, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING );
	FN_DB_STR( fnRemoteAddr, pConnectionDB, PIDBTYPE_STRING, KEY_INT_REMOTEADDR );
	FN_DB_STR( fnRemoteHost, pConnectionDB, PIDBTYPE_STRING, KEY_INT_REMOTEHOST );
	FN_DB_STR( fnRemoteUser, pResponseDB, PIDBTYPE_STRING, KEY_INT_REMOTEUSER );
	FN_DB_STR( fnRequestMethod, pRequestDB, PIDBTYPE_STRING, KEY_HTTP_METHOD );
	FN_DB_STR( fnScriptName, pResponseDB, PIDBTYPE_STRING, KEY_INT_SCRIPTNAME );
	FN_DB_STR( fnServerProtocol, pRequestDB, PIDBTYPE_STRING, KEY_HTTP_PROTOCOL );
	FN_DB_STR( fnCookie, pRequestDB, PIDBTYPE_RFC822, KEY_HTTP_COOKIE );
	FN_STRING( fnServerName, HTTPUtil_getHostName(pPIHTTP) );
	FN_STRING( fnServerPort, HTTPUtil_getHostPort(pPIHTTP) );
	FN_STRING( fnServerSoftware, HTTPCore_getServerStamp() );
#undef FN_DB_STR
#undef FN_STRING

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "PHP4: ";
		if ( !PIUtil_stricmp( KEY_CONF_EXTRAHDRPREFIX, pVariable ) )
			{
				sExtraHeadersPrefix = (const char *)DeQuote( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_BEHAVIOR, pVariable ) )
			{
			/* ---
			cycle through the list of available flags
			comparing them with this one
			--- */
			int i=0;
			for( ; aFlagMap[i].pName; i++ )
				{
				if ( !PIUtil_stricmp( aFlagMap[i].pName,
					DeQuote( pValue )))	{ break; };
				};
			if ( !aFlagMap[i].pName )
				{
				/* --- flag not found --- */
				os << "Unknown PHP4 mode '" << pValue << "'." << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};
			iBehavior = aFlagMap[i].iFlag;
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		return 1;
		};

public:
	PHP4( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBasePHP4( pObject ),
		pTypes( PIDB_new( 0, "Types" )),
		iBehavior( PHP_MODE_STANDARD )
		{
		ReadParameters( iArgc, ppArgv );
		/* ---
		Insert DB types for 'GetServerVariable'
		--- */
#define AT( name, function ) \
		PIDB_add( pTypes, PIDBTYPE_OPAQUE, (name), (void *)function, 0 );

		AT( "ALL_HTTP", fnAllHttp );
		AT( "AUTH_TYPE", fnAuthType );
		AT( "CONTENT_LENGTH", fnContentLength );
		AT( "CONTENT_TYPE", fnContentType );
		AT( "GATEWAY_INTERFACE", fnGatewayInterface );
		AT( "PATH_INFO", fnPathInfo );
		AT( "PATH_TRANSLATED", fnPathTranslated );
		AT( "QUERY_STRING", fnQueryString );
		AT( "REMOTE_ADDR", fnRemoteAddr );
		AT( "REMOTE_HOST", fnRemoteHost );
		AT( "REMOTE_USER", fnRemoteUser );
		AT( "REQUEST_METHOD", fnRequestMethod );
		AT( "SCRIPT_NAME", fnScriptName );
		AT( "SERVER_NAME", fnServerName );
		AT( "SERVER_PORT", fnServerPort );
		AT( "SERVER_PROTOCOL", fnServerProtocol );
		AT( "SERVER_SOFTWARE", fnServerSoftware );
		AT( "HTTP_COOKIE", fnCookie );
#undef AT
		};

	virtual ~PHP4()
		{
		PIDB_delete( pTypes );
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		if ( iPhase != PH_HANDLE) return PIAPI_ERROR;

		/* ---
		Create new PHP4 context
		--- */
		PHP4Context tContext( this, &tPIHTTP, iBehavior );
		const char *pPath = (const char *)PIDB_lookup( tPIHTTP.pResponseDB,
			PIDBTYPE_STRING, KEY_INT_PATH, 0 );

		/* ---
		Read in stdin (if any)
		--- */
		if ( tContext.cbTotalBytes )
			{
			/* ---
			See how much we can read on the first pass and leave the	
			extension for call ReadClient() for the rest		
			--- */
			int iRead = 0;
			const char *pBuffer = PIIOBuffer_read( tPIHTTP.pBuffer, &iRead );
			if ( !pBuffer || !iRead )
				{
				/* ---
				This is an error
				--- */
				HTTPCore_logError( &tPIHTTP, "PHP4: Could not read stdin \
(%ld bytes) for PHP3 script '%s'.", (long)tContext.cbTotalBytes, pPath );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			
			/* --- 
			Allocate a buffer and put data in it, buffer size will be
			no more that internal buffer in PIIOBuffer object.
			(usually between 4 and 16 K).
			--- */
			tContext.lpbData = (LPBYTE)PIHTTP_allocMem( &tPIHTTP, iRead );
			memcpy( tContext.lpbData, pBuffer, iRead );
			tContext.cbAvailable = iRead;
			};

		/* ---
		Fire off PHP4 Wrapper
		--- */
		int iRet = PHP4_wrapper( &tContext );

		/* ---
		Check for error message
		--- */
		if ( tContext.lpszLogData && *(tContext.lpszLogData) )
			{
			HTTPCore_logError( &tPIHTTP, "PHP4: Script with path '%s' \
logged an error message: ________________________________________________________________________________\n%s\n________________________________________________________________________________", pPath, tContext.lpszLogData );
			};
		
		tPIHTTP.iStatus = tContext.dwHttpStatusCode;

		switch( iRet )
			{
			case PIAPI_COMPLETED:
				if (iBehavior == PHP_MODE_LINT )
					{
#define MSG "\
<BODY BGCOLOR=\"#FFFFFF\" BACKGROUND=\"/icons/Pi3Tile.gif\">\
<H2>PHP4 Syntax correct</H2>\
The syntax check for the called PHP4 resource succeeded.</BODY>"
						unsigned long ulLen;
						ulLen = sizeof(MSG)-1;
						char szBuf[16];
						sprintf( szBuf, "%lu", ulLen );					
						PIDB_replace( tPIHTTP.pResponseDB, PIDBTYPE_RFC822,
							KEY_HTTP_CONTENTLENGTH, szBuf, 0 );
						PIIOBuffer_write( tPIHTTP.pBuffer, MSG, ulLen, PIIOBUF_NONE );
						PIIOBuffer_flush( tPIHTTP.pBuffer );
#undef MSG
					}
				break;
			case PIAPI_ERROR:
				if (iBehavior == PHP_MODE_LINT)
					{
#define MSG "\
<BODY BGCOLOR=\"#FFFFFF\" BACKGROUND=\"/icons/Pi3Tile.gif\">\
<H2>PHP4 Syntax error</H2>\
The syntax check for the called PHP4 resource failed.</BODY>"
					unsigned long ulLen = sizeof(MSG)-1;
					char szBuf[16];
					sprintf( szBuf, "%lu", ulLen );					
					PIDB_replace( tPIHTTP.pResponseDB, PIDBTYPE_RFC822,
						KEY_HTTP_CONTENTLENGTH, szBuf, 0 );
					PIIOBuffer_write( tPIHTTP.pBuffer, MSG, ulLen, PIIOBUF_NONE );
					PIIOBuffer_flush( tPIHTTP.pBuffer );
#undef MSG
					}
				else
					{
					HTTPCore_logError( &tPIHTTP, "PHP4: Script wrapper with path %s \
returned with a general error.", pPath );
					};
				break;
			default:
				HTTPCore_logError( &tPIHTTP, "PHP4: Script wrapper with path %s \
returned an unknown code.", pPath );
			};

		return tContext.iRc;
		};

	BOOL GetServerVariable( PIHTTP *pPIHTTP, LPSTR lpszVariableName,
		LPVOID lpvBuffer, LPDWORD lpdwSize )
		{
		if ( !lpszVariableName || !lpdwSize )
			{
			PIPlatform_setLastError( PIAPI_EINVAL );
			return PIAPI_FALSE;
			};

		/* ---
		Convert variable to upper case
		--- */
		PIString sTmp( lpszVariableName );
		sTmp.ConvertToUpperCase();

		/* ---
		Lookup function corresponding to this environment variable
		--- */
		PMFN fnEnv = (PMFN)PIDB_lookup( pTypes, PIDBTYPE_OPAQUE, sTmp, 0 );

		/* ---
		No lookup function?
		--- */
		if ( !fnEnv )
			{
			PIPlatform_setLastError( PIAPI_EINVAL );
			return PIAPI_FALSE;
			};

		return (fnEnv)( this, pPIHTTP, lpvBuffer, lpdwSize );
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL fnGetServerVariable( HCONN hConn, LPSTR lpszVariableName,
	LPVOID lpvBuffer, LPDWORD lpdwSize )
{
	PHP4Context *pContext = (PHP4Context *)hConn;
	int iRet = pContext->pPHP4->GetServerVariable( pContext->pPIHTTP,
		lpszVariableName, lpvBuffer, lpdwSize );
	if ( HTTPCore_debugEnabled() )
		{
		HTTPCore_logDebug( DBG_MED, "PHP4: Return code<%d>, variable<%s>, \
value<%x>, size<%d>.",
			iRet, lpszVariableName ? lpszVariableName : "",
			(void *)lpvBuffer, lpdwSize ? *lpdwSize : 0 );
		};
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL fnWriteClient( HCONN ConnID, LPVOID Buffer, LPDWORD	lpdwBytes,
	DWORD dwReserved )
{
	if ( !ConnID || !lpdwBytes || !Buffer )
		{
		PIPlatform_setLastError( PIAPI_EINVAL );
		return PIAPI_FALSE;
		};

	PHP4Context *pContext = (PHP4Context *)ConnID;
	int iWritten = PIIOBuffer_write( pContext->pPIHTTP->pBuffer,
		(char *)Buffer, *lpdwBytes, PIIOBUF_NONE );
	if ( iWritten==-1 )
		{
		/*
		** Socket must have been closed
		*/
		PIPlatform_setLastError( PIAPI_EINVAL );
		return PIAPI_FALSE;
		};
	*lpdwBytes = iWritten;
	return PIAPI_TRUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL fnReadClient( HCONN ConnID, LPVOID lpvBuffer, LPDWORD lpdwSize )
{
	if ( !ConnID || !lpdwSize )
		{
		PIPlatform_setLastError( PIAPI_EINVAL );
		return PIAPI_FALSE;
		};

	PHP4Context *pContext = (PHP4Context *)ConnID;
	int iRet = PIIOBuffer_readToBuffer( pContext->pPIHTTP->pBuffer,
		(char *)lpvBuffer, *lpdwSize );
	if ( iRet>0 )
		{
		*lpdwSize = iRet;
		return PIAPI_TRUE;
		}
	else
		{
		return PIAPI_FALSE;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL fnSendHeaderFunction( HCONN ConnID, LPDWORD lpdwSize, LPDWORD lpdwDataType )
{
	if ( !ConnID )
		{
		PIPlatform_setLastError( PIAPI_EINVAL );
		return PIAPI_FALSE;
		};

	PHP4Context *pContext = (PHP4Context *)ConnID;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;

//	switch( pContext->dwHttpStatusCode )
//		{
//		case ST_PERMANENTREDIRECT:
//		case 302:
//			PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_RFC822,
//				KEY_HTTP_LOCATION, (void *)"302", 0 );
//			pContext->dwHttpStatusCode = ST_PERMANENTREDIRECT;
//			pContext->iRc = PIAPI_COMPLETED;
//			return PIAPI_TRUE;
//		default:
			/*
			** Assume Content-Length could not be set, disable keep open
			*/
			PIDB_replace( pPIHTTP->pConnectionDB, PIDBTYPE_OPAQUE,
				KEY_INT_KEEPOPEN, 0, 0 );
			pPIHTTP->iStatus = pContext->dwHttpStatusCode;
			pContext->iRc = PIAPI_COMPLETED;

			/*
			** Send server specific headers
			*/
			if ( HTTPCore_sendGeneralHeaders( pPIHTTP ) != PIAPI_COMPLETED )
				{
				return PIAPI_FALSE;
				}

			if ( lpdwSize && *lpdwSize && lpdwDataType )
				{
				if ( PIIOBuffer_write( pPIHTTP->pBuffer, (const char *)lpdwDataType,
					*lpdwSize, PIIOBUF_NONE ) != *lpdwSize ) return PIAPI_FALSE;
				};

			return PIAPI_TRUE;
//		};
//	return PIAPI_FALSE;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PHP4_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBasePHP4_constructor( pObj, PI_NEW( PHP4( pObj,
		iArgc, ppArgv ) ) );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PHP4_destructor( PIObject *pObj, int,
	const char *[] )
{
	delete (PHP4 *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}


#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name PHP4Class
		Type LogicExtension
		Library PHP4Wrapper
		OnClassLoad HandlerBasePHP4_onClassLoad
		Constructor PHP4_constructor
		CopyConstructor HandlerBasePHP4_copyConstructor
		Destructor PHP4_destructor
		Execute HandlerBasePHP4_execute
	</Class>

	<Object>
		Name PHP4
		Class PHP4Class
		ExtraHeadersPrefix "HTTP_"
#		PHPMode "Highlight"
	</Object>

/*___+++CNF_END+++___*/
#endif

