/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 $Source: /cvsroot/pi3web/Pi3Web_200/Source/Parse/HTMLPars.cpp,v $
 $Date: 2003/05/13 18:42:08 $

 Description:
 *----------------------------------------------------------------------------*/
//$SourceTop:$

#include <assert.h>
#include <ctype.h>

#include "HTMLPars.h"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool HTMLParser::PutCh(char c)
{
	if ( iLen<BUF_SIZE )
		szBuf[iLen++]=c;
	else
		{
		assert( 0 );
		};
	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void HTMLParser::PutBackCh( int iCh )
{
	iPutBackCh=iCh;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTMLParser::GetCh2()
{
	if ( ppPtr && ppPtr<pePtr )
		{
		char ch=*ppPtr;
		ppPtr++;
		return ch;
		}
	else
		{
		unsigned long ulSize;
		ppPtr=(char *)pReader->Read( &ulSize );
		if ( ppPtr )
			{
			assert( ulSize );
			pePtr=&( ppPtr[ulSize] );
			return GetCh2();
			}
		else
			return PI_EOF;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Track line and position in line.
\*____________________________________________________________________________*/
int HTMLParser::GetCh1()
{
	ulX++;
	return GetCh2();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Could cause a stack overflow if its not inline.
\*____________________________________________________________________________*/
int HTMLParser::GetCh()
{
	char ch;
	if ( iPutBackCh!=PI_EOF ) 
		{ 
		ch=iPutBackCh;
		iPutBackCh=PI_EOF; 
		}
	else
		ch=GetCh1();

	if ( ch=='\r' ) return GetCh();
	return PutCh( ch ) ? ch : PI_EOF;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
HTMLParser::HTMLParser( const char *pFileName )
:	sFileName( pFileName ),
	pReader( PI_NEW( FileReader ) ),
	ppPtr(0), pePtr(0), iLen(0), bOK( false ),
	iPutBackCh( PI_EOF )
{
	bOK=Init();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
HTMLParser::~HTMLParser()
{
	PI_DELETE( pReader );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool HTMLParser::Init()
{
	assert( pReader );
	if ( !pReader->Open( sFileName ) )
		return false;

	/* --- reset variables --- */
	NF();
	eInputState=normal;

	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTMLParser::Lex()
{
	if ( !bOK ) return PI_EOF;
	iLen=0;
	for(;;)
		{
		char ch;
		if ( iLen==BUF_SIZE ) return TEXT;
		ch=GetCh();
		if ( ch=='\n' )
			{
			NL();
			};
		switch( eInputState )
			{
			case normal:
				if ( ch=='<' )
					{ PutBackCh( ch ); eInputState=start_markup; return TEXT; }
				else if ( ch==PI_EOF )
					{ return PI_EOF; };
				break;

			case start_markup:
				if ( ch=='<' )
					{ eInputState=markup_c1; }
				else
					{ eInputState=normal; }
				break;

			case markup_c1:
				if ( ch=='!' )
					{ eInputState=markup_c2; }
				else
					{ eInputState=normal; }
				break;

			case markup_c2:
				if ( ch=='-' )
					{ eInputState=markup_c3; }
				else
					{ eInputState=normal; };
				break;

			case markup_c3:
				if ( ch=='-' )
					{ eInputState=markup_comment; }
				else
					{ eInputState=normal; };
				break;

			case markup_comment:
				if ( ch=='-' )
					{ eInputState=markup_e1; };
				break;

			case markup_e1:
				if ( ch=='-' )
					{ eInputState=markup_e2; }
				else
					{ eInputState=markup_comment; };
				break;

			case markup_e2:
				if ( ch=='>' )
					{  eInputState=normal; return MARKUP_COMMENT; }
				else
					{ eInputState=markup_comment; };
				break;

			default:
				assert(0/* programming error */);
			};
	};
}

