/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Parse/HTMLPars.h,v $
 * $Date: 2003/05/13 18:42:08 $
 *
 Description:
\*____________________________________________________________________________*/
/* $HeaderTop:$ */

#ifndef HTMLPARS_H_
#define HTMLPARS_H_

#include <assert.h>
#include <ctype.h>

#include "FileRead.h"
#include "PIString.h"

#define PI_EOF (-1)

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class HTMLParser
{
private:
	PIString sFileName;
	FileReader *pReader;
	unsigned long ulX, ulY;
	inline void NF()	{ ulY=1; ulX=0; };
	inline void NL()	{ ulY++; ulX=0; };
	enum {	
		normal, 
		start_markup, 	
		markup_c1,
		markup_c2,
		markup_c3,
		markup_e1,
		markup_e2,
		markup_comment,
		nl
		} eInputState;

	/* --- input block management --- */
	char *ppPtr;		/* current pointer */
	char *pePtr;		/* end pointer */

	/* --- internal buffer --- */
	enum { BUF_SIZE=4096 };
	char szBuf[BUF_SIZE];
	int iLen;
	inline
	bool PutCh(char c);
	inline
	int GetCh2();
	inline
	int GetCh1();
	inline
	int GetCh();
	inline 
	void PutBackCh( int iCh );
	bool Init();
	bool bOK;
	int iPutBackCh;

protected:
public:
	HTMLParser( const char *pFileName );
	virtual ~HTMLParser();
	inline bool IsOK() const	{ return bOK; };

	/* --- lex and results --- */
	enum { TEXT, MARKUP_COMMENT };
	int Lex();

	/* --- data on current position --- */
	inline const char *GetPtr()	{ return szBuf; };
	inline unsigned long GetX()	{ return ulX; };
	inline unsigned long GetY()	{ return ulY; };
	inline int GetLen()		{ assert(iLen); return iLen-1; };

};

#endif	/* HTMLPARS_H_ */

