/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Pi2API/_PIObj.h,v $
 * $Date: 2003/05/13 18:42:09 $
 *
 Description:
\*____________________________________________________________________________*/
/* $HeaderTop:$ */

#ifndef _PIOBJECT_H_
#define _PIOBJECT_H_

#include "_PIClass.h"
#include "Pi2API.h"
#include "PIConfDB.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class _PIObject
{
private:
	PIString sName;
	const _PIClass *pClass;
	PIDB *pDB;
	PIConfigDB *pConfig;
	void *pData;
	bool bConstructed;

public:
	_PIObject( const char *pName, const _PIClass *pTheClass,
		_PIDB *pTheDB, PIConfigDB *pTheConfig )
	:	sName( pName ),
		pClass( pTheClass ),
		pDB( pTheDB ),
		pConfig( pTheConfig ),
		pData( 0 ),
		bConstructed( false )
		{
		};
	_PIObject( const _PIObject &tObject )
	:	sName( tObject.sName ),
		pClass( tObject.pClass ),
		pDB( tObject.pDB ),
		pConfig( tObject.pConfig ),
		pData( tObject.pData ),
		bConstructed( false )
		{
		};
	virtual ~_PIObject()
		{
		assert( !bConstructed );
		};
	const PIString &GetName() const			{ return sName; };
	PIDB *GetDB() const						{ return pDB; };
	void SetDB( PIDB *pTheDB )				{ pDB = pTheDB; };
	PIConfigDB *GetConfigurationDB() const	{ return pConfig; };
	void SetConfigurationDB( PIConfig *pTheConfig )
											{ pConfig = pTheConfig; };
	const _PIClass *GetClass() const		{ return pClass; };
	void *GetUserData() const				{ return pData; };
	void SetUserData( void *pTheData )		{ pData=pTheData; };

	/* --- static methods --- */
	static _PIObject *Load(_PIDB *pDB, PIConfigDB *pConfig, const char *pName,
		int iArgc, const char *ppArgv[] );
	static _PIObject *LoadFromLine( _PIDB *pDB, PIConfigDB *pConfig,
		const char *pObjectLine );
	static _PIObject *LoadEx( const _PIObject *pObject, const char *pKey );
	static int ReadParameters( _PIObject *pObject, int iArgc,
		const char *ppArgv[],
		int (*)(void *, const char *, const char *, const char *),
		void *pParam );
	static _PIObject *New( const _PIClass *pClass, _PIDB *pDB,
		PIConfigDB *pConfig, const char *pName, int iArgc,
		const char *ppArgv[] );
	static _PIObject *Copy( const _PIObject *pObject, int iArgc,
		const char *ppArgv[] );
	static int Delete( _PIObject *pObject, int iArgc, const char *ppArgv[] );

	/* --- friends --- */
	friend class _PIClass;
};

#endif	/* _PIOBJECT_H_ */

