/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Pi3API/HTTPCore.cpp,v $
 * $Date: 2003/06/01 12:10:14 $
 *
 Description:
	HTTP core facilities.

	TODO:
		i) The status and headers request line parsing doesn't allow for
			line continuation with tabs.

		ii) Some of error response codes should be changed. There a little
			overuse of HTTP BAD REQUEST. Some HTTP/1.1 gateway error 
			response codes should be used instead.

\*____________________________________________________________________________*/
//$SourceTop:$

#define TEMP 1

#if !defined(NDEBUG)
#	include <iostream.h>
#endif
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <fstream.h>
#include <stdio.h>
#include <stdarg.h>

#include "PIString.h"
#include "HTTPDefs.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PICompat.h"
#include "MiscUtil.h"
#include "DeQuote.h"
#include "PIStrStr.h"
#include "StrToken.h"
#include "Pi3Expr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_HANDLERS				"Handlers"
#define KEY_CONF_SERVERROOT				"ServerRoot"
#define KEY_CONF_SERVERSTAMP			"ServerStamp"
#define KEY_CONF_CLIENTSTAMP			"ClientStamp"
#define KEY_CONF_ERRORLOGFILE			"ErrorLogFile"
#define KEY_CONF_DEBUGLOGFILE			"DebugLogFile"
#define KEY_CONF_DEBUGBEFOREHANDLER		"DebugBeforeHandler"
#define KEY_CONF_DEBUGAFTERHANDLER		"DebugAfterHandler"
#define KEY_CONF_DEFAULTMIMETYPE		"DefaultMIMEType"

/*
#define D { cerr << __FILE__ << ", " << __LINE__ << ": " << endl; }
*/
#define D
enum { LOCAL_BUFFER_LEN=4095 };	/* --- used for read RFC822, URI etc --- */

/* --- maximum number of internal redirects --- */
#define MAX_REDIRECTS					10

#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }

/*____________________________________________________________________________*\
 *
 Description:
    Global fast keys.
\*____________________________________________________________________________*/
static bool bFastKeysInitialized = false;
static const char *pFKCLF = 0;
static const char *pFKMethod = 0;
static const char *pFKURI = 0;
static const char *pFKPath = 0;
static const char *pFKQueryString = 0;
static const char *pFKProtocol = 0;
static const char *pFKMethodNumber = 0;
static const char *pFKProtocolNumber = 0;
static const char *pFKProxyProtocol = 0;
static const char *pFKProxyHost = 0;
static const char *pFKProxyPort = 0;
static const char *pFKKeepOpenO = 0;
static const char *pFKKeepOpenH = 0;
static const char *pFKBytesSent = 0;
static const char *pFKHostRoot = 0;
static const char *pFKStatusCode = 0;
static const char *pFKTransferEnc = 0;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Internal_initFastKeys()
{
    if ( bFastKeysInitialized )
        {
        return;
        };

	pFKCLF = PIDB_getFastKey( KEY_HTTP_CLF, PIDBTYPE_STRING );
	pFKMethod = PIDB_getFastKey( KEY_HTTP_METHOD, PIDBTYPE_STRING );
	pFKURI = PIDB_getFastKey( KEY_HTTP_URI, PIDBTYPE_STRING );
	pFKPath = PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING );
	pFKQueryString = PIDB_getFastKey( KEY_HTTP_QUERYSTRING, PIDBTYPE_STRING );
	pFKProtocol = PIDB_getFastKey( KEY_HTTP_PROTOCOL, PIDBTYPE_STRING );
	pFKMethodNumber = PIDB_getFastKey( KEY_HTTP_METHOD, PIDBTYPE_OPAQUE );
	pFKProtocolNumber = PIDB_getFastKey( KEY_HTTP_PROTOCOL, PIDBTYPE_OPAQUE );
	pFKProxyProtocol = PIDB_getFastKey( KEY_HTTP_PROXYPROTOCOL,
		PIDBTYPE_STRING );
	pFKProxyHost = PIDB_getFastKey( KEY_HTTP_PROXYHOST, PIDBTYPE_STRING );
	pFKProxyPort = PIDB_getFastKey( KEY_HTTP_PROXYPORT, PIDBTYPE_STRING );
	pFKKeepOpenO = PIDB_getFastKey( KEY_INT_KEEPOPEN, PIDBTYPE_OPAQUE );
	pFKKeepOpenH = PIDB_getFastKey( KEY_INT_KEEPOPEN, PIDBTYPE_RFC822 );
	pFKBytesSent = PIDB_getFastKey( KEY_INT_BYTESSENT, PIDBTYPE_OPAQUE );
	pFKHostRoot = PIDB_getFastKey( KEY_INT_HOSTROOT, PIDBTYPE_STRING );
	pFKStatusCode = PIDB_getFastKey( KEY_HTTP_STATUSCODE, PIDBTYPE_OPAQUE );
	pFKTransferEnc = PIDB_getFastKey( KEY_HTTP_TRANSFERENCODING, PIDBTYPE_RFC822 );

    bFastKeysInitialized = true;

    assert( pFKMethod && pFKURI && pFKPath && pFKQueryString &&
		pFKProtocol &&
		pFKMethodNumber && pFKProtocolNumber && 
		pFKProxyProtocol && pFKProxyHost && pFKProxyPort &&
		pFKKeepOpenO && pFKKeepOpenH && pFKBytesSent &&
		pFKHostRoot && pFKStatusCode
		);
}

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ServerGlobals
{
private:
	/* --- forbid copy constructor --- */
	ServerGlobals( const ServerGlobals & )
		{ assert( 0 ); };

public:
	PIString sServerRoot;
	PIString sServerStamp;
	PIDB *pMIMEDB;
	PIString sDefaultMIMEType;
#if TEMP
	PIPLATFORM_FD tErrorLog;
#else
	FILE *pErrorLog;
#endif
	FILE *pDebugLog;
	Pi3Expression *pBeforeHandler;
	Pi3Expression *pAfterHandler;
	DblList lHandlers;

	ServerGlobals()
		:	pMIMEDB( 0 ),
#if TEMP
			tErrorLog( PIPLATFORM_FD_INVALID ),
#else
			pErrorLog( 0 ),
#endif
			pDebugLog( 0 ),
			pBeforeHandler( 0 ),
			pAfterHandler( 0 )
		{};

	~ServerGlobals()
		{
		for( DblListIterator i( lHandlers ); !i.BadIndex(); i++ )
			{ PIObject_delete( (PIObject *)i.Current(), 0, 0 ); };
#if TEMP
		if ( tErrorLog!=PIPLATFORM_FD_INVALID )
			{
			PIFile_close( tErrorLog );
			tErrorLog = PIPLATFORM_FD_INVALID;
			};
#else
		if ( pErrorLog )
			{
			::fclose( pErrorLog );
			pErrorLog = 0;
			};
#endif
		if ( pDebugLog )
			{
			::fclose( pDebugLog );
			pDebugLog = 0;
			};
		Pi3Expression_delete( pBeforeHandler );
		Pi3Expression_delete( pAfterHandler );
		PIDB_delete( pMIMEDB );
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ClientGlobals
{
private:
	/* --- forbid copy constructor --- */
	ClientGlobals( const ClientGlobals & )
		{ assert( 0 ); };

public:
	PIString sClientStamp;

    ClientGlobals()
		{};

	~ClientGlobals()
		{
		};
};

/*____________________________________________________________________________*\
 *
 Description:
	Globals
\*____________________________________________________________________________*/
static ServerGlobals *pSG = 0;
static ClientGlobals *pCG = 0;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	0 if not loaded, -1 on error, 1 if loaded.
\*____________________________________________________________________________*/
int Internal_LoadExpression( Pi3Expression **ppExpression, const char *pName,
		PIObject *pObject )
{
	const char *pValue = (const char *)PIConfig_lookupValue(
		PIObject_getConfigurationDB( pObject ), pName, 0, 0 );
	if ( !pValue )
		{ return 0; };
	DeQuote tValue( pValue );

	assert( ppExpression );
	PIString sError;
	Pi3Expression *pExpression = 0;
	pExpression = Pi3Expression_new( tValue, 0, &sError );
	if ( !pExpression )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ),
			PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR,
			"HTTPCore_init: Error parsing expression: %s",
				(const char *)sError );
		return -1;
		};
	*ppExpression = pExpression;
	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_addMIMEEntry( PIDB *pMIMEDB, const char *pLine, PIString &sMsg )
{
	assert( pLine );
	int i;
	/*
	** Scan past initial space 
	*/
	i = 0;
	for( ; pLine[i] && isspace( pLine[i] ); i++ );
	if ( !pLine[i] )
		{
		sMsg  = "HTTP1_0: Syntax error in MIME entry.";
		return 0;
		};
	pLine = &( pLine[i] );
	/*
	** Scan to first whitespace
	*/
	for( i=0; pLine[i] && !isspace( pLine[i] ); i++ );
	PIString sExtn( pLine, i );

#if WIN32
	/*
	** Silently convert extension to upper case
	*/
	sExtn.ConvertToUpperCase();
#endif
	const char *pExtn = sExtn;

	/*
	** Skip over all space
	*/
	for( ; pLine[i] && isspace( pLine[i] ); i++ );
	if ( !pLine[i] )
		{
		sMsg = "HTTP1_0: Syntax error in MIME entry.";
		return 0;
		};
	pLine = &( pLine[i] );
	for( i=0; pLine[i] && !isspace( pLine[i] ); i++ );

	/*
	** New extension is in sExtn, MIME type starts at pLine and has
	** length i
	*/
	PIString sMIMEType( pLine, i );
	const char *pMIMEType = sMIMEType;

	const char *pPreviousMIMEType = (const char *)PIDB_lookup(
		pMIMEDB, PIDBTYPE_STRING, pExtn, 0 );
	if ( pPreviousMIMEType )
		{
		sMsg = "HTTP1_0: Extension '";
		sMsg.Concatenate( pExtn );
		sMsg.Concatenate( "' is already mapped to MIME type '" );
		sMsg.Concatenate( pPreviousMIMEType );
		sMsg.Concatenate( "'." );
		return 0;
		};

	if (PIDB_add( pMIMEDB, PIDBTYPE_STRING, pExtn,
		(void *)pMIMEType, 0 ))
		{
		sMsg = "HTTP1_0: internal error adding MIME type.";
		return 0;
		};

	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_loadAllMIMETypes( PIDB *pMIMEDB, PIObject *pObject, PIDB *pConf )
{
	PIDBIterator *pIter = PIDB_getIterator( pConf, PIDBTYPE_USER,
		KEY_CONF_ADDMIMEENTRY, 0 );
	if ( !pIter ) { return 0; };
	for( ;	PIDBIterator_atValidElement( pIter ); 
			PIDBIterator_next( pIter ) )
		{
		const char *pVariable = 0;
		PIConfigString *pString = (PIConfigString *)PIUserType_getValue(
			(PIUserType *)PIDBIterator_current( pIter, &pVariable ) );
		const char *pValue = PIConfigString_getString( pString );

		/* --- generate location string --- */
		PIOStrStream os;
#if 0
/* NOTE: This error information is really terse */
		long lLine = pString->GetY();
			
		/* --- if line position is 0, this is the previous line --- */
		if ( !pString->GetX() && lLine>0 )
			{ lLine--; };

		os << "[" << lLine << "] ";
#endif

		/* --- no variable --- */
		if ( !pVariable || !*pVariable )
			{
			os << "HTTP1_0: Internal error adding MIME type '" << 
				pVariable << "'" << ends;
			CONFIG_ERR( pObject, os.str() );
			PIDBIterator_delete( pIter );
			return 0;
			};

		/* --- no value --- */
		if ( !pValue || !*pValue )
			{
			os << "HTTP1_0: No value for variable '" <<
				pVariable << "'" << ends;
			CONFIG_ERR( pObject, os.str() );
			PIDBIterator_delete( pIter );
			return 0;
			};

		/* --- ignore 'Name' or 'Class' parameters --- */
		if (	!PIUtil_stricmp( PIDBKEY_CONF_NAME, pVariable ) ||
				!PIUtil_stricmp( PIDBKEY_CONF_CLASS, pVariable ) )
			{ continue; };
			
		DeQuote tUnQuoted( pValue );

		PIString sMsg;
		if ( !Internal_addMIMEEntry( pMIMEDB, tUnQuoted.GetString(), sMsg ) )
			{
			os << sMsg << ends;
			CONFIG_ERR( pObject, os.str() );
			PIDBIterator_delete( pIter );
			return 0;
			};
		};
	PIDBIterator_delete( pIter );

	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_RelativeToAbsolutePath( const char *pRoot,
    const char *pRelativePath, PIString &sResult )
{
	assert( pRoot );
	assert( pRelativePath );

    if ( !pRoot )
        { pRoot = ""; };
	if ( RelativeToAbsolutePath( pRoot, pRelativePath, sResult ) )
		{
#if WIN32
        /*
        ** Convert '/' to '\\' if this is WIN32
        */
        char *pStr = (char *)(const char *)sResult; /* break encapsulation! */
        int iLen = strlen( pStr );
        for( int j=0; j<iLen; j++ )
            {
            if ( pStr[j]=='/' ) { pStr[j]='\\'; };
            };
#endif

        return 1;
        };

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return zero on success
\*____________________________________________________________________________*/
int HTTPCore_initServer( PIObject *pObject )
{
    if ( pSG )
        {
        /* already setup */
        return PIAPI_COMPLETED;
        };

    /*
    ** Make sure fast keys are setup
    */
    Internal_initFastKeys();

	pSG = PI_NEW( ServerGlobals() );

	assert( pObject );
	if ( !pObject )
        {
		CONFIG_ERR( pObject, "HTTPCore_initServer: NULL object \
(Internal Error)." );
		return -1;
        };

	/* ---
	Read the server root. Server root is mandatory
	--- */
	const char *pServerRoot = (const char *)PIConfig_lookupValue( 
		PIObject_getConfigurationDB( pObject ), KEY_CONF_SERVERROOT, 0, 0 );
	pSG->sServerRoot = pServerRoot;

	/* ---
	If the server root specified in the configuration file is relative
	then lookup the current directory and store it in absolute form,
	this avoids complications that may occur if the current directory
	is changed in the lifetime of the server
	--- */
	enum { DIR_BUF=1023 };
	char szBuf[DIR_BUF+1];
	*szBuf = '\0';
	PIPlatform_getCurrentDirectory( szBuf, DIR_BUF );
	szBuf[DIR_BUF] = '\0';
	if ( !RelativeToAbsolutePath( szBuf, pServerRoot,
		pSG->sServerRoot ) )
		{
		CONFIG_ERR( pObject, "HTTPCore_initServer: Could not set \
'ServerRoot' (Internal Error)." );
		return -1;
		};

	/* ---
	Contract the path to remove any '..' sequences 
	--- */
	strncpy( szBuf, pSG->sServerRoot, DIR_BUF );
	if ( HTTPUtil_contractPath( szBuf ) )
		{
		/* --- too many '..' sequences --- */
		CONFIG_ERR( pObject, "HTTPCore_initServer: 'ServerRoot' contains too \
many '..' sequences." );
		return -1;
		};
	pSG->sServerRoot = szBuf;

	if ( pSG->sServerRoot==PIString::Empty() )
		{
		CONFIG_ERR( pObject, "HTTPCore_initServer: No definition for \
'ServerRoot'" );
		return -1;
		};

	pSG->sServerStamp = (const char *)PIConfig_lookupValue( 
		PIObject_getConfigurationDB( pObject ), KEY_CONF_SERVERSTAMP, 0, 0 );

	/* ---
	open error log
	--- */
	const char *pErrorLog = (const char *)PIConfig_lookupValue(
		PIObject_getConfigurationDB( pObject ), KEY_CONF_ERRORLOGFILE, 0, 0 );
	if ( !pErrorLog )
		{
		CONFIG_ERR( pObject, "HTTPCore_initServer: No definition for \
'ErrorLogFile'" );
		return -1;
		};
	DeQuote tErrorLogFile( pErrorLog );
	PIString sErrorLogFile( pErrorLog );
	Internal_RelativeToAbsolutePath( pSG->sServerRoot,
        (const char *)tErrorLogFile, sErrorLogFile );
#if TEMP
	pSG->tErrorLog = PIFile_open( sErrorLogFile, "w" );
	if ( pSG->tErrorLog==PIPLATFORM_FD_INVALID )
#else
	pSG->pErrorLog = ::fopen( sErrorLogFile, "w" );
	if ( !pSG->pErrorLog )
#endif
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), 
			PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR,
			"HTTPCore_initServer: Could not open error logfile '%s'.",
			(const char *)sErrorLogFile );
		return -1;
		};
	/* ---
	Log initial message
	--- */
	HTTPCore_logError( 0, "Server error log started" );

	/* ---
	open debug log
	--- */
	const char *pDebugLog = (const char *)PIConfig_lookupValue(
		PIObject_getConfigurationDB( pObject ), KEY_CONF_DEBUGLOGFILE, 0, 0 );
	if ( pDebugLog )
		{
		DeQuote tDebugLogFile( pDebugLog );
		PIString sDebugLogFile( pDebugLog );
		Internal_RelativeToAbsolutePath( pSG->sServerRoot,
            (const char *)tDebugLogFile, sDebugLogFile );
		pSG->pDebugLog = ::fopen( sDebugLogFile, "w" );
		if ( !pSG->pDebugLog )
			{
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR,
				"HTTPCore_initServer: Could not open error logfile '%s'.",
					(const char *)sErrorLogFile );
			return -1;
			};
		HTTPCore_logDebug( DBG_HIGH, "Server debug log started" );
		};

	/* --- load the debug expressions --- */
	if ( Internal_LoadExpression( &(pSG->pBeforeHandler),
		KEY_CONF_DEBUGBEFOREHANDLER, pObject )==-1 )
		{ return -1; };
	if ( Internal_LoadExpression( &(pSG->pAfterHandler),
		KEY_CONF_DEBUGAFTERHANDLER, pObject )==-1 )
		{ return -1; };

	/* ---
	Create MIME DB and read in MIME file.
	--- */
	pSG->pMIMEDB = PIDB_new( 0, "MIME DB" );
	const char *pMIMEFile = (const char *)PIConfig_lookupValue( 
		PIObject_getConfigurationDB( pObject ), KEY_CONF_MIMEFILE, 0, 0);
	if ( pMIMEFile )
		{
		DeQuote tDeQuote( pMIMEFile );
		PIString sMIMEFile;
		Internal_RelativeToAbsolutePath( pSG->sServerRoot,
            (const char *)tDeQuote, sMIMEFile );
		if ( !HTTPCore_loadMIMEFile( pObject, pSG->pMIMEDB, sMIMEFile ) ) 
			{
			/* --- HTTPCore_loadMIMEFile() generates error messages --- */
			return -1;
			};
		};

	/* ---
	DefaultMIMEType
	--- */
	DeQuote tDefaultMIMEType( (const char *)PIConfig_lookupValue( 
		PIObject_getConfigurationDB( pObject ), KEY_CONF_DEFAULTMIMETYPE, 0,
		0 ) );
	pSG->sDefaultMIMEType = (const PIString &)tDefaultMIMEType;
	if ( pSG->sDefaultMIMEType==PIString::Empty() )
		{
		pSG->sDefaultMIMEType==HTTP_DEFAULT_MIME_TYPE;
		};

	/* ---
	Load other MIME types
	--- */
	if ( !Internal_loadAllMIMETypes( pSG->pMIMEDB, pObject,
		(PIDB *)PIObject_getConfigurationDB( pObject ) ) )
		{ return -1; /* --- error already given --- */ };

	PIDB *pDB = PIObject_getDB( pObject );

	/* ---
	Load top level handlers
	--- */
	const char *pHandlerNames = PIConfig_lookupValue(
		PIObject_getConfigurationDB( pObject ),
		KEY_CONF_HANDLERS, 0, 0 );
	if ( !pHandlerNames )
		{
		CONFIG_ERR( pObject, "HTTPCore_initServer: 'Handlers' not defined" );
		return -1;
		};
	/* --- Attempt to load all handler objects --- */
	StringTokenizer tTokens( pHandlerNames, " " );
	bool bIsOK = true;
	int i;
	for( i=0; bIsOK && i<tTokens.NumTokens(); i++ )
		{
		const char *pHandlerName = tTokens.GetToken( i );
		PIObject *pHandler = PIObject_load( pDB, 0, pHandlerName, 0, 0 );
		if ( !pHandler )
			{
			/* --- message is already set by 'PIObject_load()' --- */
			return -1;
			};
		pSG->lHandlers.Append( (DblList::type)pHandler );
		};

	if ( !bIsOK )	   
		{
		CONFIG_ERR( pObject, "HTTPCore_initServer: internal error adding \
handlers to DB" );
		return -1;
		};

	/* ---
	Sanity check
	--- */
	return !( pFKMethod && pFKURI && pFKPath && pFKQueryString &&
		pFKProtocol &&
		pFKMethodNumber && pFKProtocolNumber && 
		pFKProxyProtocol && pFKProxyHost && pFKProxyPort &&
		pFKKeepOpenO && pFKKeepOpenH
		);
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return zero on success
\*____________________________________________________________________________*/
/* int HTTPCore_initClient( PIObject *pObject ) - LATER see note below */
int HTTPCore_initClient()
{
    if ( pCG )
        {
        /* already setup */
        return PIAPI_COMPLETED;
        };

    /*
    ** Make sure fast keys are setup
    */
    Internal_initFastKeys();

	pCG = PI_NEW( ClientGlobals() );

#if 0
/* LATER - perhaps */
	assert( pObject );

	if ( !pObject )
        {
		CONFIG_ERR( pObject, "HTTPCore_initClient: NULL object \
(Internal Error)." );
		return -1;
        };

	pCG->sClientStamp = (const char *)PIConfig_lookupValue( 
		PIObject_getConfigurationDB( pObject ), KEY_CONF_CLIENTSTAMP, 0, 0 );
#endif

    return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	zero on success
\*____________________________________________________________________________*/
int HTTPCore_cleanupServer()
{
#if !defined(NDEBUG)
	PIProgram_dbgDump( pSG->pMIMEDB, "./mime.dbg" );
#endif
	PI_DELETE( pSG );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	zero on success
\*____________________________________________________________________________*/
int HTTPCore_cleanupClient()
{
	PI_DELETE( pCG );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	return 0 on success
\*____________________________________________________________________________*/
static int Internal_SendHeader_Date( PIIOBuffer &tB )
{
	PIString sTmp;
	time_t tT;
	PIPlatform_beforeUnsafeBlock();
	time( &tT );
	struct tm *pTms = gmtime( &tT );

	int iRet = HTTPUtil_rFC822Time( pTms, &sTmp );
	PIPlatform_afterUnsafeBlock();

	if ( iRet )
		{ return ST_INTERNALERROR; };

#define TMP "Date: "
	tB.Write( TMP, sizeof( TMP )-1 );
#undef TMP
	tB.WriteLn( sTmp, sTmp.Len() );

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	return 0 on success
\*____________________________________________________________________________*/
static int Internal_SendHeader_Connection( PIHTTP &tPIHTTP )
{
	PIDB *pC = tPIHTTP.GetConnectionDB();
	PIDB *pQ = tPIHTTP.GetRequestDB();
	PIIOBuffer &tB = tPIHTTP.GetIOBuffer();
	
	/* ---
	Here's the connection related response headers we want to 
	send:-

		for HTTP/0.9, 	there are no headers and no keep open

		for HTTP/1.0, 
				Connection: Keep-Alive	
			to notify connection will be kept open

		for HTTP/1.1
				Connection: Close
			if connection is not going to be kept open

	--- */

	/* --- get protocol --- */
	int iProtocol = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, pFKProtocolNumber,
		PIDBFLAG_FASTKEY );

	/* --- get keep open flag --- */
	int iKeepOpen = (int)PIDB_lookup( pC, PIDBTYPE_OPAQUE, pFKKeepOpenO,
		PIDBFLAG_FASTKEY );
	
	if ( iProtocol == PR_HTTP10 && iKeepOpen )
		{
		/* --- notify that the connection will be kept open --- */
#define TMP "Connection: Keep-Alive"
		tB.WriteLn( TMP, sizeof( TMP )-1  );
#undef TMP
		}
	else if ( iProtocol == PR_HTTP11 && !iKeepOpen )
		{
		/* --- notify that the connection will be closed --- */
#define TMP "Connection: Close"
		tB.WriteLn( TMP, sizeof( TMP )-1  );
#undef TMP
		};

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Verify the HTTP response line and divide it into components. Returns
	INT_CONTINUE on success or an HTTP error status code on error.

	Assumes pLine is NULL terminated at pLine[iLineLen].

	On success instantiates all values to at least an empty string.

\*____________________________________________________________________________*/
static int Internal_SplitResponseLine(
	char *pLine, 
	int iLineLen,
	const char **ppProtocol,
	int *piStatusCode
	)
{
	assert( pLine );
	assert( iLineLen );
	assert( ppProtocol );
	assert( piStatusCode );

	*ppProtocol = "";	/* --- this is done because protcol is optional, i.e.
						HTTP/0.9 doesn't send one
						--- */
	/* ---
	protocol
	--- */
	int i = 0;
	for( ; i<iLineLen && !(isspace(pLine[i])); i++);
	*ppProtocol = pLine;
	if ( i==iLineLen ) { return ST_BADREQUEST; };
	pLine[i]='\0'; i++;
	for(; i<iLineLen && (isspace(pLine[i])); i++);
	pLine = &( pLine[i] ); iLineLen -= i; i=0; 

	/* ---
	pLine now points at the statuscode following the protocol
	--- */

	/* --- Statuscode must exist --- */
	if ( i==iLineLen ) { return ST_BADREQUEST; };

	for(; i<iLineLen && !(isspace(pLine[i])); i++);
	const char *pStatusCode = pLine;
	pLine[i]='\0'; i++;

	*piStatusCode = atoi(pStatusCode);

	/* ---
	We ignore the status text and the whole rest
	--- */

	/* -- OK --- */
	return INT_CONTINUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Verify the HTTP request line and divide it into components. Returns
	INT_CONTINUE on success or an HTTP error status code on error.

	Assumes pLine is NULL terminated at pLine[iLineLen].

	On success instantiates all values to at least an empty string with the
	exception of the proxy values which will all be instantiated if
	ppProxyProtocol is instantiated.

\*____________________________________________________________________________*/
static int Internal_SplitRequestLine(
	char *pLine, 
	int iLineLen,
	const char **ppMethod,
	const char **ppProxyProtocol,
	const char **ppProxyHost,
	int *piProxyHostLen,
	const char **ppProxyPort,
	int *piProxyPortLen,
	const char **ppPath,
	const char **ppQueryString,
	const char **ppProtocol
	)
{
	assert( pLine );
	assert( iLineLen );
	assert( ppMethod );
	assert( ppProxyProtocol );
	assert( ppProxyHost );
	assert( piProxyHostLen );
	assert( ppProxyPort );
	assert( piProxyPortLen );
	assert( ppPath );
	assert( ppProtocol );

	*ppProtocol = "";	/* --- this is done because protcol is optional, i.e.
						HTTP/0.9 doesn't send one
						--- */
	/* 
	**  check URI for exluded characters acc. to RFC2396 2.4.3
	**  (silently checking the whole request line instead of URI only
	**  and intentional omitting \x00, \x20, '#' and '%' from this check)
	*/
#define deniedchars "\
\x7F\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F\
\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\
<>\"{}|\\^[]`"
	if ( strcspn( pLine, deniedchars ) < strlen( pLine ))
		{ return ST_BADREQUEST; };

	/* ---
	method
	--- */
	int i = 0;
	for( ; i<iLineLen && !(isspace(pLine[i])); i++);
	*ppMethod = pLine;
	if ( i==iLineLen ) { return ST_BADREQUEST; };
	pLine[i]='\0'; i++;
	for(; i<iLineLen && (isspace(pLine[i])); i++);
	pLine = &( pLine[i] ); iLineLen -= i; i=0; 

	/* ---
	pLine now points at the component following the method
	--- */

	/* --- URL path must exist --- */
	if ( i==iLineLen )
		{ return ST_BADREQUEST; };

	/* ---
	proxy host and port
	--- */
	if ( (*pLine!='/') && (*pLine!='*'))
		{
		/* --- assume proxy host and port specification --- */
		for(;
			(i+2)<iLineLen &&
			strncmp(&(pLine[i]), "://", 3) && !(isspace(pLine[i])); i++)
			{ pLine[i] = tolower( pLine[i] ); };
		if ( !strncmp( &(pLine[i]), "://", 3 ) )
			{
			*ppProxyProtocol = pLine; pLine[i] = '\0';
			i += 3;
			pLine = &( pLine[i] ); iLineLen -= i; i=0;

			/* --- pLine now points at proxy host --- */
			*ppProxyHost = pLine;
			for(;	i<iLineLen &&
					pLine[i]!=':' &&
					pLine[i]!='/' &&
					!(isspace(pLine[i])); i++);
			*piProxyHostLen = i;
			if ( pLine[i]==':' ) 
				{
				i++;
				pLine = &( pLine[i] ); iLineLen -= i; i=0;

				/* --- pLine now points at proxy port --- */
				*ppProxyPort = pLine;
				for(; i<iLineLen && pLine[i]!='/' && !(isspace(pLine[i])); i++);
				*piProxyPortLen = i;
				}
			else
				{
				*ppProxyPort = "";
				*piProxyPortLen = 0;
				};
			pLine = &( pLine[i] ); iLineLen -= i; i=0;
			}
		else
			{
			/* ---
			The URL path didn't start with '/' but it wasn't a proxy request

			NOTE: Maybe this should raise a gateway error instead of the
			generic bad request error.

			--- */
			return ST_BADREQUEST;
			};
		};

	/* ---	
	pLine now points at first character of URL path
	--- */

	/* ---
	path and query string.
	--- */
	for(; i<iLineLen && pLine[i]!='?' && !(isspace(pLine[i])); i++);
	*ppPath = pLine;

	if ( pLine[i]=='?' )
		{ 
		pLine[i++] = '\0';
		*ppQueryString = &(pLine[i]);
		for(; i<iLineLen && !(isspace(pLine[i])); i++);
		}
	else
		{ *ppQueryString = ""; };

	// decode long file names with spaces
    char *pdest = strchr( *ppPath, '+' );
    while (pdest != NULL) {
		if (strlen(*ppQueryString) && (pdest > *ppQueryString)) break;
		pdest[0] = ' ';
		pdest = strchr( *ppPath, '+' );
	};

	if ( i==iLineLen ) { return INT_CONTINUE; };
	pLine[i++]='\0';
	for(; i<iLineLen && (isspace(pLine[i])); i++);
	pLine = &( pLine[i] ); iLineLen -= i; i=0; 

	/* ---
	pLine now points at the component following the URL path and
	query string.
	--- */

	/* ---
	protocol is last component on the line, skip components until the last
	--- */
	for(;;)
		{
		for(; i<iLineLen && !(isspace(pLine[i])); i++);
		*ppProtocol = pLine;
		for(; i<iLineLen && (isspace(pLine[i])); i++);
		if ( i==iLineLen ) { return INT_CONTINUE; }; /* this should be true */

		/* --- I guess not --- */
		pLine = &( pLine[i] ); iLineLen -= i; i=0;
		};

	/* -- OK --- */
	return INT_CONTINUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns INT_CONTINUE on success, otherwise HTTPCore_readHeaders() returns
	an HTTP error status code.
\*____________________________________________________________________________*/
int HTTPCore_readHeaders( 
	PIIOBuffer *pB,	/* Buffer */
	PIDB *pDB,		/* Request DB */
	int iFlags )
{
	assert( pB );
	PIIOBuffer &tB = *pB;

	char szLocalBuffer[ LOCAL_BUFFER_LEN+1 ];
	szLocalBuffer[ LOCAL_BUFFER_LEN ]='\0';

	int iLen = tB.GetLine( szLocalBuffer, LOCAL_BUFFER_LEN );
	if ( iLen<0 )
		{ return ST_INTERNALERROR; }
	else if ( iLen==0 )
		{ return ST_BADREQUEST; }
	if ( iLen==LOCAL_BUFFER_LEN )
		/* --- HTTP/1.1 status code --- */
		{ return ST_REQUESTURITOOLONG; }

	szLocalBuffer[iLen]='\0';

	/* ---
	Get status line
	--- */
	const char *pMethod = 0;
	const char *pProxyProtocol = 0;
	const char *pProxyHost = 0;
	int iProxyHostLen = 0;
	const char *pProxyPort = 0;
	int iProxyPortLen = 0;
	const char *pURI = 0;
	const char *pQueryString = 0;
	const char *pProtocol = 0;
		
	int iStatus;
	if ( iFlags==RH_CLIENTREQUEST )
		{
		/* ---
		Save the raw request line in the CLF header field
		--- */
		PIDB_add( pDB, PIDBTYPE_STRING, pFKCLF, szLocalBuffer,
			PIDBFLAG_FASTKEY );

		/* --- parse the line --- */
		iStatus = Internal_SplitRequestLine(
			szLocalBuffer,
			iLen,
			&pMethod,
			&pProxyProtocol,
			&pProxyHost,
			&iProxyHostLen,
			&pProxyPort,
			&iProxyPortLen,
			&pURI,
			&pQueryString,
			&pProtocol
			);
		
		if ( iStatus )
			{ return iStatus; };
		if ( !pMethod || !pURI || !pQueryString || !pProtocol )
			{ return ST_BADREQUEST; };

		/* --- store these headers --- */
		PIDB_add( pDB, PIDBTYPE_STRING, pFKMethod, (void *)pMethod,
			PIDBFLAG_FASTKEY );
		PIDB_add( pDB, PIDBTYPE_STRING, pFKURI, (void *)pURI, 
			PIDBFLAG_FASTKEY );
		PIDB_add( pDB, PIDBTYPE_STRING, pFKQueryString, (void *)pQueryString, 
			PIDBFLAG_FASTKEY );
		PIDB_add( pDB, PIDBTYPE_STRING, pFKProtocol, (void *)pProtocol,
			PIDBFLAG_FASTKEY );

		/* --- Is this a proxy request? --- */
		if ( pProxyProtocol )
			{
			PIDB_add( pDB, PIDBTYPE_STRING, pFKProxyProtocol,
				(void *)pProxyProtocol, PIDBFLAG_FASTKEY );
			PIString sProxyHost( pProxyHost, iProxyHostLen );
			PIDB_add( pDB, PIDBTYPE_STRING, pFKProxyHost,
				(void *)(const char *)sProxyHost, PIDBFLAG_FASTKEY );
			PIString sProxyPort( pProxyPort, iProxyPortLen );
			PIDB_add( pDB, PIDBTYPE_STRING, pFKProxyPort,
				(void *)(const char *)sProxyPort, PIDBFLAG_FASTKEY );
			};

		/* ---
		Convert method and protocol into internal numeric form	
		--- */
		int iMethod = MD_UNKNOWN;
		int iProtocol = PR_UNKNOWN;
		if ( !PIUtil_stricmp( pMethod, MD_NAME_GET ) )
			{ iMethod = MD_GET; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_POST ) )
			{ iMethod = MD_POST; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_HEAD ) )
			{ iMethod = MD_HEAD; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_PUT ) )
			{ iMethod = MD_PUT; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_DELETE ) )
			{ iMethod = MD_DELETE; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_OPTIONS ) )
			{ iMethod = MD_OPTIONS; }
		else if ( !PIUtil_stricmp( pMethod, MD_NAME_TRACE ) )
			{ iMethod = MD_TRACE; };
		if ( *pProtocol )
			{
			/* --- any protocol is not HTTP/0.9 --- */
			if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP10 ) )
				{ iProtocol = PR_HTTP10; }
			else if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP11 ) )
				{ iProtocol = PR_HTTP11; };
			}
		else
			{ iProtocol = PR_HTTP09; };

		/* --- Do we support this method and protocol ? --- */
		if ( iMethod == MD_UNKNOWN )
			{ return ST_NOTIMPLEMENTED; };
		if ( iProtocol == PR_UNKNOWN )
			{ iProtocol = PR_HTTP10; };

		/* --- Do we support the method for version < HTTP/1.1 ? --- */
		// Removed because to strict, e.g. for uploads with Netscape
#if 0
		if ((iProtocol != PR_HTTP11) && (iMethod > MD_HEAD))
			{ return ST_BADREQUEST; }; // maybe better ST_NOTIMPLEMENTED?
#endif
		/* --- set the numeric method and protocol --- */
		PIDB_replace( pDB, PIDBTYPE_OPAQUE, pFKMethodNumber, (void *)iMethod,
				PIDBFLAG_FASTKEY );
		PIDB_replace( pDB, PIDBTYPE_OPAQUE, pFKProtocolNumber,
			(void *)iProtocol, PIDBFLAG_FASTKEY );
		/* --- 
		Don't get the rest of the RFC822 headers if the protocol
		is HTTP/0.9
		--- */
		if ( iProtocol==PR_HTTP09 )
			{ return INT_CONTINUE; };

		iLen = tB.GetLine( szLocalBuffer, LOCAL_BUFFER_LEN );	
		}
	else if ( iFlags==RH_RESPONSE )
		{
		/* ---
		determine if this is a full response:
			This is considered true if both the following are true	
			- The line starts with 'HTTP/' or 'http/'
			- The line does not contain the ':' character
		--- */
		if ( !PIUtil_strncmpi( szLocalBuffer, "HTTP/", 5 ) )
			{
			int i=0;
			for(; i<iLen && szLocalBuffer[i]!=':'; i++);
			if ( i==iLen )
				{
				/* ---
				this is a full response, add this line to the DB as CLF
				and get the next line
				--- */
				PIDB_add( pDB, PIDBTYPE_STRING, pFKCLF, szLocalBuffer,
					PIDBFLAG_FASTKEY );

				int iStatusCode = 0;
				
				iStatus = Internal_SplitResponseLine(
					szLocalBuffer,
					iLen,
					&pProtocol,
					&iStatusCode
				);

				if ( iStatus )
					{ return iStatus; };
				if ( !pProtocol || iStatusCode < 100 || iStatusCode > 599 )
					{ return ST_BADREQUEST; };

				/* ---
				Convert protocol into internal numeric form	
				--- */
				int iProtocol = PR_UNKNOWN;
				if ( *pProtocol )
					{
					/* --- any protocol is not HTTP/0.9 --- */
					if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP10 ) )
						{ iProtocol = PR_HTTP10; }
					else if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP11 ) )
						{ iProtocol = PR_HTTP11; };
					}
				else
					{ iProtocol = PR_HTTP09; };

				if ( iProtocol == PR_UNKNOWN )
					{ iProtocol = PR_HTTP10; };

				/* --- store the protocol name --- */
				PIDB_add( pDB, PIDBTYPE_STRING, pFKProtocol, (void *)pProtocol,
					PIDBFLAG_FASTKEY );

				/* --- set the numeric protocol and statuscode --- */
				PIDB_add( pDB, PIDBTYPE_OPAQUE, pFKProtocolNumber,
					(void *)iProtocol, PIDBFLAG_FASTKEY );

				PIDB_add( pDB, PIDBTYPE_OPAQUE, pFKStatusCode,
					(void *)iStatusCode, PIDBFLAG_FASTKEY );

				iLen = tB.GetLine( szLocalBuffer, LOCAL_BUFFER_LEN );
				};
			};
		};
	while( iLen>0 )
		{
		int i=0;
		for(; i<iLen && szLocalBuffer[i]!=':'; i++);
		szLocalBuffer[i++]='\0';
		for(; i<iLen && (isspace(szLocalBuffer[i])); i++);
		PIDB_add( pDB,
			PIDBTYPE_RFC822,
			szLocalBuffer,
			(void *)&(szLocalBuffer[i]),
			0
			);
		iLen = tB.GetLine( szLocalBuffer, LOCAL_BUFFER_LEN );	
		};

	/* --- The 'Host' Header is mandatory for HTTP/1.1 --- */
	if ((iFlags==RH_CLIENTREQUEST) && ((const int)PIDB_lookup(pDB,
		PIDBTYPE_OPAQUE, pFKProtocolNumber, PIDBFLAG_FASTKEY) == PR_HTTP11)
	&& !PIDB_lookup(pDB, PIDBTYPE_RFC822, KEY_HTTP_HOST, 0)) {
		return ST_BADREQUEST;
	};

	/* ---
	Done!
	--- */
	return INT_CONTINUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_sendGeneralHeaders( PIHTTP *pPIHTTP )
{
	assert( pPIHTTP );
	PIHTTP &tPIHTTP = *pPIHTTP;

	/* --- 
	Don't do anything for subrequests
	--- */
	if ( tPIHTTP.IsSubRequest() )
		{ return PIAPI_COMPLETED; };

	PIIOBuffer &tB = tPIHTTP.GetIOBuffer();
	PIDB *pQ = tPIHTTP.GetRequestDB();
	PIDB *pR = tPIHTTP.GetResponseDB();

	/* --- get status line and headers --- */
	int iStatus = tPIHTTP.iStatus;
	if ( !iStatus )
		{
		/* ---
		If the status is 0 then silently change it to 200 OK
		--- */
		tPIHTTP.iStatus = ST_OK;
		iStatus = ST_OK;
		};

	int iProtocol = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, pFKProtocolNumber,
		PIDBFLAG_FASTKEY );	

	/* --- send status line and headers if this is not HTTP/0.9 --- */
	if ( iProtocol!=PR_HTTP09 )
		{
		switch( iProtocol )
			{
			case PR_UNKNOWN:
			case PR_HTTP10:
				tB.Write( PR_NAME_HTTP10, sizeof( PR_NAME_HTTP10 )-1 ); break;
			case PR_HTTP11:
				tB.Write( PR_NAME_HTTP11, sizeof( PR_NAME_HTTP11 )-1 ); break;
			default:
				assert( 0 );
				return PIAPI_ERROR;
			};

		switch( iStatus )
			{
#define TMP( pattern ) { \
			tB.WriteLn( (pattern), sizeof( pattern )-1 ); break; }
			/* ---
			HTTP 1.0
			--- */
			case ST_OK:						TMP( " 200 OK" );
			case ST_PERMANENTREDIRECT:		TMP( " 301 Permanent Redirect" );
			case ST_FOUND:				if (iProtocol == PR_HTTP11) {
											TMP( " 302 Found" );
										} else {
											TMP( " 302 Temporary Redirect" );
										};
			case ST_NOTMODIFIED:			TMP( " 304 Not Modified" );
			case ST_BADREQUEST:				TMP( " 400 Bad Request" );
			case ST_UNAUTHORIZED:			TMP( " 401 Unauthorized" );
			case ST_FORBIDDEN:				TMP( " 403 Forbidden" );
			case ST_NOTFOUND:				TMP( " 404 Not Found" );
			case ST_INTERNALERROR:			TMP( " 500 Internal Error" );
			case ST_NOTIMPLEMENTED:			TMP( " 501 Not Implemented" );
			case ST_BADGATEWAY:				TMP( " 502 Bad Gateway" );
			case ST_SERVICEUNAVAILABLE:		TMP( " 503 Service Unavailable" );
			case ST_CREATED:				TMP( " 201 Created" );
			case ST_ACCEPTED:				TMP( " 202 Accepted" );
			case ST_NOCONTENT:				TMP( " 204 No Content" );
			/* ---
			HTTP 1.1 
			--- */
			case ST_CONTINUE:				TMP( " 100 Continue" );
			case ST_SWITCHINGPROTOCOLS:		TMP( " 101" );

			case ST_NONAUTHORITATIVEINFO:	TMP( " 203 None-Authoritative Information" );
			case ST_RESETCONTENT:			TMP( " 205 Reset Content" );
			case ST_PARTIALCONTENT:			TMP( " 206 Partial Content" );

			case ST_MULTIPLECHOICES:		TMP( " 300 Multiple Choices" );
			case ST_SEEOTHER:				TMP( " 303 See Other" );
			case ST_USEPROXY:				TMP( " 305 Use Proxy" );
			case ST_UNUSED:					TMP( " 306 (Unused)" );
			case ST_TEMPORARYREDIRECT:		TMP( " 307 Temporary Redirect" );

			case ST_PAYMENTREQUIRED:		TMP( " 402 Payment Required" );
			case ST_METHODNOTALLOWED:		TMP( " 405 Method Not Allowed" );
			case ST_NOTACCEPTABLE:			TMP( " 406 Not Acceptable" );
			case ST_PROXYAUTHREQUIRED:		TMP( " 407 Proxy Authentication Required" );
			case ST_REQUESTTIMEOUT:			TMP( " 408 Request Timeout" );
			case ST_CONFLICT:				TMP( " 409 Conflict" );
			case ST_GONE:					TMP( " 410 Gone" );
			case ST_LENGTHREQUIRED:			TMP( " 411 Length Required" );
			case ST_PRECONDITIONFAILED:		TMP( " 412 Precondition Failed" );
			case ST_REQUESTENTITYTOOLARGE:	TMP( " 413 Request Entity Too Large" );
			case ST_REQUESTURITOOLONG:		TMP( " 414 Request-URI Too Long" );
			case ST_UNSUPPORTEDMEDIATYPE:	TMP( " 415 Unsupported Media Type" );
			case ST_RANGENOTSATISFIABLE:	TMP( " 416 Request Range Not Satisfiable" );
			case ST_EXPECTATIONFAILED:		TMP( " 417 Expectation Failed" );

			case ST_GATEWAYTIMEOUT:			TMP( " 504 Gateway Timeout" );
			case ST_UNSUPPORTEDVERSION:		TMP( " 505 HTTP Version Not supported" );

			default:
#undef TMP
#if !defined(NDEBUG)
				cerr << iStatus << endl;
#endif
				assert( 0 );
				return PIAPI_ERROR;
			};

		/* ---
		Send general header fields 
		--- */

		/* --- server --- */
		if ( pSG->sServerStamp!=PIString::Empty() )
			{
#define TMP "Server: "
			tB.Write( TMP, sizeof( TMP )-1  );
			tB.WriteLn( pSG->sServerStamp, -1 );
			};
#undef TMP		
		/* --- date --- */
		Internal_SendHeader_Date( tB );
		
		/* --- Connection --- */
		Internal_SendHeader_Connection( tPIHTTP );

		/* --- TransferEncoding --- */
#define TMP "Transfer-Encoding: "
		if (const char *pTE = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			pFKTransferEnc, PIDBFLAG_FASTKEY ))
			{
			tB.Write( TMP, sizeof( TMP )-1 );
			tB.WriteLn( pTE, -1 );
			};
#undef TMP
		};	

	/* --- 
	Done!
	--- */
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_sendEntityHeaders( PIHTTP *pPIHTTP, PIDB *pR )
{
	assert( pPIHTTP );
	PIHTTP &tPIHTTP = *pPIHTTP;

	/* --- 
	Don't do anything for subrequests
	--- */
	if ( tPIHTTP.IsSubRequest() )
		{ return PIAPI_COMPLETED; };

	PIDB *pQ = tPIHTTP.GetRequestDB();
	PIIOBuffer &tB = tPIHTTP.GetIOBuffer();

	int iProtocol = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, pFKProtocolNumber,
		PIDBFLAG_FASTKEY );	

	if ( iProtocol==PR_HTTP09 )			/* --- no header for HTTP/0.9 --- */
		{ return PIAPI_COMPLETED; };

	/* ----
	Do RFC822 response headers
	--- */
	PIDBIterator *pIter = PIDB_getIterator( pR, PIDBTYPE_RFC822, 0, 0 );
	if ( pIter )
		{
		for(;
				PIDBIterator_atValidElement( pIter );
				PIDBIterator_next( pIter ) )
			{
			const char *pKey;
			const char *pValue = (const char *)
			PIDBIterator_current( pIter,  &pKey );
			if (!PIUtil_stricmp(pKey, KEY_HTTP_SERVER)) { continue; };
			if (!PIUtil_stricmp(pKey, KEY_HTTP_DATE)) { continue; };
			if (!PIUtil_stricmp(pKey, KEY_HTTP_CONNECTION)) { continue; };
			if (!PIUtil_stricmp(pKey, KEY_HTTP_TRANSFERENCODING)) { continue; };
			if (!pValue || !*pValue) { continue; };
			tB.Write( pKey, -1 );
			tB.Write( ": ", 2 );
			tB.WriteLn( pValue, -1 );
			};
		PIDBIterator_delete( pIter );
		};
	tB.WriteLn( "", 0 );

	/* --- 
	Done!
	--- */
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_loadMIMEFile( PIObject *pObject, PIDB *pMIMEDB,
	const char *pMIMEFile )
{
	assert( pMIMEFile );
	if ( !pMIMEFile )
		{ return 0; };

	ifstream ifs( pMIMEFile );
	enum { MIME_LINE_MAX=2047 };
	char szMIMEBuf[MIME_LINE_MAX+1];
	szMIMEBuf[MIME_LINE_MAX]='\0';
	if ( !ifs.good() || ifs.eof() )
		{
		enum { TMP_BUF=2048 };
		char szTmpBuf[TMP_BUF];
		sprintf( szTmpBuf, "HTTP1_0: mime file '%s' \
is empty or does not exist.", pMIMEFile );
		CONFIG_ERR( pObject, szTmpBuf );
		return 0;
		};
	while( ifs.good() )
		{
		ifs.getline( szMIMEBuf, MIME_LINE_MAX );
		char *pszMIMEBuf = szMIMEBuf;
		if ( ifs.eof() )
			{ break; };
		int i=0;

		/*
		** Scan to first non-space 
		*/
		for(; pszMIMEBuf[i] && (isspace(pszMIMEBuf[i])); i++);

		/*
		** Is this line empty
		*/
		if ( !pszMIMEBuf[i] || pszMIMEBuf[i]=='#' )
			{ continue; /* comment line */ };

		/*
		** Make pMIMEType point to start of media definition
		*/
		const char *pMIMEType=&( pszMIMEBuf[i] );

		/*
		** Scan to end of media definition (first whitespace)
		*/
		for(; pszMIMEBuf[i] && !(isspace(pszMIMEBuf[i])); i++);
		int bMoreExtns = 0;
		if ( pszMIMEBuf[i] )
			{
			bMoreExtns = 1;
			pszMIMEBuf[i++] = '\0';
			};
		for(; bMoreExtns;)
			{
			/*
			** Scan past whitespace before extension
			*/
			for(; pszMIMEBuf[i] && (isspace(pszMIMEBuf[i])); i++);

			/* if line is empty, skip it */
			if ( !pszMIMEBuf[i] || pszMIMEBuf[i]=='#' )
				{ break; /* line is empty or comment line */ };

			/*
			** Mark start of extension
			*/
			pszMIMEBuf = &( pszMIMEBuf[i] ); i=0;

			/*
			** Skip to end of extension
			*/
			for(; pszMIMEBuf[i] && !(isspace(pszMIMEBuf[i])); i++);
			if ( pszMIMEBuf[i] )
				{
				bMoreExtns = 1;
				pszMIMEBuf[i++] = '\0';
				};
			PIString sExtn( pszMIMEBuf );
#if WIN32
			/*
			** Silently convert to upper case
			*/
			sExtn.ConvertToUpperCase();
#endif
			const char *pExtn = sExtn;
			const char *pPreviousMIMEType = (const char *)PIDB_lookup(
				pMIMEDB, PIDBTYPE_STRING, pExtn, 0 );
			if ( pPreviousMIMEType )
				{
				enum { TMP_BUF=2048 };
				char szTmpBuf[TMP_BUF];
				sprintf( szTmpBuf, "HTTP1_0: reading mime file '%s' \
extension '%s' is already mapped to MIME type '%s'.", pMIMEFile, pExtn,
				pPreviousMIMEType );
				CONFIG_ERR( pObject, szTmpBuf );
				return 0;
				};
			if (PIDB_add( pMIMEDB, PIDBTYPE_STRING, pExtn,
				(void *)pMIMEType, 0 ))
				{
				CONFIG_ERR( pObject, "HTTP1_0: internal error \
dding MIME-type." );
				return 0;
				};
			};
		};
	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PIFInfo *HTTPCore_getCachedFile( const char *pPath )
{
	return PIFInfo_new( pPath );
} 

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void HTTPCore_releaseCachedFile( PIFInfo *pFInfo )
{
	PIFInfo_delete( pFInfo );
} 

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_relativeToAbsolutePath( PIDB *pDB,
    const char *pRelativePath, Pi3String *pResult )
{
    assert( pRelativePath );
    assert( pResult );
    assert( pSG );

    /*
    ** Get the server or Host root from the DB
    */
    const char *pRoot = 0;
    if ( pDB )
        {
        pRoot = (const char *)PIDB_lookup( pDB, PIDBTYPE_STRING,
            pFKHostRoot, PIDBFLAG_FASTKEY | PIDBFLAG_PROPAGATEUP );
        };

    if ( !pRoot )
        {
        pRoot = pSG->sServerRoot;
        };

    return Internal_RelativeToAbsolutePath( pRoot, pRelativePath, *pResult );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPCore_getServerStamp()
{
	assert( pSG );
	return pSG->sServerStamp;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns the MIME type for a file extension
\*____________________________________________________________________________*/
const char *HTTPCore_getMIMETypeFromExtension( const char *pExtn )
{
	assert( pSG );
	const char *pMIME = 0;
#if WIN32
	/*
	** Everything is upper case
	*/
	PIString sExtn( pExtn );
	sExtn.ConvertToUpperCase();
	const char *pExtension = sExtn;
#else
	const char *pExtension = pExtn;
#endif
	if ( !pExtension ||
		!( pMIME = (const char *)PIDB_lookup( pSG->pMIMEDB, PIDBTYPE_STRING,
		pExtension, 0 ) ) )
		{ return pSG->sDefaultMIMEType; };
	return pMIME;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns 0 on success.
\*____________________________________________________________________________*/
int HTTPCore_logError( PIHTTP *pPIHTTP, const char *pFormat, ... )
{
	assert( pSG );
#if TEMP
	PIPLATFORM_FD tError = pSG->tErrorLog;
#else
	FILE *pError = pSG->pErrorLog;
#endif

	va_list tvList;
	
	/* --- get time stamp --- */
	enum { BUF_SIZE=255 };
	char szBuf[BUF_SIZE+1];	

	time_t tT;

	/* --- assume time(), gmtime() and strftime() are not thread safe --- */
	PIPlatform_beforeUnsafeBlock();
	time( &tT ); 

	/* ---
	Don't know of a standard for error logfile time, so the following
	time format will be used

		"Thu Mar 27 19:28:21 1997 GMT"

	--- */
	strftime( szBuf, BUF_SIZE, "[%a %b %d %H:%M:%S %Y GMT] ", gmtime( &tT ) );
	PIPlatform_afterUnsafeBlock();

#if TEMP
	/* ---
	Generating an error message is comparatively infrequent, so can do
	things a little less efficientlty
	--- */
    enum { INCREMENT = 8192 };
    int iMaxLen = INCREMENT;
	int iLen = strlen( szBuf );
again:
	char *pBuffer = PI_NEW( char[iMaxLen+1] );	/* unfortunately we don't 
									        	have snprintf, well we
                                                do under MSVC ... */
    if ( iMaxLen < iLen )
        {
        iMaxLen += ( iLen - iMaxLen ) + INCREMENT;
        delete [] pBuffer;
        goto again;
        };
	strcpy( pBuffer, szBuf );
    iMaxLen -= iLen;

	/* --- print message --- */
	va_start( tvList, pFormat );
    if ( iMaxLen > 0 )
        {
#if WIN32
        if ( _vsnprintf( &(pBuffer[iLen]), iMaxLen, pFormat, tvList )==-1 )
#else
        if ( vsprintf( &(pBuffer[iLen]), pFormat, tvList )==-1 )
#endif
            {
            /*
            ** Not enough space for data 
            */
            iMaxLen += INCREMENT;
            iMaxLen *= 2;           /* make memory increments go up
                                    binary exponentially
                                    */
            delete [] pBuffer;
            goto again;
            };
        };
	va_end( tvList );

	iLen += strlen( &(pBuffer[iLen]) );
#if WIN32	
	pBuffer[iLen++] = '\r';
#endif
	pBuffer[iLen++] = '\n';
	pBuffer[iLen] = '\0';
	PIFile_writeAtomic( tError, iLen, pBuffer );

	/* ---
	Add the message to the response DB
	--- */
	if ( pPIHTTP && pPIHTTP->pResponseDB )
		{
		PIDB_add( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_ERRORMESSAGE,
			(void *)pBuffer, 0 );
		};

	PI_DELETE( [] pBuffer );

#else
	/* --- print time stamp --- */
	fputs( szBuf, pError );

	/* --- print message --- */
	va_start( tvList, pFormat );
	vfprintf( pError, pFormat, tvList );
	va_end( tvList );

	/* --- print carriage return and flush message --- */
	fputc( '\n', pError );
	fflush( pError );
#endif

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns 0 on success.
\*____________________________________________________________________________*/
int HTTPCore_logDebug( int /* iLevel */, const char *pFormat, ... )
{
	assert( pSG );
	FILE *pDebug = pSG->pDebugLog;
	if ( !pDebug )
		{ return 0; /* --- debugging not switched on --- */ };

	va_list tvList;
	
	/* --- print message --- */
	va_start( tvList, pFormat );
	vfprintf( pDebug, pFormat, tvList );
	va_end( tvList );

	/* --- print carriage return and flush message --- */
	fputc( '\n', pDebug );
	fflush( pDebug );

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_executePhase(
	PIHTTP *pPIHTTP,
	PIObject *(* nextFn)(void *),
	void *pArg )
{
	PIObject *pHandler = 0;
	pHandler = (nextFn)( pArg );
	int iResult = PIAPI_CONTINUE;
	while( pHandler )
		{
		/* --- buffer for debug logging --- */
		enum { BUF_SIZE=1023 };
		char szBuf[BUF_SIZE+1];

		/* --- set debug information --- */
		pPIHTTP->pHandlerName = PIObject_getName( pHandler );
		pPIHTTP->iLastResultCode = PIAPI_EINVAL;

		/* --- delta information for profiling --- */
		long lSeconds = 0;
		long lMicroSeconds = 0;

		/* --- log before handler debug --- */
		if ( pSG->pDebugLog )
			{
			if ( pSG->pBeforeHandler )
				{
				int iLen = Pi3Expression_write( pSG->pBeforeHandler, pPIHTTP, 0,
					szBuf, BUF_SIZE );
				if ( iLen!=-1 )
					{
					if ( iLen>BUF_SIZE ) { iLen=BUF_SIZE; };
					szBuf[iLen] = '\0';	
					HTTPCore_logDebug( DBG_LOW, "%s", szBuf );
					};
				};
			pPIHTTP->iDepth++;
			PITimer_elapsed( pPIHTTP->pTimer, &lSeconds, &lMicroSeconds );
			};

		iResult = PILogic_execute( pHandler, 1, (const char **)&pPIHTTP );

		/* --- log after handler debug --- */
		if ( pSG->pDebugLog )
			{
 			if ( pSG->pAfterHandler )
				{
				/* --- get handler time delta for profiling --- */
				long lSeconds2 = 0;
				long lMicroSeconds2 = 0;
				PITimer_elapsed( pPIHTTP->pTimer, &lSeconds2, &lMicroSeconds2 );
				pPIHTTP->iDelta = (int)( ( lMicroSeconds2 - lMicroSeconds ) +
					( ( lSeconds2 - lSeconds ) * 1000000L ) );
				};

			pPIHTTP->iDepth--;
			if ( pSG->pAfterHandler )
				{
				/* --- set debug information name --- */
				pPIHTTP->pHandlerName = PIObject_getName( pHandler );
				pPIHTTP->iLastResultCode = iResult;

				int iLen = Pi3Expression_write( pSG->pAfterHandler, pPIHTTP, 0,
					szBuf, BUF_SIZE );
				if ( iLen!=-1 )
					{
					if ( iLen>BUF_SIZE ) { iLen=BUF_SIZE; };
					szBuf[iLen] = '\0';	
					HTTPCore_logDebug( DBG_LOW, "%s", szBuf );
					};
				};
			};

		switch( iResult )
			{
			case PIAPI_CONTINUE:
				goto done_handler;

			case PIAPI_COMPLETED:
			case PIAPI_ERROR:
			case PIAPI_ABORT:
			case INT_REDIRECT:
				goto done_phase;

			default:
				assert( 0 );
				goto done_phase;
			};	/* end switch */
done_handler:;
		pHandler = (nextFn)( pArg );
		};	/* end while */
done_phase:

	return iResult;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Determine the error code if any for PIAPI_CONTINUE returned
	from a handler phase (no action taken). HTTP error status code
	is returned. A return code of 0 indicates no error.
\*____________________________________________________________________________*/
int Internal_ContinueStatus( int iPhase )
{
	switch( iPhase )
		{
		case PH_INVALID:		return ST_INTERNALERROR;
		case PH_INIT:			return ST_INTERNALERROR;
		case PH_HEADERS:		return ST_INTERNALERROR;
		case PH_HOSTMAP:		return 0;
		case PH_MAPPING:		return ST_INTERNALERROR;
		case PH_CHECKPATH:		return ST_NOTFOUND;
		case PH_CHECKAUTH:		return ST_UNAUTHORIZED;
		case PH_CHECKACCESS:	return ST_FORBIDDEN;
		case PH_CHECKTYPE:		return ST_INTERNALERROR;
		case PH_HANDLE:			return ST_INTERNALERROR;
		case PH_LOG:			return 0;
		case PH_DESTROY:		return 0;
		default:
			return ST_INTERNALERROR;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_SetPath( PIHTTP &tPIHTTP )
{
	int iRet = PIAPI_COMPLETED;

	/* --- 
	Get the URI path
	Use HTTPCore_contractPath() to remove '..' and other evil elements
	Set the 'Path' response variable to the contracted path or
	Set the status to 403 if the path attempted to read up into
	the directory root.
	--- */
	const char *pURI = (const char *)PIDB_lookup( tPIHTTP.GetRequestDB(), 
				PIDBTYPE_STRING, pFKURI, PIDBFLAG_FASTKEY );
	if ( pURI ) 
		{
		char *pTmp = new char[strlen(pURI)+1];
		strcpy( pTmp, pURI ); 

		/* ---
		Then contract it to avoid side effects of '.' and '..' in
		the URL
		--- */
		if ( HTTPUtil_contractPath( pTmp )>0 )
			{
			/* --- force a 403 (forbidden) error --- */
			tPIHTTP.iStatus = ST_FORBIDDEN;
			iRet = INT_REDIRECT;
			}
		else
			{
			PIDB_add( tPIHTTP.GetResponseDB(), PIDBTYPE_STRING,
				pFKPath, pTmp, PIDBFLAG_FASTKEY );
			};

		delete [] pTmp;
		}
	else
		{
		HTTPCore_logError( &tPIHTTP, "Could not set request path\n" );
		iRet = HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
		};

	return iRet;
}

/*____________________________________________________________________________*\
*
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static PIObject *Internal_ExecutePhaseCallback( void *pArg )
{
	DblListIterator *pIter = (DblListIterator *)pArg;
	assert( pIter );
	if ( pIter->BadIndex() )
		{ return 0; };
	PIObject *pObject = (PIObject *)pIter->Current();
	assert( pObject );
	(*pIter)++;
	return pObject;
}

/*____________________________________________________________________________*\
*
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int Internal_ExecutePhase( PIHTTP *pPIHTTP, DblList &lHandlers )
{
	DblListIterator i( lHandlers );
	return HTTPCore_executePhase( pPIHTTP, Internal_ExecutePhaseCallback, &i);
}

/*____________________________________________________________________________*\
*
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_dispatch(
	PIHTTP *pPIHTTP,
	int iStartPhase,
	int iEndPhase )
{
	assert( pPIHTTP );
	PIDB *pC = pPIHTTP->GetConnectionDB();
/*	PIDB *pR = pPIHTTP->GetResponseDB(); */
	PIIOBuffer &tB = pPIHTTP->GetIOBuffer();
	int iRet = PIAPI_COMPLETED;

	int iNumRedirects = 0;
again:
	for( int iPhase = iStartPhase; iPhase<=iEndPhase; iPhase++ )
		{
		/* --- cast away const --- */
		*( (int *)&(pPIHTTP->ciPhase) ) = iPhase;

		/* ---
		If this is the log phase, then flush all output and set bytes sent
		--- */
		if ( iPhase == PH_LOG )
			{
			tB.Flush();
			PIDB_replace( pC, PIDBTYPE_OPAQUE, pFKBytesSent,
				(void *)tB.GetBytesSent(), PIDBFLAG_FASTKEY );
			};

		/* --- execute phase --- */
		iRet = Internal_ExecutePhase( pPIHTTP, pSG->lHandlers );

		/* ---
		Determine action based on return code, iRet, and status, iStatus.
		--- */
		switch( iRet )
			{
			case PIAPI_CONTINUE:
				{
				int iStatus = Internal_ContinueStatus( iPhase );
				if ( !iStatus ) { break; };
				pPIHTTP->iStatus = iStatus;

				/* --- set an error message --- */
				HTTPCore_logError( pPIHTTP, "HTTPCore_dispatch: No processing \
occurred for phase '%s', response status set to '%d'.", \
					HTTPUtil_phaseNumberToName( iPhase ), iStatus );
				};
				/* --- fall through to redirect --- */

			case INT_REDIRECT:
				if ( iPhase<=PH_HANDLE && iNumRedirects<=MAX_REDIRECTS )
					{
					iStartPhase = PH_MAPPING;
					iNumRedirects++;
					goto again;
					};
				return PIAPI_ERROR;

			case PIAPI_COMPLETED:
				if ( iPhase==PH_HEADERS && !pPIHTTP->iStatus &&
					Internal_SetPath( *pPIHTTP )==INT_REDIRECT )
					{
					if ( iNumRedirects<=MAX_REDIRECTS )
						{	
						iStartPhase = PH_MAPPING;
						iNumRedirects++;
						goto again;
						}
					else
						{
						return PIAPI_ERROR;
						};
					};
				break;

			/* --- error conditions --- */
			case PIAPI_ERROR:
			case PIAPI_ABORT:
				return iRet;

			default:
				assert( 0 );
				return PIAPI_ABORT;

			};	/* --- end of switch on phase return code --- */
		};	/* --- end of this phase --- */

	/* --- if the last handler left 'CONTINUE', change to 'COMPLETED' --- */
	if ( iRet==PIAPI_CONTINUE )
		{ iRet = PIAPI_COMPLETED; };

	return iRet;
}

/*____________________________________________________________________________*\
*
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPCore_debugEnabled()
{
	return ( pSG->pDebugLog ) ? 1 : 0 ;
}

