/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Platform/Win16OSE.cpp,v $
 * $Date: 2003/05/13 18:42:14 $
 *
 Description:
	This file provides the implementation of some of the abstract
	platform objects for the Win16 os interface.

\*____________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"

#if defined(CONFIG_OS_WIN16)

#include "Win16OSE.h"
#include "Platform.h"
#include "errno.h"
#include "Utils.h"
#include "PlatDefs.h"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *Win16FATFileSysType::GetExtension( const char *pName ) const
{
	assert( pName );
	int i = strlen( pName ) -1;
	for(; i>=0; i--)
		{
		if ( pName[i]=='.' )
			{ return &( pName[i+1] ); };
		};
	return "";
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	
 Description:
\*____________________________________________________________________________*/
Win16DLL::Win16DLL( const char *pPath )
:	hInstance( 0 ), sPath( pPath ), sLastErrorMsg( "No error" )
{
	assert( pPath );

	/*
	** LATER - library in mellicous path?
	*/
	hInstance = LoadLibrary( pPath );
	if ( hInstance < HINSTANCE_ERROR)
		{
		CaptureWin16Error( hInstance );
		hInstance = 0;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
Win16DLL::~Win16DLL()
{
	if ( hInstance )
		{
		/* WIN16: void FreeLibrary( HINSTANCE ) */
		FreeLibrary( hInstance );
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void *Win16DLL::GetAddress( const char *pSymbolName )
{
	assert( hInstance && pSymbolName );
	if ( !hInstance || !pSymbolName ) 
		{ return 0; };
	void *pAddr = GetProcAddress( hInstance, pSymbolName );
	if ( !pAddr )
		{
		CaptureWin16Error( 0 );
		};
	return pAddr;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public
 Description:
\*____________________________________________________________________________*/
void Win16DLL::CaptureWin16Error( HINSTANCE hInstance )
{
	DWORD dwError = hInstance;
	PIRETERR( dwError );
	PIString sTmp;
	pi_ltoa( dwError, 10, sTmp );
	sLastErrorMsg = "Win16 error code: ";
	sLastErrorMsg.Concatenate( sTmp );	
}

#if 0
// LATER no file mapping in win 16
/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
Win16MappedFile::Win16MappedFile( FileSystemEntity &tFSE )
:	hFile( HFILE_ERROR ),
	hMapping( 0 ),
	lpAddress( 0 ),
	dwLength( 0 )
{
	Setup( tFSE.GetPath(), tFSE.GetFileSize() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
Win16MappedFile::~Win16MappedFile()
{
	if ( lpAddress )
		{ DoUnlock(); };

	if ( hMapping != 0 )
		{
		if ( ::_lclose( hMapping ) != TRUE )
			{ PIOSERR; };
		};

	if ( hFile != HFILE_ERROR )
		{
		if ( ::_lclose( hFile ) != TRUE )
			{ PIOSERR; };
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		private:
 Description:
\*____________________________________________________________________________*/
void Win16MappedFile::Setup( const char *pFileName, DWORD dwTheLength )
{
	assert( pFileName );
	if ( !pFileName ) { PIERROR( PIAPI_EINVAL ); return; };

	/* ---
	Create a file handle
	--- */
	hFile = CreateFile(
		pFileName,
		GENERIC_READ,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		0 );
	if ( hFile == HFILE_ERROR )
		{ PIOSERR; return; };

	/* ---
	Create a file mapping
	--- */
	hMapping = CreateFileMapping(
		hFile,
		0,
		PAGE_READONLY,
		0,
		0,
		0 );
	if ( hMapping == 0 )
		{ PIOSERR; return ; };

	dwLength = dwTheLength;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		private
 Description:
\*____________________________________________________________________________*/
void *Win16MappedFile::DoLock( int *piLength )
{
	if ( hMapping == 0 )
		{ PIERROR( PIAPI_EINVAL ); return 0; };

	if ( lpAddress )
		{ return lpAddress; };

	/* ---
	Map the file into memory
	--- */
	lpAddress = MapViewOfFile(
		hMapping,
		FILE_MAP_READ,
		0,
		0,
		0 /* Entire file */ );
	if ( !lpAddress )
		{ PIOSERR; return 0; };
	
	if ( piLength )
		{ *piLength = dwLength;	};
	return lpAddress;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		private
 Description:
\*____________________________________________________________________________*/
bool Win16MappedFile::DoUnlock()
{
	if ( !lpAddress )
		{ return false; };

	if ( UnmapViewOfFile( lpAddress )==0 )
		{ PIOSERR; return false; };	

	lpAddress = 0;
	return true;
}
#endif // LATER - no file mapping in win16

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
Win16Timer::Win16Timer()
{
	Reset();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
bool Win16Timer::Reset()
{
	dwTime = ::GetTickCount();
	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
bool Win16Timer::Elapsed( long *plSeconds, long *plMicroSeconds )
{
	DWORD dwElapsed = ::GetTickCount() - dwTime;
	if ( plSeconds )
		{ *plSeconds = dwElapsed / 1000; };
	if ( plMicroSeconds )
		{ *plMicroSeconds = (dwElapsed % 1000) * 1000; };
	return true;
}

#endif	/* if defined(CONFIG_OS_WIN16) */

