/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/JspHandler.cpp,v $
 * $Date: 2003/05/13 18:42:18 $
 *
 Description:
	Execute an servlet extension.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <stdarg.h>

#include "JspHandler.h"
#include "FlexLexer.h"
#include "jsp_parser.h"
#include "jsp_scanner.h"

#ifdef POSIX
#	include <unistd.h>
#	include <errno.h>
#	include <sys/wait.h>
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	Jsp

Description:
	Parse, compile and execute a JSP page. You can find more informations
	about JSP and Servlets here <A HREF="http://java.sun.com/">
    http://java.sun.com/</A>.

Options:

<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>JavaCompiler
<TD>"javac"
<TD>Quoted string
<TD>Optional command line for the java compiler
<TD>"/usr/local/jdk1.3.1/bin/javac"

<TR>
<TD>ClassPath
<TD>-
<TD>Quoted string
<TD>Optional classpath set for the java compiler
<TD>"Pi3Servlet.jar;servlet.jar"

<TR>
<TD>Context
<TD>+
<TD>Valid Pi3 object Reference
<TD>Reference to servlet container
<TD>ServletContainer

<TR>
<TD>Extends
<TD>-
<TD>Quoted string
<TD>Valid base class used by this Jsp.
<TD>"Pi3HttpJspPage"

<TR>
<TD>Init
<TD>-
<TD>Quoted string
<TD>Servlet config option(s)
<TD>"color=red"

<TR>
<TD>Encoding
<TD>-
<TD>Quoted string
<TD>Encoding of the java source files
<TD>"UTF8"

<TR>
<TD>OutputPath
<TD>-
<TD>Quoted string
<TD>Valid path name, where the compiled Jsp source files (.class) will be stored
<TD>"c:\Program Files\jdk1.3.1\jre\lib"

<TR>
<TD>SourcePath
<TD>-
<TD>Quoted string
<TD>Valid path name, where the Jsp source files (.java) will be stored
<TD>"..\Jsp"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>Context, Init</H5>
This options are the same as used for the class Servlet.

<H5>JavaCompiler</H5>
The java compiler will be invoked on POSIX systems by this command line.
On Windows, the <i>sun.tools.javac.Main</i> interface is used.

<H5>ClassPath</H5>
The java compiler will use this as the class path, if set, when a Jsp source is compiled.

<H5>Extends</H5>
This parameter is a valid class name, to be used as the base class of a Jsp
by the Jsp parser. This value is optional and needs to be overwritten only
in special cases.

<H5>Encoding</H5>
This optional parameter sets the encoding of the .java files for the javac compiler.
The default value is UTF8.

<H5>OutputPath</H5>
The java compiler needs a valid path name, where the compiled Jsp source files
(.class) will be stored. This path must be part of the CLASSPATH.

<H5>SourcePath</H5>
The built in JSP parser will write the output Jsp source files (.java) into this
folder. The java compiler looks for this source files in this folder.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED, PIAPI_ERROR or INT_REDIRECT according to the result of
	JNI calls and the status returned by the servlet.

Note:
Example:
	<PRE>
	&lt;Object&gt;
		Name Jsp
		Class JspClass
		Context ServletContainer
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle Condition="&cmp(&dblookup(response,string,ObjectMap,SERVLET),Servlet)" Jsp
		...
	&lt;/Object&gt;
	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif


int Jsp::Parameter( const char *pVariable, const char *pValue,
	const char *pWhere )
	{
	assert( pVariable && pValue );
	PIOStrStream os;
	os << pWhere << GetLogName() << ": ";

	if ( !PIUtil_stricmp( KEY_CONF_EXTENDS, pVariable ) )
		{
		sExtendsName = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_SOURCEPATH, pVariable ) )
		{
		sSourcePath = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_OUTPUTPATH, pVariable ) )
		{
		sOutputPath = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_ENCODING, pVariable ) )
		{
		sEncoding = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_CLASSPATH, pVariable ) )
		{
		sClassPath = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_COMPILER, pVariable ) )
		{
		sCompiler = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_CLASSNAME, pVariable ) )
		{
		}
	else if ( !PIUtil_stricmp( KEY_CONF_SERVLETNAME, pVariable ) )
		{
		}
	else if ( !PIUtil_stricmp( KEY_CONF_INITPARAM, pVariable ) )
		{
		}
	else if ( !PIUtil_stricmp( KEY_CONF_SERVLETCONTEXT, pVariable ) )
		{
		}
	else 
		{
		return Servlet::Parameter( pVariable, pValue, pWhere );
		};

	return 1;
	};

Jsp::Jsp( PIObject *pObject, int iArgc, const char *ppArgv[], const char *pLogName )
	:	Servlet( pObject, iArgc, ppArgv, pLogName ),
		sExtendsName( DEFAULTEXTENDS ),
		sClassPath( 0 ),
		sSourcePath( 0 ),
		sOutputPath( 0 ),
		sEncoding( "UTF8" ),
		sCompiler( "javac" )
{
	if (IsOK()) { ReadParameters( iArgc, ppArgv ); };
};


int Jsp::Compile( PIHTTP &tPIHTTP, const char *pSourcePath )
{
	int iRet = PIAPI_COMPLETED;
#ifdef WIN32
    JNIEnv *env = 0;       /* pointer to native method interface */

	jclass ServletUtilsClass = 0;
	jobject ServletUtilsObject = 0;
	jclass JspCompilerClass = 0;
	jobject JspCompilerObject = 0;
	jthrowable Exception = 0;

	PIDB *pR = tPIHTTP.pResponseDB;

	if (!PIHTTP_isChild( &tPIHTTP )) {
		iRet = JVM()->AttachCurrentThread((void **)&env, NULL);
		if (iRet) {
			Log( &tPIHTTP, "%s: Error when trying to instantiate JVM.", GetLogName());
			return PIAPI_ERROR;
		} else {
			Log( &tPIHTTP, "%s: JVM attached.", GetLogName());
		};
	} else {
		env = (JNIEnv *)PIDB_lookup( pR, PIDBTYPE_OPAQUE, "SERVLET_ENV", 0 );
	};
	
	/* get the ServletUtils class */
	ServletUtilsClass = env->FindClass("org/pi3/servlet/util/Pi3ServletUtils");
	if (!ServletUtilsClass) { goto init_error; }

	/* get the ServletUtils constructor */
	jmethodID ServletUtilsMID;
	ServletUtilsMID = env->GetMethodID(ServletUtilsClass, "<init>", "(II)V");
	if (!ServletUtilsMID) { goto init_error; };

	/* get the ServletUtils object */
	ServletUtilsObject = env->NewObject( ServletUtilsClass, ServletUtilsMID,
		(int)&tPIHTTP, (int)this);
	if (!ServletUtilsObject) { goto init_error; };

	/* get the Pi3ServletUtils.grabStackTrace method */
	jmethodID grabStackTraceMID;
	grabStackTraceMID = env->GetMethodID(ServletUtilsClass, "grabStackTrace",
		"(Ljava/lang/Throwable;)Ljava/lang/String;");
	if (!grabStackTraceMID) { goto init_error; };

	/* get the JspCompiler class */
	JspCompilerClass = env->FindClass("org/pi3/jsp/core/Pi3JspCompiler");
	if (!JspCompilerClass) { goto init_error; }

	/* get the JspCompiler constructor */
	jmethodID JspCompilerMID;
	JspCompilerMID = env->GetMethodID(JspCompilerClass, "<init>", "()V");
	if (!JspCompilerMID) { goto init_error; };

	/* get the JspCompiler object */
	JspCompilerObject = env->NewObject( JspCompilerClass, JspCompilerMID );
	if (!JspCompilerObject) { goto init_error; };

	/* get the Pi3JspCompiler.compile method */
	jmethodID compileMID;
	compileMID = env->GetMethodID(JspCompilerClass, "compile",
		"(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Z");
	if (!compileMID) { goto init_error; };

    /* invoke the Pi3JspCompiler.compile method */
	jboolean bRes;
	bRes = env->CallBooleanMethod(JspCompilerObject, compileMID,
		env->NewStringUTF( (const char *)sEncoding ),
		env->NewStringUTF( (const char *)sClassPath ),
		env->NewStringUTF( (const char *)sOutputPath ),
		env->NewStringUTF( pSourcePath ));

	if (!bRes)
		{
		Log( &tPIHTTP, "Pi3JspCompiler.compile() method failed for %s.",
			pSourcePath );
		}
	else
		{
	Log( &tPIHTTP, "Pi3JspCompiler.compile() method invoked for %s.",
		pSourcePath );
		goto done;
		}

init_error:
	Exception = env->ExceptionOccurred();
	if ( Exception && ServletUtilsObject && grabStackTraceMID )
		{
		Log( &tPIHTTP, "%s: Exception occured: ", GetLogName());

		/* invoke ServletUtils.grabStackTrace */
		jstring jmsg = (jstring)env->CallObjectMethod(ServletUtilsObject,
			grabStackTraceMID, Exception);
		if (jmsg) {
			jboolean isCopy;
			const char *msg = env->GetStringUTFChars(jmsg, &isCopy);
			Log( &tPIHTTP, msg);
			HTTPCore_logError(&tPIHTTP, msg);
			env->ReleaseStringUTFChars(jmsg, msg);
			}
		else
			{
			HTTPCore_logError( &tPIHTTP, "%s: \
An unknown exception occured during invokation of java compiler.", GetLogName());
			}
		env->ExceptionDescribe();
		env->ExceptionClear();
	} else {
		HTTPCore_logError(&tPIHTTP, "%s: Check your CLASSPATH settings.",
			GetLogName());
	};

	iRet = HTTPUtil_doHTTPError(&tPIHTTP, ST_INTERNALERROR);

done:
    /* We are done. */
	if (!PIHTTP_isChild( &tPIHTTP )) {
		if (!JVM()->DetachCurrentThread()/* !jvm->DestroyJavaVM() **/) {
			Log( &tPIHTTP, "%s: JVM detached.", GetLogName());
		} else {
			Log( &tPIHTTP, "%s: Error when trying to detach JVM.", GetLogName());
		};
	};
#endif // WIN32

#ifdef POSIX
	pid_t iPid;
	PIString sEnc("-encoding");

	iPid = ::fork();
	switch( iPid )
		{
		case -1:
			break;

		case 0:
			/* ---
			This is the child
			--- */
			enum { MAX_PARAMS = 7 };
			char *ppParams[MAX_PARAMS];

			sEnc.Concatenate(sEncoding);
			if (sClassPath.Len())
				{
				ppParams[0] = (char *)(const char *)sEnc;
				ppParams[1] = "-classpath";				
				ppParams[2] = (char *)(const char *)sClassPath;
				ppParams[3] = "-d";
				ppParams[4] = (char *)(const char *)sOutputPath;
				ppParams[5] = (char *)pSourcePath;
				ppParams[6] = NULL;
				}
			else
				{
				ppParams[0] = (char *)(const char *)sEnc;
				ppParams[1] = "-d";
				ppParams[2] = (char *)(const char *)sOutputPath;
				ppParams[3] = (char *)pSourcePath;
				ppParams[4] = NULL;
				};

			if ( ::execv((const char *)sCompiler, (char *const *)ppParams)==-1 )
				{
				perror( "Java Compiler, execv() failed." );
				fprintf( stderr, "Path is '%s'\n", pSourcePath );
				fflush( stderr );
				::_exit( 1 );
				};

			/* none of the following should happen */
			::_exit( -1 );
			assert( 0 );
			break;

		default:
			/* --- this is the parent --- */
			waitpid( iPid, &iRet, WUNTRACED );

			/* --- log debug information about the process --- */
			HTTPCore_logDebug( DBG_MED, "Process created!, \n\tPid %d", iPid );
		};
#endif // POSIX
	return iRet;
}

int Jsp::Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &pB/* tBuffer */ )
{
	if ( iPhase != PH_HANDLE ) return PIAPI_ERROR;
	/**
	* Assume we can't keep open connection due to
	* unknow Content-Length because of dynamic Jsp
	**/
	PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, 0, 0 );

	PIDB *pR = tPIHTTP.pResponseDB;
	const char *pServletName = 0;
	int iRet = PIAPI_ERROR;

	PIString sServletPath;
	PIString sClassFile;
	const char *Ext;
	PIFInfo *pFI;
	PIFInfo *pTmpFI;
	// File modification times
	PIPLATFORM_TIME tJsp;
	PIPLATFORM_TIME tJava;
	PIPLATFORM_TIME tClass;

	// Ascertain Jsp name from path and compile Jsp
	const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
		KEY_INT_PATH, 0 );
	if (!pPath) return PIAPI_ERROR;

	pFI = PIFInfo_new( pPath );
	
	if (!PIFInfo_exists( pFI ) || !PIFInfo_isRegular( pFI ))
		{
		PIFInfo_delete( pFI );
		return HTTPUtil_doHTTPError( &tPIHTTP,  ST_NOTFOUND );
		}

	Ext = PIFInfo_getExtension( pFI );

	if ( Ext && !PIUtil_stricmp(Ext, "jsp")) 
		{
		const char *pServletName = PIFInfo_getName( pFI );
		if ( pServletName )
			{ 
			sClassName = PIString( pServletName, strlen( pServletName ) - 4 );
			}
		else
			{
			goto init_error;
			};
		}
	else
		{
		goto init_error;
		};

	sServletPath = sSourcePath;
	sServletPath.Concatenate( PIPlatform_getDirectorySeparator() );
	sServletPath.Concatenate( sClassName );
	sServletPath.Concatenate( ".java" );

	PISync_lock( pMutex );
	tJsp = PIFInfo_getLastModified( pFI );

	pTmpFI = PIFInfo_new( (const char *)sServletPath );
	tJava = PIFInfo_getLastModified( pTmpFI );
	PIFInfo_delete( pTmpFI );

	// no java source or older than jsp source
	if ( tJsp > tJava )
		{
		JspFlexLexer scanner(pPath, sServletPath, sClassName, sExtendsName);
		iRet = scanner.exec();
		if (iRet) {
			iRet = PIAPI_ERROR;
			HTTPCore_logError( &tPIHTTP,  "Jsp parser failed for \
Input: %s, ServletPath: %s, ClassName: %s, Extends: %",
				pPath,
				(const char *)sServletPath,
				(const char *)sClassName,
				(const char *)sExtendsName
				) ;
			goto init_error;
		};
		pTmpFI = PIFInfo_new( sServletPath );
		tJava = PIFInfo_getLastModified( pTmpFI );
		PIFInfo_delete( pTmpFI );
		}

	PIFInfo_delete( pFI );

	sClassFile = sOutputPath;
	sClassFile.Concatenate( PIPlatform_getDirectorySeparator() );
	sClassFile.Concatenate( sClassName );
	sClassFile.Concatenate( ".class" );

	pFI = PIFInfo_new( (const char *)sClassFile );
	tClass = PIFInfo_getLastModified( pFI );

	// no java class or older than java source
	if ( tJava > tClass )
		{
		// call the java compiler
		iRet = Compile( tPIHTTP, sServletPath );

		if (iRet != PIAPI_COMPLETED ) {
			HTTPCore_logError( &tPIHTTP,
				"Java compiler failed for: \"%s\"", (const char *)sServletPath) ;
			goto init_error;
		};
	};

	iRet = Servlet::Handle( iPhase, tPIHTTP, pB );

init_error:
	PISync_unlock( pMutex );
	PIFInfo_delete( pFI );
	if (iRet) {
		return HTTPUtil_doHTTPError( &tPIHTTP,  ST_INTERNALERROR );
	}
	return iRet;
};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Jsp_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseServlet_constructor( pObj, PI_NEW( Jsp( pObj,
		iArgc, ppArgv, "Jsp handler" ) ) );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int Jsp_destructor( PIObject *pObj, int,
	const char *[] )
{
	delete (Servlet *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name JspClass
		Type LogicExtension
		Library Servlet
		OnClassLoad HandlerBaseServlet_onClassLoad
		Constructor Jsp_constructor
		CopyConstructor HandlerBaseServlet_copyConstructor
		Destructor Jsp_destructor
		Execute HandlerBaseServlet_execute
	</Class>

	<Object>
		Name Jsp
		Class JspClass
		Context ServletContainer
	</Object>

/*___+++CNF_END+++___*/
#endif


