/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/Pi3HttpServletResponse.cpp,v $
 * $Date: 2003/05/13 18:42:19 $
 *
 Description:
	Execute an servlet extension.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <jni.h>                                         /* where everything is defined */
#include "org_pi3_servlet_core_Pi3HttpServletResponse.h" /* automatically created JNI header */
#include "PiAPI.h"
#include "Pi2API.h"
#include "Pi3API.h"
#include "ServletHandler.h"

/*
 * Class:     org_pi3_servlet_core_Pi3HttpServletResponse
 * Method:    containsHeader
 * Signature: (Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_org_pi3_servlet_core_Pi3HttpServletResponse_containsHeader
(JNIEnv *env, jobject response, jstring hdr){

	jclass cls = env->GetObjectClass(response);
	jmethodID getPiHttp;
	getPiHttp = env->GetMethodID( cls, "getPiHttp", "()I");
	// Errorhandling!
	if (!getPiHttp) { return false; };
	env->ExceptionClear();

	int PiHttp = env->CallIntMethod(response, getPiHttp);
	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return false;
		};

	PIHTTP *pPIHTTP = (PIHTTP *)PiHttp;
	jboolean isCopy;
	const char *szBuf = env->GetStringUTFChars(hdr, &isCopy);
	void *test = PIDB_lookup( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, szBuf, 0 );
	return (test != NULL);
};

/*
 * Class:     org_pi3_servlet_core_Pi3HttpServletResponse
 * Method:    setContentLength
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_org_pi3_servlet_core_Pi3HttpServletResponse_setContentLength
(JNIEnv *env, jobject response, jint cl) {

	jclass cls = env->GetObjectClass(response);
	jmethodID getPiHttp;
	getPiHttp = env->GetMethodID( cls, "getPiHttp", "()I");
	// Errorhandling!
	if (!getPiHttp) { return; };
	env->ExceptionClear();

	int PiHttp = env->CallIntMethod(response, getPiHttp);
	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return;
		};

	PIHTTP *pPIHTTP = (PIHTTP *)PiHttp;

	/* --- content-length --- */
	enum { BUF_SIZE=63 };
	char szBuf[BUF_SIZE+1];
	sprintf( szBuf, "%lu", cl );
	PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, (void *)szBuf, 0 );
};

/*
 * Class:     org_pi3_servlet_core_Pi3HttpServletResponse
 * Method:    setContentType
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_pi3_servlet_core_Pi3HttpServletResponse_setContentType
  (JNIEnv *env, jobject response, jstring ct){

	jclass cls = env->GetObjectClass(response);
	jmethodID getPiHttp;
	getPiHttp = env->GetMethodID( cls, "getPiHttp", "()I");
	// Errorhandling!
	if (!getPiHttp) { return; };
	env->ExceptionClear();

	int PiHttp = env->CallIntMethod(response, getPiHttp);
	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return;
		};

	PIHTTP *pPIHTTP = (PIHTTP *)PiHttp;

	/* --- content-type --- */
	jboolean isCopy;
	const char *szBuf = env->GetStringUTFChars(ct, &isCopy);
	PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, (void *)szBuf, 0 );
	env->ReleaseStringUTFChars(ct, szBuf);
};

/*
 * Class:     org_pi3_servlet_core_Pi3HttpServletResponse
 * Method:    setStatus
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_org_pi3_servlet_core_Pi3HttpServletResponse_setStatus
(JNIEnv *env, jobject response, jint sc) {

	jclass cls = env->GetObjectClass(response);
	jmethodID getPiHttp;
	getPiHttp = env->GetMethodID( cls, "getPiHttp", "()I");
	// Errorhandling!
	if (!getPiHttp) { return; };
	env->ExceptionClear();

	int PiHttp = env->CallIntMethod(response, getPiHttp);
	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return;
		};

	PIHTTP *pPIHTTP = (PIHTTP *)PiHttp;
	pPIHTTP->iStatus = sc;
}