/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/org/pi3/servlet/core/Pi3HttpServletResponse.java,v $
 * $Date: 2003/05/13 18:42:21 $
 *
 Description:
	Implementation of both servlet API ServletResponse, HttpServletResponse interfaces.

\*____________________________________________________________________________*/

package org.pi3.servlet.core;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.BufferedOutputStream;
import java.lang.reflect.Array;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Enumeration;
import java.util.Locale;
import java.util.StringTokenizer;
import javax.servlet.ServletResponse;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpUtils;
import org.pi3.servlet.util.Pi3ServletUtils;
import org.pi3.servlet.util.Pi3BufferedOutputStream;

/**
 *
 * Implements both the <code>ServletResponse</code> and
 * <code>javax.servlet.http.HttpServletResponse</code>
 * interface to provide HTTP-specific functionality in
 * sending a response back to the client.
 *
 * Defines an HTTP servlet response that a servlet running on
 * Pi3Web sends to a client. This interface allows the servlet's
 * <code>service</code> method to access HTTP headers and return
 * data to its client.
 *
 * 
 * @author	Various
 * @version	$Version$
 *
 *
 */

final public class Pi3HttpServletResponse implements ServletResponse, HttpServletResponse {


    private int hobj;
    private int hsrv;
	private int sresult = 0;
	private boolean usedPW = false;
	private boolean usedOS = false;
    private Pi3ServletOutputStream sout;
    private PrintWriter pw;
    private Locale locale;
	Pi3ServletUtils su;
	HttpServletRequest req;

	/** This constructor is used by the native side */
    protected Pi3HttpServletResponse(int pihttp, int hservlet,
	  Pi3ServletUtils sutils, HttpServletRequest request) {
        hobj = pihttp;
        hsrv = hservlet;
		req = request;
        sout = new Pi3ServletOutputStream(hobj, hsrv);
        pw   = new PrintWriter(new Pi3BufferedOutputStream(sout));
		su   = sutils;
    }




    /** The method to get the PiHttp object from the native side */
    protected int getPiHttp() {
        return hobj;
    }




    /** The method to get the Servlet handler object from the native side */
    protected int getServletHandler() {
        return hsrv;
    }




    /** The method to get the result of calling native methods from the native side */
    protected int getServerResult() {
	    return sresult;
	}




    /** The method to get the availability to use the output ressources from the native side */
    protected boolean isCommitted() {
	    return sout.isCommitted();
	}




    private String toAbsoluteURL(String location) throws MalformedURLException {

        if (location == null) return null;
        URL url;
	    try {
            url = new URL(location);
        } catch (MalformedURLException e1) {
            String requrl = HttpUtils.getRequestURL(req).toString();
            try {
                url = new URL(new URL(requrl), location);
            } catch (MalformedURLException e2) {
                return (location);
            }
        }
        return (url.toString());
    }




    private boolean useURLRewriting(String location) {

        if (!req.isRequestedSessionIdValid()) return (false);

        if (req.isRequestedSessionIdFromCookie()) return (false);

        URL url;
        try {
            url = new URL(toAbsoluteURL(location));
        } catch (MalformedURLException e) {
            return (false);
        }

        if (!req.getScheme().equalsIgnoreCase(url.getProtocol())) return (false);
        if (!req.getServerName().equalsIgnoreCase(url.getHost())) return (false);
        if ((url.getPort() > 0) && (req.getServerPort() != url.getPort())) return (false);

/*        String contextPath = req.getContext().getPath();
        if ((contextPath != null) && (contextPath.length() > 0)) {
            String file = url.getFile();
            if ((file == null) || !file.startsWith(contextPath)) return (false);
        }
*/
        return (true);
    }



    private String rewriteURL(String location) {

        if (location == null) return null;

        Pi3HttpSession ses = (Pi3HttpSession)req.getSession(false);
        if ( ses == null) return null;

        String path = null;
        String query = null;
        int question = location.indexOf("?");
        if (question < 0) {
            path = location;
        } else {
            path = location.substring(0, question);
            query = location.substring(question);
        }
        StringBuffer sb = new StringBuffer(path);
        sb.append(";");
		sb.append(Constants.SESSION_PARAMETER_NAME);
        sb.append("=");
        sb.append(ses.getSID());
        if (query != null) sb.append(query);
        return (sb.toString());
    }



    /**
     * Returns the name of the charset used for
     * the MIME body sent in this response.
     *
     * <p>If no charset has been assigned, it is implicitly
     * set to <code>ISO-8859-1</code> (<code>Latin-1</code>).
     *
     * <p>See RFC 2047 (http://ds.internic.net/rfc/rfc2045.txt)
     * for more information about character encoding and MIME.
     *
     * @return		a <code>String</code> specifying the
     *			name of the charset, for
     *			example, <code>ISO-8859-1</code>
     *
     */
  
    public String getCharacterEncoding() {

        if (containsHeader("Content-Type")) {
            String ct = su.getResponseVariable("Content-Type");
			if (ct != null) {
                StringTokenizer st = new StringTokenizer(ct,";");
			    String to;
                while (st.hasMoreTokens()) {
                    to = st.nextToken();
				    if (to.trim().toUpperCase().startsWith("CHARSET=")) {
                        return to.substring(to.indexOf("=") + 1);
					}
				}
            }
        }
        return su.getDefaultCharset();
    }
    

    

    /**
     * Returns a {@link ServletOutputStream} suitable for writing binary 
     * data in the response. The servlet container does not encode the
     * binary data.  Either this method or {@link #getWriter} may 
     * be called to write the body, not both.
     *
     * @return				a {@link ServletOutputStream} for writing binary data	
     *
     * @exception IllegalStateException if the <code>getWriter</code> method
     * 					has been called on this response
     *
     * @exception IOException 		if an input or output exception occurred
     *
     * @see 				#getWriter
     *
     */

    public ServletOutputStream getOutputStream() throws IOException {
	    if (usedPW) {
            IllegalStateException e = new IllegalStateException();
			throw e;
		}
	    usedOS = true;
        return sout;
    }


    

    /**
     * Returns a <code>PrintWriter</code> object that 
     * can send character text to the client. 
     * The character encoding used is the one specified 
     * in the <code>charset=</code> property of the
     * {@link #setContentType} method, which must be called
     * <i>before</i> calling this method for the charset to take effect. 
     *
     * <p>If necessary, the MIME type of the response is 
     * modified to reflect the character encoding used.
     *
     * <p>Either this method or {@link #getOutputStream} may be called
     * to write the body, not both.
     *
     * 
     * @return 				a <code>PrintWriter</code> object that 
     *					can return character data to the client 
     *
     * @exception UnsupportedEncodingException  if the charset specified in
     *						<code>setContentType</code> cannot be
     *						used
     *
     * @exception IllegalStateException    	if the <code>getOutputStream</code>
     * 						method has already been called for this 
     *						response object
     *
     * @exception IOException   		if an input or output exception occurred
     *
     * @see 					#getOutputStream
     * @see 					#setContentType
     *
     */

    public PrintWriter getWriter() throws IOException {
	    if (usedOS) {
            IllegalStateException e = new IllegalStateException();
			throw e;
		}
	    usedPW = true;
        return pw;
    }
    
    
    
    
    /**
     * Sets the length of the content body in the response
     * In HTTP servlets, this method sets the HTTP Content-Length header.
     *
     *
     * @param len 	an integer specifying the length of the 
     * 			content being returned to the client; sets
     *			the Content-Length header
     *
     */

    public native void setContentLength(int len);
    
    


    /**
     * Sets the content type of the response being sent to
     * the client. The content type may include the type of character
     * encoding used, for example, <code>text/html; charset=ISO-8859-4</code>.
     *
     * <p>If obtaining a <code>PrintWriter</code>, this method should be 
     * called first.
     *
     *
     * @param type 	a <code>String</code> specifying the MIME 
     *			type of the content
     *
     * @see 		#getOutputStream
     * @see 		#getWriter
     *
     */

    public native void setContentType(String type);





    /***********************
     * HttpServletResponse *
     ***********************/

    /**
     * Adds the specified cookie to the response.  This method can be called
     * multiple times to set more than one cookie.
     *
     * @param cookie the Cookie to return to the client
     *
     */

    public void addCookie(Cookie cookie) {
	    su.addCookie(cookie);
	};




    /**
     * Returns a boolean indicating whether the named response header 
     * has already been set.
     * 
     * @param	name	the header name
     * @return		<code>true</code> if the named response header 
     *			has already been set; 
     * 			<code>false</code> otherwise
     */

    public native boolean containsHeader(String name);





    /**
     * Encodes the specified URL by including the session ID in it,
     * or, if encoding is not needed, returns the URL unchanged.
     * The implementation of this method includes the logic to
     * determine whether the session ID needs to be encoded in the URL.
     * For example, if the browser supports cookies, or session
     * tracking is turned off, URL encoding is unnecessary.
     * 
     * <p>For robust session tracking, all URLs emitted by a servlet 
     * should be run through this
     * method.  Otherwise, URL rewriting cannot be used with browsers
     * which do not support cookies.
     *
     * @param	url	the url to be encoded.
     * @return		the encoded URL if encoding is needed;
     * 			the unchanged URL otherwise.
     */

    public String encodeURL(String url) {

        if (useURLRewriting(url)) {
		    url = rewriteURL(url);
		};
		return url;
    }




    /**
     * Encodes the specified URL for use in the
     * <code>sendRedirect</code> method or, if encoding is not needed,
     * returns the URL unchanged.  The implementation of this method
     * includes the logic to determine whether the session ID
     * needs to be encoded in the URL.  Because the rules for making
     * this determination can differ from those used to decide whether to
     * encode a normal link, this method is seperate from the
     * <code>encodeURL</code> method.
     * 
     * <p>All URLs sent to the <code>HttpServletResponse.sendRedirect</code>
     * method should be run through this method.  Otherwise, URL
     * rewriting cannot be used with browsers which do not support
     * cookies.
     *
     * @param	url	the url to be encoded.
     * @return		the encoded URL if encoding is needed;
     * 			the unchanged URL otherwise.
     *
     * @see #sendRedirect
     * @see #encodeUrl
     */

    public String encodeRedirectURL(String url) {

        if (useURLRewriting(url)) {
		    url = rewriteURL(url);
		};
		return url;
    }




    /**
     * @deprecated	As of version 2.1, use encodeURL(String url) instead
     *
     * @param	url	the url to be encoded.
     * @return		the encoded URL if encoding is needed; 
     * 			the unchanged URL otherwise.
     */

    public String encodeUrl(String url) {
        return encodeURL(url);
    }



    
    /**
     * @deprecated	As of version 2.1, use 
     *			encodeRedirectURL(String url) instead
     *
     * @param	url	the url to be encoded.
     * @return		the encoded URL if encoding is needed; 
     * 			the unchanged URL otherwise.
     */

    public String encodeRedirectUrl(String url) {
        return encodeRedirectURL(url);
    }




    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.  The server generally creates the 
     * response to look like a normal server error page.
     *
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param	sc	the error status code
     * @param	msg	the descriptive message
     * @exception	IOException	If an input or output exception occurs
     * @exception	IllegalStateException	If the response was committed
     *						before this method call
     */

    public void sendError(int sc, String msg) throws IOException {
	    try {
		    sendError(sc);
		}
		catch (IOException e)
		{
		    throw e;
		};
	}




    /**
     * Sends an error response to the client using the specified
     * status.  The server generally creates the 
     * response to look like a normal server error page.
     *
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param	sc	the error status code
     * @exception	IOException	If an input or output exception occurs
     * @exception	IllegalStateException	If the response was committed
     */

    public void sendError(int sc) throws IOException {
	    if (sout.isCommitted()) {
			IllegalStateException e = new IllegalStateException();
			throw e;
		}
	    try {
			sresult = su.sendError(sc);
			sout.doCommit();
		}
		catch (IOException e)
		{
		    throw e;
		};
	}




    /**
     * Sends a temporary redirect response to the client using the
     * specified redirect location URL.  This method can accept relative URLs;
     * the servlet container will convert the relative URL to an absolute URL
     * before sending the response to the client.
     *
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param		location	the redirect location URL
     * @exception	IOException	If an input or output exception occurs
     * @exception	IllegalStateException	If the response was committed
     */

    public void sendRedirect(String location) throws IOException {
		if (sout.isCommitted()) {
			IllegalStateException e = new IllegalStateException();
			throw e;
		}
		su.setResponseVariable("Location", toAbsoluteURL(location));
		setStatus(SC_MOVED_PERMANENTLY);
		sout.sendHeaders();
	}


    
    /**
     * 
     * Sets a response header with the given name and
     * date-value.  The date is specified in terms of
     * milliseconds since the epoch.  If the header had already
     * been set, the new value overwrites the previous one.  The
     * <code>containsHeader</code> method can be used to test for the
     * presence of a header before setting its value.
     * 
     * @param	name	the name of the header to set
     * @param	value	the assigned date value
     * 
     * @see #containsHeader
     * @see #addDateHeader
     */

    public void setDateHeader(String name, long date) {
		su.setResponseVariable(name, su.getHttpDate(date));
    }



    
    /**
     * 
     * Adds a response header with the given name and
     * date-value.  The date is specified in terms of
     * milliseconds since the epoch.  This method allows response headers 
     * to have multiple values.
     * 
     * @param	name	the name of the header to set
     * @param	value	the additional date value
     * 
     * @see #setDateHeader
     */

    public void addDateHeader(String name, long date) {
        su.addResponseVariable(name, su.getHttpDate(date));
    }



    
    /**
     *
     * Sets a response header with the given name and value.
     * If the header had already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be
     * used to test for the presence of a header before setting its
     * value.
     * 
     * @param	name	the name of the header
     * @param	value	the header value
     *
     * @see #containsHeader
     * @see #addHeader
     */

    public void setHeader(String name, String value) {
        su.setResponseVariable(name, value);
    }



    
    /**
     * Adds a response header with the given name and value.
     * This method allows response headers to have multiple values.
     * 
     * @param	name	the name of the header
     * @param	value	the additional header value
     *
     * @see #setHeader
     */

    public void addHeader(String name, String value) {
        su.addResponseVariable(name, value);
    }




    /**
     * Sets a response header with the given name and
     * integer value.  If the header had already been set, the new value
     * overwrites the previous one.  The <code>containsHeader</code>
     * method can be used to test for the presence of a header before
     * setting its value.
     *
     * @param	name	the name of the header
     * @param	value	the assigned integer value
     *
     * @see #containsHeader
     * @see #addIntHeader
     */

    public void setIntHeader(String name, int value) {
		Integer i = new Integer(value);
        su.setResponseVariable(name, i.toString());
    }




    /**
     * Adds a response header with the given name and
     * integer value.  This method allows response headers to have multiple
     * values.
     *
     * @param	name	the name of the header
     * @param	value	the assigned integer value
     *
     * @see #setIntHeader
     */

    public void addIntHeader(String name, int value) {
		Integer i = new Integer(value);
        su.addResponseVariable(name, i.toString());

    }



    
    /**
     * Sets the status code for this response.  This method is used to
     * set the return status code when there is no error (for example,
     * for the status codes SC_OK or SC_MOVED_TEMPORARILY).  If there
     * is an error, the <code>sendError</code> method should be used
     * instead.
     *
     * @param	sc	the status code
     *
     * @see #sendError
     */

    public native void setStatus(int sc);



  
    /**
     * @deprecated As of version 2.1, due to ambiguous meaning of the 
     * message parameter. To set a status code 
     * use <code>setStatus(int)</code>, to send an error with a description
     * use <code>sendError(int, String)</code>.
     *
     * Sets the status code and message for this response.
     * 
     * @param	sc	the status code
     * @param	sm	the status message
     */

    public void setStatus(int sc, String sm) {
	    setStatus(sc);
    }

    static {
        System.loadLibrary("Pi3ServletJNI");
    }

}