/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/org/pi3/servlet/core/Pi3ServletInputStream.java,v $
 * $Date: 2003/05/13 18:42:21 $
 *
 Description:
	Implementation of servlet API ServletInputStream interface.

\*____________________________________________________________________________*/

package org.pi3.servlet.core;

import javax.servlet.ServletInputStream;
import java.lang.reflect.Array;
import java.io.IOException;
import java.lang.reflect.Array;


/**
 * Provides an input stream for reading binary data sent by the
 * client. You can access an object of this class by using the
 * {@link Pi3HttpServletRequest#getInputStream} method.
 *
 * <p>This class implements the <code>java.io.InputStream.read()</code>
 * abstract method.
 *
 * 
 * @author 	Holger Zimmermann [zimpel@pi3.org]
 * @version	$Revision: 1.5 $
 *
 * @see 	Pi3HttpServletRequest#getInputStream
 *
 */

final public class Pi3ServletInputStream extends ServletInputStream {

	private int hobj;
	private int hsrv;
	private boolean bCommitted;

	private native int internalReadA(byte[] cbuf, int off, int len) throws IOException;
	private native int internalReadB() throws IOException;




	/** This constructor is used by the native side */
	protected Pi3ServletInputStream(int pihttp, int hservlet) {
		hobj = pihttp;
		hsrv = hservlet;
	}




	/** The method to get the PiHttp object from the native side */
	protected int getPiHttp() {
		return hobj;
	}




	/** The method to get the Servlet handler object from the native side */
	protected int getServletEngine() {
		return hsrv;
	}

    public int read() throws IOException {
        if (available() == 0) return -1;
	    try {
	        return internalReadB();
		}
		catch (IOException e)
		{
		    throw e;
		};
	};




    public int read(byte[] cbuf, int off, int len) throws IOException {
        if (available() == 0) return -1;
	    try {
	        return internalReadA(cbuf, off, len);
		}
		catch (IOException e)
		{
		    throw e;
		};
	};




	public native int available();

    static {
        System.loadLibrary("Pi3ServletJNI");
    }
}
