/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/org/pi3/servlet/core/Pi3ServletOutputStream.java,v $
 * $Date: 2003/05/13 18:42:21 $
 *
 Description:
	Implementation of servlet API ServletOutputStream interface.

\*____________________________________________________________________________*/

package org.pi3.servlet.core;

import javax.servlet.ServletOutputStream;
import java.lang.reflect.Array;
import java.io.IOException;


/**
 * Provides an output stream for sending binary data to the
 * client. You can access an object of this class
 * by using the {@link Pi3HttpServletResponse#getOutputStream}
 * method.
 *
 * <p>This class implements the <code>java.io.OutputStream.write(int)
 * </code> abstract method.
 *
 * 
 * @author 	Holger Zimmermann [zimpel@pi3.org]
 * @version	$Revision: 1.5 $
 *
 * @see 	Pi3HttpServletResponse
 *
 */

final public class Pi3ServletOutputStream extends ServletOutputStream {

	private int hobj;
	private int hsrv;
	private boolean bCommitted;

    private native void internalWriteA(byte[] b, int off, int len) throws IOException;
	private native void internalWriteB(byte b) throws IOException;
    private native void internalSendHeaders() throws IOException;




	/** This constructor is used by the native side */
	protected Pi3ServletOutputStream(int pihttp, int hservlet) {
		hobj = pihttp;
		hsrv = hservlet;
	}




	/** The method to get the PiHttp object from the native side */
	protected int getPiHttp() {
		return hobj;
	}




	/** The method to get the Servlet handler object from the native side */
	protected int getServletHandler() {
		return hsrv;
	}




    public void close() throws IOException {
        try {
            flush();
        }
        catch (IOException e) {
            throw e;
        };
    }




    public boolean isCommitted() {
        return bCommitted;
    }




    public void doCommit() {
        bCommitted = true;
    }




    public void write(int b) throws IOException {
        if ( !bCommitted ) {
		    sendHeaders();
		    bCommitted = true;
        };
	    try {
	        internalWriteB((byte)b);
		}
		catch (IOException e)
		{
		    throw e;
		};
	}




    public native void flush() throws IOException;



    
	public void write(byte[] b)  throws IOException {
	    if ( !bCommitted ) {
		    sendHeaders();
		    bCommitted = true;
        };
	    try {
            internalWriteA(b, 0, b.length);
        }
        catch (IOException e) {
            throw e;
        };
	};




    public void write(byte[] b, int off, int len) throws IOException {

	    /* array to small, sorry! */
        if (b.length < off + len) throw new IOException();

        if ( !bCommitted ) {
		    sendHeaders();
		    bCommitted = true;
        };

        try {
            internalWriteA(b, off, len);
        }
        catch (IOException e) {
            throw e;
        };
	};




    public void sendHeaders() throws IOException {
        if ( !bCommitted ) {
		    internalSendHeaders();
		    bCommitted = true;
        } else {
		    IllegalStateException e = new IllegalStateException();
		    throw e;
		}
	}

    static {
        System.loadLibrary("Pi3ServletJNI");
    }
}
