/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Win32App/Pi3Srv32.cpp,v $
 * $Date: 2003/05/29 16:44:34 $
 *
 Description:
	Win32 service functions.
\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <windows.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>

#include "Pi2API.h"
#include "PIStrStr.h"
#include <windowsx.h>
#include "resource.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define PI3_SERVICE_NAME	"Pi3Service"
void ServiceMain( DWORD dwArgc, LPTSTR *lpszArgv );
void ServiceControlHandler( DWORD dwCtrlCode );
void StopService( LPTSTR lpszMsg );
HANDLE hTerminateEvent = 0;
SERVICE_STATUS_HANDLE sshStatusHandle;
SERVICE_STATUS ssStatus;
DWORD dwCheck;
int iRet = -1;
#define PI3_SERVICE_ERROR_FILE		"Pi3Error.log"
#define PI3_SERVICE_WARNING_FILE	"Pi3Warn.log"
#define PI3_LOCAL_PARAMFILE			"SOFTWARE\\Pi3\\Pi3Svc"
#define PI3_CONFIGPATH_KEY			"ConfigPath"
#define PI3_DEFAULT_CONFIGPATH		"..\\Conf\\Config.pi3"
// #define D(x) { MessageBox( NULL, (x), __FILE__, MB_OK ); };
#define D(x)

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void main()
{
	SERVICE_TABLE_ENTRY steDispatcherTable[] =
		{
		{ TEXT( PI3_SERVICE_NAME ), (LPSERVICE_MAIN_FUNCTION) ServiceMain },
		{ NULL, NULL }
		};

	if ( !StartServiceCtrlDispatcher( steDispatcherTable ) )
		{
		DWORD dwError = GetLastError();
		cerr << "Win32 error #" << dwError << endl;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void StopService( LPTSTR lpszMsg )
{
	LPTSTR lpStrings[2];
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	HANDLE hEventSource = RegisterEventSource( NULL, TEXT(PI3_SERVICE_NAME) );
	assert( hEventSource );
	sprintf( szBuf, "Pi3Service Error: %d", GetLastError() );
	lpStrings[0] = szBuf;
	lpStrings[1] = lpszMsg;
	if ( hEventSource )
		{
		if ( !ReportEvent(
			hEventSource,
			EVENTLOG_ERROR_TYPE,
			0,
			0,
			NULL,
			2,
			0,
			(const char **)lpStrings,
			NULL ) )
			{
			}
		else
			{
			};
		DeregisterEventSource( hEventSource );
		};	

	if ( hTerminateEvent )
		{ SetEvent( hTerminateEvent ); };
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
DWORD dwTheArgc;
LPTSTR *lpszTheArgv;
PIObject *___pObject;
PIDB *___pDB;
unsigned long Pi3Fn( unsigned long )
{
	if ( PILogic_execute( ___pObject, dwTheArgc, (const char **)lpszTheArgv ) )
		{
		remove( PI3_SERVICE_ERROR_FILE );
		PILog_writeDiagnostics( ___pDB, PILOG_ALL, PI3_SERVICE_ERROR_FILE );
		};
	
	return 0UL;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void App( PIDB *pDB )
{
	PIThread *pThread = 0;
	PIConfig *pConfigDB = 0;
	const char *pErr = 0;

	___pDB = pDB;

	sshStatusHandle = RegisterServiceCtrlHandler(
		TEXT( PI3_SERVICE_NAME ),
		(LPHANDLER_FUNCTION)ServiceControlHandler );

	if ( sshStatusHandle == (SERVICE_STATUS_HANDLE)NULL )
		{
		pErr = "RegisterServiceCtrlHandler() failed";
		goto DO_ERROR;
		};

	dwCheck = 1;
	ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	ssStatus.dwServiceSpecificExitCode = 0;

	/* --- notify service is pending, checkpoint 1 --- */
	ssStatus.dwControlsAccepted = 0;
	ssStatus.dwCurrentState = SERVICE_START_PENDING;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = dwCheck++;
	ssStatus.dwWaitHint = 2000;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); goto CLEANUP; };

	/* read in parameters from parameter file in current directory */
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	HKEY hKey;
	if ( RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,		// handle of an open key 
		PI3_LOCAL_PARAMFILE,	// address of subkey name		
		0,						// reserved
		KEY_EXECUTE |
		KEY_QUERY_VALUE,		// desired security access 
		&hKey			 		// address of buffer for opened handle 
	) == ERROR_SUCCESS )
		{
		DWORD dwSize = BUF_SIZE+1;
		DWORD dwType = REG_SZ;
		if ( RegQueryValueEx(
			hKey,				// handle of key to query 
			PI3_CONFIGPATH_KEY,	// address of name of value to query 
			NULL,				// reserved 
			&dwType,			// address of buffer for value type 
			(LPBYTE)&szBuf[0],	// address of data buffer 
			&dwSize				// address of data buffer size 
		) != ERROR_SUCCESS )
			{
			strcpy( szBuf, PI3_DEFAULT_CONFIGPATH );
			}
		}
	else
		{
		strcpy( szBuf, PI3_DEFAULT_CONFIGPATH );
		}

	if ( !*szBuf )
		{
		pErr = "Could not read configuration filename";
		goto DO_ERROR;
		};

	hTerminateEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
	if ( !hTerminateEvent ) 
		{ goto CLEANUP; };

	/* ---
	notify service is pending, checkpoint 2

	Loading the configuration files and setting up the objects
	and libraries can take a while.

	--- */
	ssStatus.dwControlsAccepted = 0;
	ssStatus.dwCurrentState = SERVICE_START_PENDING;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = dwCheck++;
	ssStatus.dwWaitHint = 10000;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); goto CLEANUP; };

	/* --- load up the configuration file --- */
	if ( PIConfig_loadConfigurationFile( pDB, szBuf, &pConfigDB ) )
		{
		remove( PI3_SERVICE_ERROR_FILE );
		PILog_writeDiagnostics( pDB, PILOG_ALL, PI3_SERVICE_ERROR_FILE );
		StopService("Could not load configuration file");
		goto CLEANUP;
		};

	ssStatus.dwControlsAccepted = 0;
	ssStatus.dwCurrentState = SERVICE_START_PENDING;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = dwCheck++;
	ssStatus.dwWaitHint = 10000;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); goto CLEANUP; };

	/* --- load up the Pi3 main object --- */
	___pObject = PIObject_load( pDB, pConfigDB, "Main", 0, 0 );
	if ( !___pObject )
		{
		remove( PI3_SERVICE_ERROR_FILE );
		PILog_writeDiagnostics( pDB, PILOG_ALL, PI3_SERVICE_ERROR_FILE );
		StopService("Could not load main Pi3 object");
		goto CLEANUP;
		};

	remove( PI3_SERVICE_WARNING_FILE );
	PILog_writeDiagnostics( pDB, PILOG_WARNING, PI3_SERVICE_WARNING_FILE );

	/* --- notify service is pending, checkpoint 3 --- */
	ssStatus.dwControlsAccepted = 0;
	ssStatus.dwCurrentState = SERVICE_START_PENDING;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = dwCheck++;
	ssStatus.dwWaitHint = 10000;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); goto CLEANUP; };

	/* --- start another thread to run the logic of the program --- */
	pThread = PIThread_new( 0, 0 );
	if ( !pThread )
		{ StopService( "Error creating service thread" ); goto CLEANUP; };

	if ( PIThread_begin( pThread, Pi3Fn, 0, PITHREAD_PRIORITY_MED, 0 ))
		{ StopService( "Error begin service thread" ); goto CLEANUP; };
	
	/* --- notify service is running --- */
	ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	ssStatus.dwCurrentState = SERVICE_RUNNING;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = 0;
	ssStatus.dwWaitHint = 0;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); goto CLEANUP; };

	/* --- wait for completion --- */
	WaitForSingleObject( hTerminateEvent, INFINITE );
	
	goto CLEANUP;

DO_ERROR:
	StopService( (char *)(pErr ? pErr : "") );

CLEANUP:
	if ( pThread )
		{ PIThread_delete( pThread ); };
	
	if ( ___pObject )
		{ PIObject_delete( ___pObject, 0, 0 ); };

	if ( hTerminateEvent )
		{ CloseHandle( hTerminateEvent ); };
	
	if ( sshStatusHandle )
		{
		ssStatus.dwControlsAccepted = 0;
		ssStatus.dwCurrentState = SERVICE_STOPPED;
		ssStatus.dwWin32ExitCode = GetLastError();
		ssStatus.dwCheckPoint = 0;
		ssStatus.dwWaitHint = 0;
		if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
			{ StopService( "SetServiceStatus" ); };
		};

	iRet = (int)ssStatus.dwWin32ExitCode;
	return;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static HWND __hWndDlgOK = 0;
static HWND __hWndDlgPasswd = 0;
struct passphrase {
	char **buf;
	int bufsize;
} * __passphrase;

static BOOL CALLBACK fnPassDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
{
	switch (uMsg)
		{
		case WM_INITDIALOG:
			__hWndDlgOK = GetDlgItem( hDlg, IDOK );
			__hWndDlgPasswd = GetDlgItem( hDlg, IDC_PASSWORD );
			__passphrase = (passphrase *)lParam;
			assert( __hWndDlgOK && __hWndDlgPasswd );
			SendMessage( __hWndDlgPasswd, EM_SETLIMITTEXT, __passphrase->bufsize, 0 );
			EnableWindow( __hWndDlgOK, FALSE );
			break;

		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
				{
				case IDOK:
					{
					/*
					** Get content of edit field
					*/
					SendMessage( __hWndDlgPasswd, WM_GETTEXT, __passphrase->bufsize,
						(LPARAM)*(__passphrase->buf) );
					EndDialog( hDlg, 1 );
					break;
					}

				case IDCANCEL:
					EndDialog( hDlg, 0 );
					break;

				case IDC_PASSWORD:
					if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE )
						{
						if ( SendMessage( GET_WM_COMMAND_HWND(wParam, lParam ), EM_LINELENGTH,
							0, 0 )>0 )
							{ EnableWindow( __hWndDlgOK, TRUE ); }
						else
							{ EnableWindow( __hWndDlgOK, FALSE ); }
						}
					else
						return (FALSE);
					break;

				default:;
				};

		default:
			return (FALSE);
		};

	return (TRUE);
}

int fnPassPhraseDlg(char *buf, int bufsize)
{
	passphrase pphrase;
	pphrase.buf = &buf;
	pphrase.bufsize = bufsize;
	if (DialogBoxParam(
		::GetModuleHandle( NULL ),
		MAKEINTRESOURCE(IDD_PASSPHRASE),
		NULL,
		(DLGPROC)fnPassDialogProc,
		(LPARAM)&pphrase
	) != -1 )
		{
		/* --- notify service is pending, increment checkpoint --- */
		ssStatus.dwControlsAccepted = 0;
		ssStatus.dwCurrentState = SERVICE_START_PENDING;
		ssStatus.dwWin32ExitCode = NO_ERROR;
		ssStatus.dwCheckPoint = dwCheck++;
		ssStatus.dwWaitHint = 10000;
		if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
			{ StopService( "SetServiceStatus" ); };
		return strlen(buf);
		}
	else
		{ StopService( "DialogBoxParam" ); };

		return -1;
}

typedef void REGPWDDLG(void *pPasswdDlg);

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void RegisterPassPhraseDialog()
{
	HINSTANCE hSSL = LoadLibrary("SSL.dll");
	if (hSSL)
	{
	REGPWDDLG *regPwdDlg = (REGPWDDLG *)GetProcAddress(hSSL, "SSL_register_PassphraseDlg");
	if (regPwdDlg!=NULL) regPwdDlg(&fnPassPhraseDlg);
	}

}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ServiceMain( DWORD dwArgc, LPTSTR *lpszArgv )
{
	/* --- save commandline parameters --- */
	dwTheArgc = dwArgc;
	lpszTheArgv = lpszArgv;

#if 0
just experimenting....
	/* --- set the error mode so as not to create messageboxes --- */
	::SetErrorMode( SEM_FAILCRITICALERRORS );
#endif

	/* ---
	set current directory to be directory where binary is.
	This is happening at this point because several debugging files will
	be created in this directory.
	--- */
	const char *pErr = 0;
	int iLen = 0;
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	if ( ::GetModuleFileName( ::GetModuleHandle( NULL ), szBuf, BUF_SIZE )<=0 )
		{
		pErr = "Could not get executable path";
		goto DO_ERROR;
		};
	szBuf[BUF_SIZE] = '\0';
	for ( iLen = strlen( szBuf )-1; iLen>=0; iLen-- )
		{
		if ( szBuf[iLen]=='/' || szBuf[iLen]=='\\' || szBuf[iLen]==':' )
			{ break; };
		};
	iLen++; szBuf[iLen]='\0';
	if ( !::SetCurrentDirectory( szBuf ) )
		{
		pErr = "Could not set current directory";
		goto DO_ERROR;
		};

	if ( !dwArgc || !lpszArgv )
		{ goto DO_ERROR; };

	RegisterPassPhraseDialog();

	if ( PIProgram_enter( lpszArgv[0], PIPROGRAM_VERSION_1_0, App ) )
		{ return; };

DO_ERROR:
	/*
	** NOTE:
	** possible do something with pErr;
	*/
	return;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ServiceControlHandler( DWORD dwControlCode )
{
	DWORD dwState = SERVICE_RUNNING;

	switch( dwControlCode )
		{
		case SERVICE_CONTROL_STOP:
			/* --- notify service is stopping --- */
			ssStatus.dwControlsAccepted = 0;
			ssStatus.dwCurrentState = SERVICE_STOP_PENDING;
			ssStatus.dwWin32ExitCode = NO_ERROR;
			ssStatus.dwCheckPoint = 1;
			ssStatus.dwWaitHint = 4000;
			if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
				{ StopService( "SetServiceStatus" ); };
			SetEvent( hTerminateEvent );
			break;

		case SERVICE_CONTROL_PAUSE:
		case SERVICE_CONTROL_CONTINUE:
		case SERVICE_CONTROL_INTERROGATE:
		default:
			break;
		};

	ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	ssStatus.dwCurrentState = dwState;
	ssStatus.dwWin32ExitCode = NO_ERROR;
	ssStatus.dwCheckPoint = 0;
	ssStatus.dwWaitHint = 0;
	if ( !SetServiceStatus( sshStatusHandle, &ssStatus ) )
		{ StopService( "SetServiceStatus" ); };
}
