/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/XSLT/XSLT.cpp,v $
 * $Date: 2003/05/13 18:42:22 $
 *
 Description:
	Execute an XSLT transformation on a XML (HTML or Docbook SGML) document
	and send back the resulting document to the client.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "DeQuote.h"
#include "PiAPI.h"
#include "StrToken.h"

#include <libxslt/libxslt.h>
#include <libexslt/exslt.h>
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxml/xmlIO.h>
#include <libxml/DOCBparser.h>
#include <libxml/HTMLparser.h>
#include <libxml/xinclude.h>
#include <libxml/catalog.h>
#include <libxml/parserInternals.h>
#include <libxml/uri.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include <libxslt/imports.h>

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define MAX_PARAMETERS 64
#define KEY_CONF_DEFAULTSTYLESHEET	"DefaultStylesheet"
#define KEY_CONF_CONTENTTYPE		"DefaultContentType"
#define KEY_CONF_OPTIONS			"Options"
#define KEY_CONF_MAXDEPTH			"MaxDepth"
#define KEY_CONF_VARIABLE			"Variable"
#define KEY_CONF_EXTRAHDRPREFIX		"ExtraHeadersPrefix"
#define KEY_CONF_EXTRAHEADERSIGNORE	"ExtraHeadersIgnore"
#define KEY_CONF_EXTRAHEADERS		"ExtraHeaders"
#define VALUE_NO					"No"
#define VALUE_YES					"Yes"


/*____________________________________________________________________________*\
 *
 Flags and Options
\*____________________________________________________________________________*/
#define FLG_NONE		0x00000000
#define FLG_NODTD		0x00000001
#define FLG_VALIDATE	0x00000002
#define FLG_NONET		0x00000004
#define FLG_WARNNET		0x00000008
#define FLG_VERBOSE		0x00000010
#define FLG_XINCLUDE	0x00000020
#define FLG_CATALOGS	0x00000040
#define FLG_HTML		0x00000080
#define FLG_DOCBOOK		0x00000100
#define FLG_TIMING		0x00000200
#define FLG_PROFILE		0x00000400
#define FLG_DEBUG		0x00000800
	
/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
			Map these flags to names

*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
struct {
	const char *pName;
	int iFlag;
} aFlagMap[] =
	{
	{ "SkipDtdLoading",	FLG_NODTD },
	{ "Validate",		FLG_VALIDATE },
	{ "NoNetEntities",	FLG_NONET },
	{ "WarnNetEntities",FLG_WARNNET },
	{ "Verbose",		FLG_VERBOSE },
	{ "XInclude",		FLG_XINCLUDE },
	{ "SgmlCatalogs",	FLG_CATALOGS },
	{ "Html",			FLG_HTML },
	{ "Docbook",		FLG_DOCBOOK },
	{ "Timing",			FLG_TIMING },
	{ "Profile",		FLG_PROFILE },
	{ "Debug",			FLG_DEBUG },

	/* --- leave this last always --- */
	{ 0, 0 }
	};
/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	XSLT

Description:
	The XSLT handler provides XSLT transformations based on libxml2
	and libxslt. Further information about these libraries is available
	at <a href="http://xmlsoft.org">http://xmlsoft.org</a>. The input
	may be either an XML, HTML or Docbook SGML document. An XSL stylsesheet
	is applied to the paresd XML document and the result is sent back to
	the client.	POST requests are also supported.

Options:

<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>DefaultStylesheet
<TD>-
<TD>&lt;path name or URL&gt;
<TD>A valid path name or URL to be used to load the XSLT stylesheet
<TD>DefaultStylesheet "./Default.xsl"

<TR>
<TD>DefaultContentType
<TD>-
<TD>&lt;a mime type string&gt;
<TD>A valid mime type to overwrite XSLT engine selection
<TD>DefaultContentType "application/rtf"

<TR>
<TD>MaxDepth
<TD>500
<TD>&lt;a number&gt;
<TD>The maximum depth of template stack
<TD>Example

<TR>
<TD>Variable
<TD>-
<TD>&lt;a Pi3Expression&gt;
<TD>A variable definition
<TD>Variable "GATEWAY_INTERFACE=CGI/1.1"

<TR>
<TD>ExtraHeaders
<TD>Yes
<TD>Yes|No
<TD>Indicates if extra headers are considered
<TD>ExtraHeaders Yes

<TR>
<TD>ExtraHeadersPrefix
<TD>-
<TD>&lt;a string&gt;
<TD>Used as prefix of each extra header
<TD>ExtraHeadersPrefix &quot;HTTP_&quot;

<TR>
<TD>ExtraHeadersIgnore
<TD>-
<TD>&lt;Space delimited strings&gt;
<TD>List of unconsidered extra headers
<TD>ExtraHeadersIgnore &quot;Content-Type Content-Length&quot;

<TR>
<TD>Options
<TD>-
<TD>Options flags as described below
<TD>Options which effect behaviour
<TD>Options "NoNetEntities | Verbose"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>DefaultStylesheet</H5>
This parameter is used to apply the given stylesheet instead the stylesheet
ignoring any stylesheet in the source file. A relative path is treated
from the current working directory, which is the directory from where the
server has been started. This is different to other external entity paths,
e.g. dtd or styleshet paths in an xml file.

<H5>DefaultContentType</H5>
This parameter is used to set the mime type used in the server response to
a given value. Otherwise the selection is done by the handler itself
supported by the XSLT engine.

<H5>MaxDepth</H5>
Adjust the maximum depth of the template stack before libxslt concludes it
is in an infinite loop. The default is 500.

<H5>Variable</H5>
Specifies a server variable expression to be set as a global stylesheet parameter.
Could be set multiple times in one configuration object up to 32.

<H5>ExtraHeaders</H5>
This configuration key defines, if extra headers are considered by
the processing of this handler as global stylesheet parameters.

<H5>ExtraHeadersPrefix</H5>
This configuration value will be the trailing string of each extra
request header.

<H5>ExtraHeadersIgnore</H5>
The headers in this list are not treated as extra headers.

<H5>Options</H5>
Specifies options which effect the behaviour of the XSLT transformations,
many available facilities are optional for performance reasons, the table
below lists the available option flags with their meanings.
<CENTER>
<TABLE BORDER=1>
<TH>Option
<TH>Meaning
<TR>
<TD>SkipDtdLoading
<TD>Skip loading the document's DTD.

<TR>
<TD>Validate
<TD>Validate the document against it's DTD.

<TR>
<TD>NoNetEntities
<TD>Do not use the Internet to fetch DTD's or entities.

<TR>
<TD>WarnNetEntities
<TD>Output notification when DTD's or entities are fetched over the Internet.
Not implemented yet.

<TR>
<TD>Verbose
<TD>Output each step taken by XSLT in processing the stylesheet and the document
in the server debug log. Debug logging must be enabled in addition.

<TR>
<TD>XInclude
<TD>Process the input document using the Xinclude specification. More details on
this can be found in the Xinclude specification: http://www.w3.org/TR/xinclude/

<TR>
<TD>SgmlCatalogs
<TD>Use catalogs to resolve the location of external entities. This speeds DTD resolution.
By having a catalog file point to a local version of the DTD, XSLT does not have to use
the Internet to fetch the DTD. XSLT uses the catalog identified by the environmental
variable SGML_CATALOG_FILES.

<TR>
<TD>Html
<TD>The input document is an HTML file. 

<TR>
<TD>Docbook
<TD>The input document is DocBook SGML. This is not to be used for DocBook XML documents.

<TR>
<TD>Timing
<TD>Display the time used for parsing the stylesheet, parsing the document and applying
the stylesheet and saving the result. Displayed in milliseconds. Not implemented yet.

<TR>
<TD>Profile
<TD>Output profiling information detailing the amount of time spent in each part of the
stylesheet. This is useful in optimizing stylesheet performance. Not implemented yet.

<TR>
<TD>Debug
<TD>Output an XML tree of the transformed document for debugging purposes on STDOUT.

</TABLE>
</CENTER>
Multiple options may be listed together by seperating them using the pipe
('|') character. Whitespace will be stripped from the option pattern. Options
are matched without regard to case sensitivity. Unknown options will raise
a configuration error. The option directive may be repeated multiple times.

Notes:
	Some of the configuration options are not reentrant. None of the options are applied,
	when the DefaultStylesheet is loaded at server startup. Embedded stylesheets work
	only with enabled option "Validate", which is probably a bug in libxml2 (2.4.20).
	The libxml 2 has been compiled _REENTRANT and with some minor changes. Thus other
	binary distribution of libxml2 might not work with Pi3Web.
	The Content-Type for the output is automatic detected, if possible. Otherwise the
	parameter "DefaultContentType" or the default Content-Type of the server is used.
	For POST requests the received form data must be of Content-Type "text/plain"
	or "application/x-www-form-urlencoded". The received form is converted into a
	generic XML format, which follows this simple DTD -

<PRE>
&lt;!DOCTYPE form [
	&lt;!ELEMENT form (field+)&gt;
	&lt;!ELEMENT field #PCDATA&gt;
	&lt;!ATTLIST field name NMTOKEN #REQUIRED&gt;
	]
&gt;
</PRE>


Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED, PIAPI_ERROR, or INT_REDIRECT accordingly
	to the status returned by the handler.

Example:
	<PRE>
	&lt;Object&gt;
		Name XSLT
		Class XSLTClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle Condition="&cmp(&dblookup(response,string,ObjectMap,XSLT),XSLT)" \
			XSLT
		...
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

#define STATE_NEW			0
#define STATE_STARTSTREAM	1
#define STATE_ENDSTREAM		2
#define STATE_END			3

#define START_FORM		"<?xml version=\"1.0\"?>\r\n<form>\r\n"
#define END_FORM		"</form>\r\n"
#define START_FIELD_A	"\t<field name=\""
#define START_FIELD_B	"\">"
#define END_FIELD		"</field>\r\n"

typedef struct {
	PIIOBuffer *iobuf;
	long clength;
	long read;
	int state;
} in_ctx;

/*____________________________________________________________________________*\
 *
 Description: Callback hook to read POST form of type text/plain
\*____________________________________________________________________________*/
static int xmlReadPlainCB (void *context, char *buffer, int len) {
	in_ctx *ctx = (in_ctx *)context;

	// new document
	if (ctx->state == STATE_NEW) {
		strcpy(buffer, START_FORM);
		ctx->state = STATE_STARTSTREAM;
		return strlen(START_FORM);
	};

	// end document
	if (ctx->state == STATE_END) {
		return 0;
	};

	// end stream
	if (ctx->state == STATE_ENDSTREAM) {
		strcpy(buffer, END_FORM);
		ctx->state = STATE_END;
		return strlen(END_FORM);
	}

	int iRet = len < ctx->clength - ctx->read ? len : ctx->clength - ctx->read;
	const char *inbuf = PIIOBuffer_read( ctx->iobuf, &iRet );

	if (iRet) {
		char *outbuf = buffer;
		char *pValue = (char *)inbuf;
		pValue[iRet] = '\0';

		StringTokenizer tTokens( inbuf, "\r", iRet );
		for(int i=0; i<tTokens.NumTokens(); i++)
			{
			const char *pToken = tTokens.GetToken( i );
			if (*pToken == '\n') pToken++;

			pValue = strchr(pToken, '=');
			if (pValue)
				{
				pValue++;

				// not in 1st form field!
				if (ctx->read||outbuf-buffer) {
					strcpy(outbuf, END_FIELD);
					outbuf += strlen(END_FIELD);
				}

				strcpy(outbuf, START_FIELD_A);
				outbuf += strlen(START_FIELD_A);

				strncpy(outbuf, pToken, (pValue-pToken)-1);
				outbuf += (pValue-pToken)-1;

				strcpy(outbuf, START_FIELD_B);
				outbuf += strlen(START_FIELD_B);

				strcpy(outbuf, pValue);
				outbuf += strlen(pValue);
				}
			else
				{
				strcpy(outbuf, pToken);
				outbuf += strlen(pToken);
				}
			}

		ctx->read += iRet;

		// last form field
		if (ctx->read == ctx->clength) {
			strcpy(outbuf, END_FIELD);
			outbuf += strlen(END_FIELD);
			ctx->state = STATE_ENDSTREAM;
		}
		return outbuf - buffer;
	};
	return 0;
}

/*____________________________________________________________________________*\
 *
 Description: Callback hook to read POST form of type
              application/x-www-form-urlencoded
\*____________________________________________________________________________*/
static int xmlReadFormCB (void *context, char *buffer, int len) {
	in_ctx *ctx = (in_ctx *)context;

	// new document
	if (ctx->state == STATE_NEW) {
		strcpy(buffer, START_FORM);
		ctx->state = STATE_STARTSTREAM;
		return strlen(START_FORM);
	};

	// end document
	if (ctx->state == STATE_END) {
		return 0;
	};

	// end stream
	if (ctx->state == STATE_ENDSTREAM) {
		strcpy(buffer, END_FORM);
		ctx->state = STATE_END;
		return strlen(END_FORM);
	}

	int iRet = len < ctx->clength - ctx->read ? len : ctx->clength - ctx->read;
	const char *inbuf = PIIOBuffer_read( ctx->iobuf, &iRet );

	if (iRet) {
		char *outbuf = buffer;

		// 1st form field
		if (!ctx->read) {
			strcpy(outbuf, START_FIELD_A);
			outbuf += strlen(START_FIELD_A);
		}

		for (int i=0;i<iRet;i++) {
			if ( inbuf[i] == '=' ) {
				strcpy(outbuf, START_FIELD_B);
				outbuf += strlen(START_FIELD_B);
				continue;
			};

			// next form field
			if ( inbuf[i] == '&') {
				strcpy(outbuf, END_FIELD);
				outbuf += strlen(END_FIELD);
				strcpy(outbuf, START_FIELD_A);
				outbuf += strlen(START_FIELD_A);
			} else {
				if (inbuf[i] == '+')
					{
					*(outbuf++) = ' ';
					}
				else if (inbuf[i] == '%')
					{
					if ( ((i+1)<iRet) && inbuf[i+1]=='%' )
						{ *(outbuf++) = inbuf[i]; i++; }
					else if ( (i+2)<iRet )
						{
						char szBuf[3];
						strncpy( szBuf, &(inbuf[i+1]), 2 );
						szBuf[2]='\0';
						*(outbuf++) = (char)strtol( szBuf, (char **)0, 16 ); 
						i+=2;
						};
					}
				else
					{
					*(outbuf++) = inbuf[i];
					}
			}
		}

		ctx->read += iRet;

		// last form field
		if (ctx->read == ctx->clength) {
			strcpy(outbuf, END_FIELD);
			outbuf += strlen(END_FIELD);
			ctx->state = STATE_ENDSTREAM;
		}
		return outbuf - buffer;
	};
	return 0;
}

/*____________________________________________________________________________*\
 *
 Description: Parser function for POST'ed input of type 
              application/x-www-form-urlencoded
\*____________________________________________________________________________*/
xmlDocPtr
xmlIOParseFormEntity( void *ioctx, xmlCharEncoding enc ) {
    xmlDocPtr ret;
    xmlParserCtxtPtr ctxt =
		xmlCreateIOParserCtxt(
			NULL,
			NULL,
			xmlReadFormCB,
			NULL,
			ioctx,
			enc
		);

    if (!ctxt) return(NULL);

    xmlParseExtParsedEnt(ctxt);

    if (ctxt->wellFormed)
	ret = ctxt->myDoc;
    else {
        ret = NULL;
        xmlFreeDoc(ctxt->myDoc);
        ctxt->myDoc = NULL;
    }

    xmlFreeParserCtxt(ctxt);
    return(ret);
}

/*____________________________________________________________________________*\
 *
 Description: Parser function for POST'ed input of type text/plain
\*____________________________________________________________________________*/
xmlDocPtr
xmlIOParsePlainEntity( void *ioctx, xmlCharEncoding enc ) {
    xmlDocPtr ret;
    xmlParserCtxtPtr ctxt =
		xmlCreateIOParserCtxt(
			NULL,
			NULL,
			xmlReadPlainCB,
			NULL,
			ioctx,
			enc
		);

    if (!ctxt) return(NULL);

    xmlParseExtParsedEnt(ctxt);

    if (ctxt->wellFormed)
	ret = ctxt->myDoc;
    else {
        ret = NULL;
        xmlFreeDoc(ctxt->myDoc);
        ctxt->myDoc = NULL;
    }

    xmlFreeParserCtxt(ctxt);
    return(ret);
}

/*____________________________________________________________________________*\
 *
 Description: Parser function for HTML file input
\*____________________________________________________________________________*/
xmlDocPtr
xmlPi3ParseHtmlFile( const char *filename, PIHTTP *pPIHTTP) {
    xmlDocPtr ret;
	docbParserCtxtPtr c =
		htmlCreateFileParserCtxt(filename, (const char *)XML_CHAR_ENCODING_NONE);

    if (c == NULL) return NULL;

	c->_private = pPIHTTP;

    htmlParseDocument(c);

    if (c->wellFormed) {
		ret = c->myDoc;
		ret->_private = (void *)c->valid;
    } else {
        ret = NULL;
        xmlFreeDoc(c->myDoc);
        c->myDoc = NULL;
    }    

    xmlFreeParserCtxt(c);
    return ret;
}

/*____________________________________________________________________________*\
 *
 Description: Parser function for Docbook SGML file input
\*____________________________________________________________________________*/
xmlDocPtr
xmlPi3ParseDocbookFile( const char *filename, PIHTTP *pPIHTTP) {
    xmlDocPtr ret;
	docbParserCtxtPtr c =
		docbCreateFileParserCtxt(filename, (const char *)XML_CHAR_ENCODING_NONE);

    if (c == NULL) return NULL;

	c->_private = pPIHTTP;

    docbParseDocument(c);

    if (c->wellFormed) {
		ret = c->myDoc;
		ret->_private = (void *)c->valid;
    } else {
        ret = NULL;
        xmlFreeDoc(c->myDoc);
        c->myDoc = NULL;
    }    

    xmlFreeParserCtxt(c);
    return ret;
}

/*____________________________________________________________________________*\
 *
 Description: Parser function for XML file input
\*____________________________________________________________________________*/
xmlDocPtr
xmlPi3ParseXmlFile( const char *filename, PIHTTP *pPIHTTP) {
    xmlDocPtr ret;
	xmlParserCtxtPtr c =
		xmlCreateFileParserCtxt(filename);

    if (c == NULL) return NULL;

	c->_private = pPIHTTP;

    xmlParseDocument(c);

    if (c->wellFormed) {
		ret = c->myDoc;
		ret->_private = (void *)c->valid;
    } else {
        ret = NULL;
        xmlFreeDoc(c->myDoc);
        c->myDoc = NULL;
    }    

    xmlFreeParserCtxt(c);
    return ret;
}

/*____________________________________________________________________________*\
 *
 Description: Callback hook to write output data to client
\*____________________________________________________________________________*/
static int xmlWriteCB (void * context, const char * buffer, int len) {
    return PIIOBuffer_write((PIIOBuffer *)context, buffer, len,
		(int)PIIOBuffer_getUserData((PIIOBuffer *)context) );
}

/*____________________________________________________________________________*\
 *
 Description: Callback hook to flush output data to client
\*____________________________________________________________________________*/
static int xmlOutCloseCB (void * context) {
    return PIIOBuffer_flush((PIIOBuffer *)context);
}

/*____________________________________________________________________________*\
 *
 Description: Callback hook to write error to client
\*____________________________________________________________________________*/
static void xmlErrorCB (void * context, const char * msg, ...) {
	enum { BUF_SIZE=1024 };
	char szBuf[BUF_SIZE+1];
	va_list tvList;
	va_start( tvList, msg );
#if WIN32
	_vsnprintf( &(szBuf[0]), BUF_SIZE, msg, tvList );
#else
	vsnprintf( &(szBuf[0]), BUF_SIZE, msg, tvList );
#endif
	va_end( tvList );
	if (szBuf[strlen(&(szBuf[0]))-1] == 10)
		szBuf[strlen(&(szBuf[0]))-1] = 0;
	HTTPCore_logError( (PIHTTP *)context, &(szBuf[0]) );
}

/*____________________________________________________________________________*\
 *
 Description: Callback hook to write data to debug log
\*____________________________________________________________________________*/
static void xmlDebugCB (void * context, const char * msg, ...) {
	enum { BUF_SIZE=1024 };
	char szBuf[BUF_SIZE+1];
	va_list tvList;
	va_start( tvList, msg );
#if WIN32
	_vsnprintf( &(szBuf[0]), BUF_SIZE, msg, tvList );
#else
	vsnprintf( &(szBuf[0]), BUF_SIZE, msg, tvList );
#endif
	va_end( tvList );
	if (szBuf[strlen(&(szBuf[0]))-1] == 10)
		szBuf[strlen(&(szBuf[0]))-1] = 0;
	HTTPCore_logDebug( DBG_MED, &(szBuf[0]) );
}

/*____________________________________________________________________________*\
 *
 Description: Helper function to create a URI from given path using escaping
\*____________________________________________________________________________*/
static PIString getFileURI(const char *pPath)  {
	if (!pPath) return NULL;
	PIString sURI = pPath;
	char *pPtr = (char *)(const char *)sURI;
	while (strchr(pPtr, '\\'))
		{
		pPtr[strchr(pPtr,'\\')-pPtr] = '/';
		}

	xmlChar *pURI = xmlURIEscape((const xmlChar *)pPtr);
	sURI = "file://";
	if (pPath[0] != '/') sURI.Concatenate('/');
	sURI.Concatenate((char *)pURI);
	return sURI; 
}

/*____________________________________________________________________________*\
 *
 Description: Static storage for origin registered entity loader function
\*____________________________________________________________________________*/
static xmlExternalEntityLoader OriginEntityLoader = 0;

/*____________________________________________________________________________*\
 *
 Description: Customized entity loader, which takes care, that external entities
 URI's don't leave the webroot of the current HTTP connection
\*____________________________________________________________________________*/
static xmlParserInputPtr xmlEntityLoader(
		const char *URL,
		const char *ID,
		xmlParserCtxtPtr context
		)
{
	if (!URL) return NULL;

	if (!strncmp(URL, "http://", 7))
		return OriginEntityLoader( URL, ID, context );

	if (!strncmp(URL, "ftp://", 6))
		return OriginEntityLoader( URL, ID, context );

	PIHTTP *pPIHTTP = (PIHTTP *)context->_private;

	if ( pPIHTTP ) {
		/* --- make sub request context --- */
		PIHTTP *pChildHTTP = PIHTTP_newChild( pPIHTTP );

		// Someone has mapped the URL already
		if (!strncmp(URL, "file://", 7)) {
			PIDB_replace( pChildHTTP->pResponseDB, PIDBTYPE_STRING,
				KEY_INT_PATH, (void *)"/", 0 );
		} else {
			PIDB_replace( pChildHTTP->pResponseDB, PIDBTYPE_STRING,
				KEY_INT_PATH, (void *)URL, 0 );
		}

		/* --- dispatch the sub request across the mapping phase --- */
		int iRet = HTTPCore_dispatch( pChildHTTP, PH_MAPPING, PH_MAPPING );

		/* --- copy the childs script name to variable script name --- */
		if ( iRet != PIAPI_COMPLETED ) {
			PIHTTP_delete( pChildHTTP );
			return NULL;
		};

		char *pPath = (char *)PIDB_lookup( pChildHTTP->pResponseDB,
			PIDBTYPE_STRING, KEY_INT_PATH, 0 );

		PIHTTP_delete( pChildHTTP );
		if (!pPath) return NULL;

		while (strchr(pPath, '\\'))
			{
			pPath[strchr(pPath,'\\')-pPath] = '/';
			}

#if WIN32 
	// assume leading slash before the drive letter
	int offset = (URL[7]== '/') ? 8 : 7;
#else
	int offset = 7;
#endif
		// check, if the URI is a subpath of the WebRoot
		return strncmp( xmlURIUnescapeString(URL, 0, NULL) + offset,
			pPath, strlen(pPath)) ? NULL : OriginEntityLoader( URL, ID, context );
#undef OFFSET
	} else {
		// assume, we're not in a web server request
		return OriginEntityLoader( URL, ID, context );
	}
}

/*____________________________________________________________________________*\
 *
 Class:			XSLT
 Description:	The XSLT handler class
\*____________________________________________________________________________*/
class XSLT : public HandlerBaseXSLT
{
private:
	/* ---
	Configuration data
	--- */
	PIDB *pTypes;					/* DB for variable types */
   	PIString sExtraHeadersPrefix;	/* Extra Header for environment values, */
									/* by default "HTTP_" */
	int iExtraHeaders;				/* whether or not to send extra headers */
	DblList lIgnoreHeaders;			/* extra headers to ignore */
	xsltStylesheetPtr default_style;
	int iOptions;
	int iParams;
	PIString sContentType;

	inline int checkIgnoreList(const char *pKey)
		{
		/* --- exclude the variables in lIgnoreHeaders --- */
		const char *pExclude;
		for( DblListIterator l( lIgnoreHeaders ); !l.BadIndex(); l++ )
			{
			pExclude = (const char *)l.Current();
			if (!PIUtil_stricmp(pKey, pExclude))
				{
				return 1;
				}
			};
		return 0;
		}
protected:

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "XSLT: ";
		if ( !PIUtil_stricmp( KEY_CONF_DEFAULTSTYLESHEET, pVariable ) )
			{
			default_style =
				xsltParseStylesheetFile((const xmlChar *)(const char *)DeQuote( pValue ));
			}
		else if ( !PIUtil_stricmp( KEY_CONF_OPTIONS, pVariable ) )
			{
			/* ---
			Split the options up by the bar ('|') character and then
			strip trailing and leading whitespace from option before
			attempting to match it.
			--- */
			StringTokenizer tTokens( pValue, "|" );
			for(int i=0; i<tTokens.NumTokens(); i++)
				{
				const char *pToken = tTokens.GetToken( i );

				/* --- skip leading whitespace --- */
				for( ; *pToken && (isspace(*pToken)); pToken++ );

				/* ---
				include only to first whitespace trailing whitespace
				--- */
				int j=0;
				for( ; pToken[j] && !(isspace(pToken[j])); j++ );
				/* ---
				j now contains the length of the first word in pToken
				--- */

				/* ---
				cycle through the list of available flags
				comparing them with this one
				--- */
				int i=0;
				for( ; aFlagMap[i].pName; i++ )
					{
					if ( !PIUtil_strncmpi( aFlagMap[i].pName, pToken, j ) )
						{ break; };
					};
				if ( !aFlagMap[i].pName )
					{
					/* --- flag not found --- */
					PIString sTmp( pToken, j );
					os << "Unknown option flag '" << sTmp << "'." << ends;
					CONFIG_ERR( Object(), os.str() );
					return 0;
					};
				iOptions |= aFlagMap[i].iFlag;
				};	/* --- loop over tokens seperated by '|' --- */

			}
		else if ( !PIUtil_stricmp( KEY_CONF_VARIABLE, pVariable ) )
			{
			if (iParams < MAX_PARAMETERS - 1)
				{
				const char *pK;
				const char *pV;
				StringTokenizer tTokens( (const char *)DeQuote(pValue), "=" );
				if (tTokens.NumTokens() < 2)
					{
					os << "Invalid Variable '" << pValue <<
						"'" << ends;
					CONFIG_ERR( Object(), os.str() );
					return 0;
					}

				pK = tTokens.GetToken( 0 );
				pV = tTokens.GetToken( 1 );

				Pi3String *pError = Pi3String_new( 0 );
				Pi3Expression *pExpr = Pi3Expression_new( pV, NULL, pError );
				if ( !pExpr )
					{
					os << Pi3String_getPtr( pError ) << ends;
					Pi3String_delete( pError );
					CONFIG_ERR( Object(), os.str() ); 
					return 0;
					};
				Pi3String_delete( pError );
				PIDB_add( pTypes, PIDBTYPE_OPAQUE, pK, (void *)pExpr, 0 );
				}
			else
				{
				os << "Too much parameters" << ends;
				CONFIG_ERR( Object(), os.str() );
				}
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHDRPREFIX, pVariable ) )
			{
				sExtraHeadersPrefix = (const char *)DeQuote( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHEADERSIGNORE, pVariable ) )
			{
			StringTokenizer tTokens( (const char *)DeQuote(pValue), " " );
			for( int i=0; i<tTokens.NumTokens(); i++)
				{
				lIgnoreHeaders.Append( (DblList::type)
					PI_NEW( PIString( tTokens.GetToken( i ) ) ) );
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHEADERS, pVariable ) )
			{
			if ( !PIUtil_stricmp( VALUE_NO, pValue ) )
				{ iExtraHeaders = 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_MAXDEPTH, pVariable ) )
			{
			xsltMaxDepth = atoi( DeQuote(pValue) );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_CONTENTTYPE, pVariable ) )
			{
			sContentType = (const char *)DeQuote( pValue );
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	XSLT( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	
		pTypes( PIDB_new( 0, "Types" ) ),
		iExtraHeaders( 1 ),
		default_style( NULL ),
		iOptions( 0 ),
		iParams( 0 ),
		sContentType( "" ),
		HandlerBaseXSLT( pObject )
		{
		xmlInitMemory();
		LIBXML_TEST_VERSION
		ReadParameters( iArgc, ppArgv );
		if ( !IsOK() ) { return; };

		// Initialization function for the XML parser. This is not reentrant.
		// Call once before processing in case of use in multithreaded programs.
		xmlInitParser();

		// some initialization stuff
		xmlLineNumbersDefault(1);
		xmlSubstituteEntitiesDefault(1);
		exsltRegisterAll();

		// -nonet
		if ( iOptions & FLG_NONET )
			xmlSetExternalEntityLoader(xmlNoNetExternalEntityLoader);

		if (!OriginEntityLoader)
			OriginEntityLoader = xmlGetExternalEntityLoader();
		xmlSetExternalEntityLoader(xmlEntityLoader);

		// -xinclude
		if ( iOptions & FLG_XINCLUDE )
			xsltSetXIncludeDefault(1);

		// -catalogs
		if ( iOptions & FLG_CATALOGS ) {
			const char *catalogs = getenv("SGML_CATALOG_FILES");
			if (catalogs) {
				xmlLoadCatalogs(catalogs);
			} else {
				CONFIG_WARN(pObject, "Environment variable $SGML_CATALOG_FILES not set.");
			}
		}

		/*
		-output			// not supported
		-noout			// not supported
		-repeat			// not supported
		-warnnet		// not supported (by libxml)
		*/
		};

	virtual ~XSLT()
		{
		PIDB_delete( pTypes );
		for( DblListIterator i( lIgnoreHeaders ); !i.BadIndex(); i++ )
			{ PI_DELETE( (PIString *)i.Current() ); };
        xsltCleanupGlobals();
        xmlCleanupParser();
		};


	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &tBuffer )
		{
		if ( iPhase!=PH_HANDLE )
			{ return PIAPI_ERROR; };

		PIDB *pQ = tPIHTTP.pRequestDB;
		PIDB *pR = tPIHTTP.pResponseDB;
		int iRet = PIAPI_ERROR;
		int iFlags;
		int iMethod;
		in_ctx ctx;
		const char *pDocPath, *pDocURI;
		const char *pCLen, *pCType;
        int nParams = 0;
		const char *pParams[MAX_PARAMETERS + 1];
		enum { BUF_SIZE=255 };
		char szBuf [BUF_SIZE+1];
		char *pBuffer = szBuf;
		const char *pKey, *pValue;
		Pi3Expression *pExpr;

        xsltStylesheetPtr cur = NULL;
        xmlDocPtr doc = NULL;
		xmlDocPtr res = NULL;
		const xmlChar *encoding;
		const xmlChar *method;
		xmlChar *buffer;
		xmlCharEncodingHandlerPtr encoder = NULL;
		xmlOutputBufferPtr obuf = NULL;
		xmlParserInputBufferPtr ibuf = NULL;

		// -novalid
		if ( iOptions & FLG_NODTD ) {
			xmlLoadExtDtdDefaultValue = 0;
		} else {
			xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
		};

		// without this embedded stylesheets don't work
		if ( iOptions & FLG_VALIDATE ) {
			xmlDoValidityCheckingDefaultValue = 1;
		} else {
			xmlDoValidityCheckingDefaultValue = 0;
		}

		/* ---
		Get server variables
		--- */
		PIDBIterator *pIter = PIDB_getIterator( pTypes, PIDBTYPE_OPAQUE, 0, 0 );
		if ( !pIter ) {
			HTTPCore_logError( &tPIHTTP, "XSLT: Internal error when creating iterator." );
			goto do_error;
		}

		// the number of parameters has been checked on startup already!
		for(; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter ) )
			{
			pExpr = (Pi3Expression *)PIDBIterator_current( pIter, &pKey );
			assert( pKey && pExpr );	

			int iWritten = Pi3Expression_write( pExpr, &tPIHTTP, NULL,
				(char *)pBuffer, BUF_SIZE );
			if ( iWritten==-1 )
				{
				pBuffer = (char *)PIHTTP_allocMem( &tPIHTTP, iWritten );
				int iWritten = Pi3Expression_write( pExpr, &tPIHTTP, NULL,
					(char *)pBuffer, BUF_SIZE );

				if ( iWritten==-1 )
					{
					HTTPCore_logError( &tPIHTTP, "XSLT: Internal error when writing \
variable expression.");
					goto do_error;
					}
				}

			pBuffer[iWritten] = 0;
			// -param
			pParams[nParams++]=(const char *)xmlStrdup((const xmlChar *)pKey);
			pParams[nParams++]=(const char *)xmlStrdup((const xmlChar *)pBuffer);
			}
		PIDBIterator_delete( pIter );

		/* --- Get extra request headers --- */
		if ( iExtraHeaders ) 
			{
			/* --- loop over all RFC822 headers in request block --- */				
			PIDBIterator *pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );	
			if ( !pIter ) {
				HTTPCore_logError( &tPIHTTP, "XSLT: Internal error when creating iterator." );
				goto do_error;
			}

			for(; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter ) )
				{
				if (nParams < MAX_PARAMETERS - 1)
					{
					pValue = (const char *)PIDBIterator_current( pIter,	&pKey );
					assert( pKey && pValue );											

					if ( !pKey || !pValue ) { /* sanity */ continue; };

					if (checkIgnoreList(pKey)) continue;

					PIString sTest(sExtraHeadersPrefix);
					sTest.Concatenate(pKey);
					char *pBuf = (char *)(const char *)sTest;
					for( int i=0; i<sTest.Len(); i++ )							
						{																
						if ( isalnum( pBuf[i] ) )										
							{ pBuf[i]=toupper(pBuf[i]); }						
						else															
							{ pBuf[i]='_'; };		
						}

					// -stringparam
					buffer = xmlStrdup((const xmlChar *)"'");
					buffer = xmlStrncat(buffer, (const xmlChar *)pValue, strlen(pValue));
					buffer = xmlStrcat(buffer, (const xmlChar *)"'");
					pParams[nParams++]=(const char *)xmlStrdup((const xmlChar *)pBuf);
					pParams[nParams++]=(const char *)buffer;
					}
				else
					{
					HTTPCore_logError( &tPIHTTP, "XSLT: Too much parameters.");
					goto do_error;
					}
				}
			PIDBIterator_delete( pIter );
			}
		pParams[nParams]=NULL;

		xsltSetGenericErrorFunc((void *)&tPIHTTP, xmlErrorCB);

		// -verbose
		if (( iOptions & FLG_VERBOSE ) && HTTPCore_debugEnabled() )
			xsltSetGenericDebugFunc((void *)&tPIHTTP, xmlDebugCB);

		/* ---
		Get file object
		--- */
		pDocPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			KEY_INT_PATH, 0 );

		pDocURI = getFileURI(pDocPath);
			
		/* ---
		Get request method
		--- */
		iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD, 0 );

		if ( iMethod == MD_POST ) {

			/* ---
			Get the Content-Length
			--- */
			pCLen = (const char *)PIDB_lookup(pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0);
			if (!pCLen) {
				HTTPCore_logError( &tPIHTTP, "XSLT: Missing Content-Length \
in POST request." );
				goto do_error;
			};

			/* ---
			Get the Content-Type
			--- */
			pCType = (const char *)PIDB_lookup(pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, 0);
			if (!pCType) {
				HTTPCore_logError( &tPIHTTP, "XSLT: Missing Content-Type \
in POST request." );
				goto do_error;
			};

			ctx.iobuf = &tBuffer;
			ctx.clength = atol(pCLen);
			ctx.state = STATE_NEW;
			ctx.read = 0;

			/* ---
			Fire off the parser
			--- */
			if (!PIUtil_stricmp(pCType, "text/plain"))
				{
				doc = xmlIOParsePlainEntity( 
					(void *)&ctx,
					XML_CHAR_ENCODING_8859_1
					);
				}
			else if (!PIUtil_stricmp(pCType, "application/x-www-form-urlencoded"))
				{
				doc = xmlIOParseFormEntity( 
					(void *)&ctx,
					XML_CHAR_ENCODING_8859_1
					);
				}
			else 
				{
				HTTPCore_logError( &tPIHTTP, "XSLT: Unknown Content-Type \
in POST request." );
				goto do_error;
				}
		} else {

			// -html, -docbook
			if ( iOptions & FLG_HTML ) {
				doc = xmlPi3ParseHtmlFile(pDocURI, NULL);
			} else if ( iOptions & FLG_DOCBOOK ) {
				doc = xmlPi3ParseDocbookFile(pDocURI, NULL);
			} else {
				doc = xmlPi3ParseXmlFile( pDocURI, &tPIHTTP );
			};
		}

		if (!doc) {
			HTTPCore_logError( &tPIHTTP, "XSLT: XML parser error in file \
with path '%s'.", pDocPath );
			goto do_error;
		};

		if ( iOptions & FLG_VALIDATE && !doc->_private) {
			HTTPCore_logDebug( DBG_MED, "XSLT: Document with path '%s' is not \
valid according to it's DTD", pDocPath);
		}
		
		if ( iOptions & FLG_XINCLUDE ) {
			xmlXIncludeProcess(doc);
		}

		cur = default_style ? default_style : xsltLoadStylesheetPI(doc);
		if (!cur) {
			HTTPCore_logError( &tPIHTTP, "XSLT: Can't load stylesheet for file \
with path '%s'.", pDocPath );
			goto do_error;
		};

		res = xsltApplyStylesheet(cur, doc, pParams);
		if (!res) {
			HTTPCore_logError( &tPIHTTP, "XSLT: Can't apply stylesheet on file \
with path '%s'.", pDocPath );
			goto do_error;
		};

		if ( iOptions & FLG_DEBUG )
			xmlDebugDumpDocument(stdout, res);

		XSLT_GET_IMPORT_PTR(encoding, cur, encoding)

		if (encoding != NULL) {
			encoder = xmlFindCharEncodingHandler((char *)encoding);
			if ((encoder != NULL) && (xmlStrEqual((const xmlChar *)encoder->name,
				(const xmlChar *) "UTF-8"))) encoder = NULL;
		};

		obuf = xmlOutputBufferCreateIO(
			xmlWriteCB,
			xmlOutCloseCB,
			(void *)&tBuffer,
			encoder
		);

		if (strlen(sContentType)) {
			// Overwrite content-type with configuration value
			PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
				(void *)(const char *)sContentType, 0);
		} else {
			/* ---  content-type --- */
			XSLT_GET_IMPORT_PTR(method, cur, method)

		    if ((method == NULL) && (res->type == XML_HTML_DOCUMENT_NODE))
				method = (const xmlChar *) "html";

			if (method != NULL) {
		        if (xmlStrEqual(method, (const xmlChar *) "xml")) {
					PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, "application/xml", 0);
				} else if (xmlStrEqual(method, (const xmlChar *) "html")) {
					PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, "text/html", 0);
				} else if (xmlStrEqual(method, (const xmlChar *) "xhtml")) {
					PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, "text/html", 0);
				} else if (xmlStrEqual(method, (const xmlChar *) "text")) {
					PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, "text/plain", 0);
				// unknown content-type, use server default
				} else {
					PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
						(void *)HTTPCore_getMIMETypeFromExtension(NULL), 0);
				}
			// unknown content-type, use server default
			} else {
				PIDB_replace(pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE,
					(void *)HTTPCore_getMIMETypeFromExtension(NULL), 0);
			}
		}

		/* --- transfer-encoding --- */
		iFlags = (!PIHTTP_isChild( &tPIHTTP) && (int)PIDB_lookup( tPIHTTP.pRequestDB,
			PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, 0 ) == 3 ) ? PIIOBUF_CHUNKED : PIIOBUF_NONE;
		if ( iFlags & PIIOBUF_CHUNKED )
			{
			PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_TRANSFERENCODING,
				(void *)KEY_HTTP_CHUNKED, 0 );
			}
		else
			{
			/* ---
			disable keep open, assuming content-length cannot be set
			--- */
			PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, 
				KEY_INT_KEEPOPEN, (void *)0, 0 );
			};

		HTTPCore_sendGeneralHeaders(&tPIHTTP);
		HTTPCore_sendEntityHeaders(&tPIHTTP, pR);
		if ( iFlags & PIIOBUF_CHUNKED ) PIIOBuffer_flush(&tBuffer);
		PIIOBuffer_setUserData(&tBuffer, (void *)iFlags);

		if (xsltSaveResultTo(obuf, res, cur) < 0)
			{
#define TMP "XSLT: Error on writing XSLT output."
			PIIOBuffer_write(&tBuffer, TMP, strlen(TMP), iFlags);
#undef TMP
			};

		iRet = PIAPI_COMPLETED;
		goto done;

do_error:
		HTTPCore_logError( &tPIHTTP, "XSLT: Using libxml %s, libxslt %s and \
libexslt %s.", xmlParserVersion, xsltEngineVersion, exsltLibraryVersion);

		HTTPCore_logError( &tPIHTTP, "XSLT: Handler was compiled against libxml %d, \
libxslt %d and libexslt %d.", LIBXML_VERSION, LIBXSLT_VERSION, LIBEXSLT_VERSION);

		HTTPCore_logError( &tPIHTTP, "XSLT: libxslt %d was compiled against libxml %d.",
			xsltLibxsltVersion, xsltLibxmlVersion);

		HTTPCore_logError( &tPIHTTP, "XSLT: libexslt %d was compiled against libxml %d \
and libxslt %d.", exsltLibexsltVersion, exsltLibxmlVersion, exsltLibxsltVersion);

		iRet = HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR);

done:
	    for (int i=0;i<nParams;i++)
			xmlFree((void *)pParams[i]);
		// don't free default stylesheet!
        if (cur && (cur != default_style)) xsltFreeStylesheet(cur);
        if (doc) xmlFreeDoc(doc);
        if (res) xmlFreeDoc(res);
		xmlCleanupParser();
		return iRet;
		};
};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int XSLT_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseXSLT_constructor( pObj, PI_NEW( XSLT( pObj,
		iArgc, ppArgv ) ) );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int XSLT_destructor( PIObject *pObj, int,
	const char *[] )
{
	delete (XSLT *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}


#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name XSLTClass
		Type LogicExtension
		Library XSLT
		OnClassLoad HandlerBaseXSLT_onClassLoad
		Constructor XSLT_constructor
		CopyConstructor HandlerBaseXSLT_copyConstructor
		Destructor XSLT_destructor
		Execute HandlerBaseXSLT_execute
	</Class>

	<Object>
		Name XSLT
		Class XSLTClass
	</Object>

/*___+++CNF_END+++___*/
#endif

